// mba.c - MASSBUS Interface Controller Routines
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS10 Emulator.
// See ReadMe for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "emu/defs.h"
#include "dev/mba/mba.h"
#include "dev/mba/mba_rp.h"

extern MBA_DEVINFO rp2_DriveInfo;

// Listing of Drive Information/Calls
MBA_DEVINFO *mba_DriveInfo[] =
{
	&rp2_DriveInfo,    // RP - Disk Pack Drive Series
//	&rm2_DriveInfo,    // RM - Disk Pack Drive Series
//	&tm2_DriveInfo,    // TM - Tape Formattor/Slave Drive Series
	NULL
};

// Create MBA device
void *mba_Create(void *dptr, MBA_CALLBACK *cptr,
	int idDevice, int argc, char **argv)
{
	MBA_DEVICE *mba;
	int idx;

	// Create new MBA device;
	mba = (MBA_DEVICE *)calloc(1, sizeof(MBA_DEVICE));
	mba->idDevice = idDevice;
	mba->pDevice  = dptr;
	mba->Callback = cptr;
	for (idx = 0; idx < MBA_MAXDRIVES; idx++) {
		mba->Drive[idx].idDrive  = idx;
		mba->Drive[idx].wDevice  = mba;
		mba->Drive[idx].pDevice  = dptr;
		mba->Drive[idx].Callback = cptr;
	}

	return (void *)mba;
}

// Create MBA drive for desired MBA device.
void *mba_CreateDrive(void *dptr, char *devName, int argc, char **argv)
{
	MBA_DEVICE  *mba = (MBA_DEVICE *)dptr;
	MBA_DRIVE   *mbaDrive;
	MBA_DEVINFO *drvInfo = NULL;
	DTYPE       *drvType = NULL;
	int         drvSlot, idx;

	if (mba == NULL) {
		printf("%s: Device not created.\n", devName);
		return NULL;
	}

	// Find a drive type like RP, RM, TM, etc.
	for (idx = 0; mba_DriveInfo[idx]->Name; idx++) {
		int len = strlen(mba_DriveInfo[idx]->Name);
		if (!strncasecmp(argv[0], mba_DriveInfo[idx]->Name, len)) {
			drvInfo = mba_DriveInfo[idx];
			break;
		}
	}
	if (drvInfo == NULL) {
		printf("%s: Drive not found.\n", devName);
		return NULL;
	}

	// Find a model like RP04, etc.
	for (idx = 0; drvInfo->dTypes[idx].Name; idx++) {
		if (!strcasecmp(argv[0], drvInfo->dTypes[idx].Name)) {
			drvType = &drvInfo->dTypes[idx];
			break;
		}
	}
	if (drvType == NULL) {
		printf("%s: Model number not found.\n", devName);
		return NULL;
	}

	// Check the drive is present or not first before set up it.
	sscanf(argv[1], "%o", &drvSlot);
	mbaDrive = &mba->Drive[drvSlot];
	if (mbaDrive->Flags < 0) {
		printf("%s: Drive present.\n", devName);
		return NULL;
	}

	// Now set up a new drive slot...
	mbaDrive->Flags = MBA_PRESENT;
	mbaDrive->dInfo = drvInfo;
	mbaDrive->dType = drvType;
	mbaDrive->dInfo->InitDrive(mbaDrive);

	printf("%s: Device %s (%s) - Created.\n",
		devName, drvType->Name, drvInfo->Name);
	return (void *)mbaDrive;
}

int mba_Attach(void *dptr, int drv, char *fileName)
{
	MBA_DEVICE *mba      = (MBA_DEVICE *)dptr;
	MBA_DRIVE  *mbaDrive = &mba->Drive[drv];

	if (mbaDrive->Flags < 0) {
		return mbaDrive->dInfo->Attach(mbaDrive, fileName);
	} else {
		printf("MBA(%d): No such drive %d.\n", mba->idDevice, drv);
		return EMU_ATTACHED;
	}	
}

int mba_Detach(void *dptr, int drv)
{
	MBA_DEVICE *mba      = (MBA_DEVICE *)dptr;
	MBA_DRIVE  *mbaDrive = &mba->Drive[drv];

	if (mbaDrive->Flags < 0) {
		return mbaDrive->dInfo->Detach(mbaDrive);
	} else {
		printf("MBA(%d): No such drive %d.\n", mba->idDevice, drv);
		return EMU_ATTACHED;
	}	
}

// Send attention signal to its controller
// to do an interrupt.
void mba_Signal(MBA_DEVICE *mba)
{
//	rh20_Signal(mba->pDevice);
}

// Load register
void mba_WriteIO(MBA_DEVICE *mba, int drv, int reg, uint16 data)
{
	MBA_DRIVE *mbaDrive = &mba->Drive[drv];

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("MBA(%d): (W) Drive %d  Register %02o  Data %06o\n",
			mba->idDevice, drv, reg, data);
#endif /* DEBUG */

	if (mbaDrive->Flags < 0) {
		mbaDrive->dInfo->WriteIO(mbaDrive, reg, data);
	}
}

void mba_ReadIO(MBA_DEVICE *mba, int drv, int reg, uint16 *data)
{
	MBA_DRIVE *mbaDrive = &mba->Drive[drv];

	if (mbaDrive->Flags < 0) {
		mbaDrive->dInfo->ReadIO(mbaDrive, reg, data);
	} else
		*data = 0;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("MBA(%d): (R) Drive %d  Register %02o  Data %06o\n",
			mba->idDevice, drv, reg, *data);
#endif /* DEBUG */
}
