// rp.c - RP MASSBUS-based disk drive routines
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS-10 Emulator.
// See README for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "emu/defs.h"
#include "dev/mba/mba.h"
#include "dev/mba/mba_rp.h"

// This controller supports many different disk drive types:
//
// Note: 16/18-bit words in sectors per surface
//
// Type     # Sectors/   # Tracks/   # Cylinders/    Total
//           Surface     Cyclinder       Drive
//
// RP04/05    22/20         19            411      =  88 MB
// RP06       22/20         19            815      = 176 MB
// RP07       50/43         32            630      = 516 MB

DTYPE rp2_dTypes[] = {
	{
		"RP04",               // Device Name
		"RP04 - 88MB Disk Pack Drive",
		UNIT_REMOVABLE|UNIT_ATTABLE|UNIT_DISABLE,
		20,                   // Number of Sectors
		14,                   // Number of Tracks
		411,                  // Number of Cylinders
		(20 * 14 * 411),      // Total Blocks
		020020,               // Device Type
		NULL,                 // Not Used

		MBA_PACK,            // RH11 Flags - Disk Pack Type
		NULL,                 // Not Used

		disk_Open,            // Low-level Open Routine
		disk_Close,           // Low-level Close Routine
		disk_Read,            // Low-level Read Routine
		disk_Write,           // Low-level Write Routine
		NULL,                 // Low-level Mark Routine
		NULL,                 // Low-level Rewind Routine
		NULL,                 // Low-level Skip Routine
		disk_Seek,            // Low-level Seek Routine
		disk_GetDiskAddr      // Low-level GetDiskAddr Routine
	},

	{
		"RP05",               // Device Name
		"RP05 - 88MB Disk Pack Drive",
		UNIT_REMOVABLE|UNIT_ATTABLE|UNIT_DISABLE,
		20,                   // Number of Sectors
		14,                   // Number of Tracks
		411,                  // Number of Cylinders
		(20 * 14 * 411),      // Total Blocks
		020021,               // Device Type
		NULL,                 // Not Used

		MBA_PACK,            // RH11 Flags - Disk Pack Type
		NULL,                 // Not Used

		disk_Open,            // Low-level Open Routine
		disk_Close,           // Low-level Close Routine
		disk_Read,            // Low-level Read Routine
		disk_Write,           // Low-level Write Routine
		NULL,                 // Low-level Mark Routine
		NULL,                 // Low-level Rewind Routine
		NULL,                 // Low-level Skip Routine
		disk_Seek,            // Low-level Seek Routine
		disk_GetDiskAddr      // Low-level GetDiskAddr Routine
	},

	{
		"RP06",               // Device Name
		"RP06 - 176MB Disk Pack Drive",
		UNIT_REMOVABLE|UNIT_ATTABLE|UNIT_DISABLE,
		20,                   // Number of Sectors
		19,                   // Number of Tracks
		815,                  // Number of Cylinders
		(20 * 19 * 815),      // Total Blocks
		020022,               // Device Type
		NULL,                 // Not Used

		MBA_PACK,            // RH11 Flags - Disk Pack Type
		NULL,                 // Not Used

		disk_Open,            // Low-level Open Routine
		disk_Close,           // Low-level Close Routine
		disk_Read,            // Low-level Read Routine
		disk_Write,           // Low-level Write Routine
		NULL,                 // Low-level Mark Routine
		NULL,                 // Low-level Rewind Routine
		NULL,                 // Low-level Skip Routine
		disk_Seek,            // Low-level Seek Routine
		disk_GetDiskAddr      // Low-level GetDiskAddr Routine
	},

	{
		"RP07",               // Device Name
		"RP07 - 516MB Fixed Disk Drive",
		UNIT_ATTABLE|UNIT_DISABLE,
		43,                   // Number of Sectors
		32,                   // Number of Tracks
		630,                  // Number of Cylinders
		(43 * 32 * 630),      // Total Blocks
		020024,               // Device Type

		NULL,                 // Not Used

		MBA_FIXED,           // RH11 Flags - Fixed Disk Type
		NULL,                 // Not Used

		disk_Open,            // Low-level Open Routine
		disk_Close,           // Low-level Close Routine
		disk_Read,            // Low-level Read Routine
		disk_Write,           // Low-level Write Routine
		NULL,                 // Low-level Mark Routine
		NULL,                 // Low-level Rewind Routine
		NULL,                 // Low-level Skip Routine
		disk_Seek,            // Low-level Seek Routine
		disk_GetDiskAddr      // Low-level GetDiskAddr Routine
	},

	{ NULL },
};

#ifdef DEBUG
static char *regNames[] =
{
	"RPCS1", // (R/W) Control and Status Register #1
	"RPDS",  // (R)   Drive Status Register
	"RPER1", // (R)   Error Register #1
	"RPMR",  // (R/W) Maintenance Register
	"RPAS",  // (R/W) Attention Summary Register
	"RPDA",  // (R/W) Desired Sector/Track Address Register
	"RPDT",  // (R)   Drive Type Register
	"RPLA",  // (R)   Look-Ahead Register
	"RPSN",  // (R)   Serial Number Register
	"RPOF",  // (R/W) Offset Register
	"RPDC",  // (R/W) Desired Cylinder Address Register
	"RPCC",  // (R/W) Current Cylinder Address Register
	"RPER2", // (R)   Error Register #2
	"RPER3", // (R/W) Error Register #3
	"RPEC1", // (R)   ECC Position Register
	"RPEC2", // (R)   ECC Pattern Register
};

static char *fncNames[] =
{
	"No Operation",          // 00
	"Unload",                // 01
	"Seek",                  // 02
	"Recalibrate",           // 03
	"Drive Clear",           // 04
	"Port Release",          // 05
	"Offset",                // 06
	"Return to Clearline",   // 07
	"Read-In Preset",        // 10
	"Pack Acknowledge",      // 11
	"*Unknown",              // 12
	"*Unknown",              // 13
	"Search",                // 14
	"*Unknown",              // 15
	"*Unknown",              // 16
	"*Unknown",              // 17
	"*Unknown",              // 20
	"*Unknown",              // 21
	"*Unknown",              // 22
	"*Unknown",              // 23
	"Verify Data",           // 24 Write Check Data
	"Verify Header",         // 25 Write Check Header
	"*Unknown",              // 26
	"*Unknown",              // 27
	"Write Data",            // 30
	"Write Header",          // 31
	"*Unknown",              // 32
	"*Unknown",              // 33
	"Read Data",             // 34
	"Read Header",           // 35
	"*Unknown",              // 36
	"*Unknown"               // 37 
};

#endif /* DEBUG */

// Valid register for access
static char rp2_regValid[] = {
	1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
	0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
};
 
static char rp2_regWrite[] = {
	1, 0, 0, 1,  1, 1, 0, 0,  0, 1, 1, 1,  0, 1, 0, 0,
	0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
};

uint8 rp2_cBuffer[RP_BLKSZ18];

void rp2_SetAttention(MBA_DRIVE *);
void rp2_SetReady(MBA_DRIVE *);

// Power Up Initialization
DEVINFO rp2_DeviceInfo;
void    rp2_Process(MBA_DRIVE *);

void rp2_InitDrive(MBA_DRIVE *mbaDrive)
{
	CLK_QUEUE *newTimer;

	mbaDrive->devInfo = &rp2_DeviceInfo;

	// Set new timer
	newTimer = (CLK_QUEUE *)calloc(1, sizeof(CLK_QUEUE));
	newTimer->Next     = NULL;
	newTimer->Flags    = 0;
	newTimer->outTimer = 10;
	newTimer->nxtTimer = 10;
	newTimer->Device   = (void *)mbaDrive;
	newTimer->Execute  = rp2_Process;
	mbaDrive->Timer    = newTimer;

	// Initialize drive registers
	mbaDrive->RPCS1 = RPCS1_DVA; // DVA - Always On.
	mbaDrive->RPDS  = RPDS_DPR;  // DPR - Always On.
	mbaDrive->RPDT  = mbaDrive->dType->dType;
	mbaDrive->RPSN  = 01000 + mbaDrive->idDrive;
	mbaDrive->RPDA  = 0;
	mbaDrive->RPER1 = 0;
	mbaDrive->RPER2 = 0;
	mbaDrive->RPER3 = 0;
}

void rp2_ResetDrive(MBA_DRIVE *mbaDrive)
{
	ts10_CancelTimer(mbaDrive->Timer);

	mbaDrive->RPCS1 = RPCS1_DVA; // DVA - Always On.
	mbaDrive->RPDA  = 0;
	mbaDrive->RPER1 = 0;
	mbaDrive->RPER2 = 0;
	mbaDrive->RPER3 = 0;
}

int rp2_Attach(MBA_DRIVE *mbaDrive, char *fileName)
{
	int Blocks;
	int rc;

	if (mbaDrive->Flags & MBA_ATTACHED)
		return EMU_ATTACHED;

	// Attach a file to that drive slot.
	Blocks  = mbaDrive->dType->Blocks;
	if (rc = disk_Open(&mbaDrive->FileRef, fileName, 0, Blocks, RP_BLKSZ18))
		return rc;

	mbaDrive->Flags |= MBA_ATTACHED;

	// Now let's set some registers for medium on-line
	mbaDrive->RPDS |= (RPDS_ATA|RPDS_MOL|RPDS_DRY|RPDS_VV);
	rp2_SetAttention(mbaDrive);
	rp2_SetReady(mbaDrive);

	return EMU_OK;
}

int rp2_Detach(MBA_DRIVE *mbaDrive)
{
	int rc = EMU_OK;

	if ((mbaDrive->Flags & MBA_ATTACHED) == 0)
		return EMU_ATTACHED;

	// Detach a file from that drive slot.
	rc = disk_Close(&mbaDrive->FileRef);

	mbaDrive->Flags &= ~MBA_ATTACHED;

	// Now let's set some registers for medium off-line
	mbaDrive->RPDS &= ~(RPDS_MOL|RPDS_WRL|RPDS_DRY|RPDS_VV);

	return rc;
}

void rp2_SetAttention(MBA_DRIVE *mbaDrive)
{
	// Set attention bit on drive status register and
	// inform its controller that attention is needed.
	mbaDrive->RPDS |= RPDS_ATA;
	mbaDrive->Callback->SetAttention(mbaDrive->pDevice);
}

void rp2_SetReady(MBA_DRIVE *mbaDrive)
{
	// Set ready/done bit on drive status register and
	// inform its controller that command or drive is
	// ready or done. 
	mbaDrive->RPDS |= RPDS_DRY;
	mbaDrive->Callback->SetReady(mbaDrive->pDevice);
}

int rp2_WriteBlock(MBA_DRIVE *mbaDrive)
{
	uint8  Data8[640];        // 640 bytes 8-bit data buffer
	uint18 blkData[RP_BLKSZ]; // 16/18-bit data buffer
	int    i, j, rc;

	rc = mbaDrive->Callback->ReadBlock(mbaDrive->pDevice, blkData, RP_BLKSZ);
	if (rc == EMU_OK) {
		for (i = 0, j = 0; i < RP_BLKSZ; j += 5, i += 2) {
			// Convert to left halfword
			Data8[j]   =  blkData[i] >> 10;
			Data8[j+1] =  blkData[i] >> 2;
			Data8[j+2] =  blkData[i] << 6;

			// Convert to right halfword
			Data8[j+2] |= blkData[i+1] >> 12;
			Data8[j+3] =  blkData[i+1] >> 4;
			Data8[j+4] =  blkData[i+1] & 0xF;
		}

		rc = disk_Write(mbaDrive->FileRef, Data8);
	}
	return rc;
}

int rp2_CheckBlock(MBA_DRIVE *mbaDrive)
{
	int36  data1, data2;
	int    idx;
	int    st = EMU_OK;

	if ((st = disk_Read(mbaDrive->FileRef, rp2_cBuffer)) >= 0) {
		for (idx = 0; idx < RP_BLKSZ18; idx += 5) {
			data1 = Convert8to36(&rp2_cBuffer[idx]);
//			rhDevice->ReadData36(rhUnit, &data2, FALSE);
			if (data1 != data2)
				return -2;
//			if (st = rhDevice->CheckWordCount(rhUnit))
//				return st;
		}
	}

	return st;
}

int rp2_ReadBlock(MBA_DRIVE *mbaDrive)
{
	uint8  Data8[640];        // 640 bytes 8-bit data buffer
	uint18 blkData[RP_BLKSZ]; // 16/18-bit data buffer
	int    i, j, rc = EMU_OK;

	if ((rc = disk_Read(mbaDrive->FileRef, Data8)) >= 0) {
		for (i = 0, j = 0; i < RP_BLKSZ; j += 5, i += 2) {
			// Convert to left halfword
			blkData[i]   =  Data8[j] << 10;
			blkData[i]   |= Data8[j+1] << 2;
			blkData[i]   |= Data8[j+2] >> 6;

			// Convert to right halfword
			blkData[i+1] =  (Data8[j+2] & 0x3F) << 12;
			blkData[i+1] |= Data8[j+3] << 4;
			blkData[i+1] |= Data8[j+4] & 0xF;
		}

		rc = mbaDrive->Callback->WriteBlock(mbaDrive->pDevice, blkData, RP_BLKSZ);
	}

	return rc;
}

void rp2_Process(MBA_DRIVE *mbaDrive)
{
	DTYPE  *mbaType  = mbaDrive->dType;
	char   *diskName = mbaType->Name;
	int    fnc       = (mbaDrive->RPCS1 & RPCS1_FUNC) >> 1;
	int    diskAddr;
	int    temp, st;
	int    savedMode = 0; // temp.

#ifdef DEBUG
	if (dbg_Check(DBG_IODATA))
		dbg_Printf("%s: Drive %d - Process: %s\n",
			diskName, mbaDrive->idDrive, fncNames[fnc]);
#endif /* DEBUG */

	if (mbaDrive->RPER1 || mbaDrive->RPER2 || mbaDrive->RPER3) {
		mbaDrive->RPDS |= RPDS_ERR;
		if (mbaDrive->RPER2 || mbaDrive->RPER3)
			mbaDrive->RPER1 |= RPER1_UNS;
	}

	switch (fnc) {
		case FNC_RECAL:
		case FNC_SEEK:
		case FNC_SEARCH:
			diskAddr = GetDiskAddr(mbaDrive->RPDC, mbaDrive->RPDA, mbaType);

			st = disk_Seek(mbaDrive->FileRef, diskAddr, SEEK_SET);
#ifdef DEBUG
			if (dbg_Check(DBG_IODATA) && st >= 0)
				dbg_Printf("%s: Unit %d - Move to disk address %08X\n",
					diskName, mbaDrive->idDrive, st);
#endif /* DEBUG */
			mbaDrive->RPCC = mbaDrive->RPDC;
			mbaDrive->RPDS &= ~RPDS_PIP;
			rp2_SetAttention(mbaDrive);
			break;

		case FNC_OFFSET:
		case FNC_RETURN:
			mbaDrive->RPDS &= ~RPDS_PIP;
			rp2_SetAttention(mbaDrive);
			break;

		case FNC_WR_DATA:
			if (mbaDrive->RPDS & RPDS_WRL) {
#ifdef DEBUG
				if (dbg_Check(DBG_IODATA))
					dbg_Printf("%s: Unit %d - Write Lock Error\n",
						diskName, mbaDrive->idDrive);
#endif /* DEBUG */
				mbaDrive->RPER1 |= RPER1_WLE;
				break;
			}

		case FNC_CHK_DATA:
		case FNC_RD_DATA:
			diskAddr = GetDiskAddr(mbaDrive->RPDC, mbaDrive->RPDA, mbaType);

			if (diskAddr >= mbaType->Blocks) {
				mbaDrive->RPER1 |= RPER1_AOE;
				break;
			}

			st = disk_Seek(mbaDrive->FileRef, diskAddr, SEEK_SET);
			if (st >= 0) {
				switch (fnc) {
					case FNC_WR_DATA:
						st = rp2_WriteBlock(mbaDrive);
						break;

					case FNC_CHK_DATA:
						st = rp2_CheckBlock(mbaDrive);
						break;

					case FNC_RD_DATA:
						st = rp2_ReadBlock(mbaDrive);
						break;
				}

				// Update RPDA and RPDC registers
				diskAddr++;
				temp = diskAddr % mbaType->Sectors;
				if (temp == (mbaType->Sectors - 1))
					mbaDrive->RPDS |= RPDS_LST;
				else
					mbaDrive->RPDS &= ~RPDS_LST;
				mbaDrive->RPDA = temp;
				temp = diskAddr / mbaType->Sectors;
				mbaDrive->RPDA |= (temp % mbaType->Tracks) << RPDA_P_TA;
				mbaDrive->RPDC = temp / mbaType->Tracks;
			}

			break;
	}

	// Check any errors. If so, set attention bit.
	if (mbaDrive->RPER1 || mbaDrive->RPER2 || mbaDrive->RPER3) {
		mbaDrive->RPDS |= RPDS_ERR;
		rp2_SetAttention(mbaDrive);
	}

	// When process is done, GO bit had been turned off.
	// Also, inform MASSBUS controller this drive is ready now.
	mbaDrive->RPCS1 &= ~RPCS1_GO;
	rp2_SetReady(mbaDrive);
}

void rp2_Execute(MBA_DRIVE *mbaDrive)
{
	DTYPE  *mbaType  = mbaDrive->dType;
	char   *diskName = mbaType->Name;
	int    fnc       = (mbaDrive->RPCS1 & RPCS1_FUNC) >> 1;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("%s: Drive %d - Execute: %s\n",
			diskName, mbaDrive->idDrive, fncNames[fnc]);
#endif /* DEBUG */

	if (fnc != FNC_DCLR) {
		if (mbaDrive->RPDS & RPDS_ERR) {
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("%s: Unit %d - Function isn't accetped when DS_ERR is set.\n",
					diskName, mbaDrive->idDrive);
#endif /* DEBUG */
			return;
		}

		if (!(mbaDrive->RPDS & RPDS_MOL)) {
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("%s: Unit %d - Medium is not on-line!\n",
					diskName, mbaDrive->idDrive);
#endif /* DEBUG */
			mbaDrive->RPER1 |= RPER1_UNS;
			mbaDrive->RPDS  |= RPDS_ERR;
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			rp2_SetAttention(mbaDrive);
			return;
		}
	}

	mbaDrive->RPDS &= ~RPDS_LST;

	switch (fnc) {
		case FNC_NOP:
		case FNC_RELEASE:
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			break;

		case FNC_RECAL:
			mbaDrive->RPDC = 0;

		case FNC_SEEK:
		case FNC_SEARCH:
			if ((GetCylinder(mbaDrive->RPDC) >= mbaType->Cylinders) ||
			    (GetTrack(mbaDrive->RPDA)    >= mbaType->Tracks)    ||
			    (GetSector(mbaDrive->RPDA)   >= mbaType->Sectors)) {
#ifdef DEBUG
				if (dbg_Check(DBG_IOREGS)) {
					dbg_Printf("%s: Unit %d - Invalid Address Error\n",
						diskName, mbaDrive->idDrive);
					dbg_Printf("%s: C %d, T %d, S %d (Max C %d, T %d, S %d)\n",
						diskName, GetCylinder(mbaDrive->RPDC), GetTrack(mbaDrive->RPDA),
						GetSector(mbaDrive->RPDA), mbaType->Cylinders,
						mbaType->Tracks, mbaType->Sectors);
				}
#endif /* DEBUG */
				mbaDrive->RPER1 |= RPER1_IAE;
				break;
			}
			mbaDrive->RPDS = (mbaDrive->RPDS & ~RPDS_DRY) | RPDS_PIP;
			ts10_SetTimer(mbaDrive->Timer);
//			rp2_Process(mbaDrive);
			break;

		case FNC_DCLR:
			rp2_ResetDrive(mbaDrive);
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			break;

		case FNC_OFFSET:
		case FNC_RETURN:
			mbaDrive->RPDS = (mbaDrive->RPDS & ~RPDS_DRY) | RPDS_PIP;
			ts10_SetTimer(mbaDrive->Timer);
//			rp2_Process(mbaDrive);
			break;

		case FNC_PRESET:
			mbaDrive->RPDA = 0;
			mbaDrive->RPOF = 0;
			mbaDrive->RPDC = 0;
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			break;

		case FNC_PACK:
			mbaDrive->RPDS |= RPDS_VV;
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			break;

		case FNC_WR_DATA:
		case FNC_CHK_DATA:
		case FNC_RD_DATA:
			if ((GetCylinder(mbaDrive->RPDC) >= mbaType->Cylinders) ||
			    (GetTrack(mbaDrive->RPDA)    >= mbaType->Tracks)    ||
			    (GetSector(mbaDrive->RPDA)   >= mbaType->Sectors)) {
#ifdef DEBUG
				if (dbg_Check(DBG_IOREGS)) {
					dbg_Printf("%s: Unit %d - Invalid Address Error\n",
						diskName, mbaDrive->idDrive);
					dbg_Printf("%s: C %d, T %d, S %d (Max C %d, T %d, S %d)\n",
						diskName, GetCylinder(mbaDrive->RPDC), GetTrack(mbaDrive->RPDA),
						GetSector(mbaDrive->RPDA), mbaType->Cylinders,
						mbaType->Tracks, mbaType->Sectors);
				}
#endif /* DEBUG */
				mbaDrive->RPER1 |= RPER1_IAE;
				rp2_SetAttention(mbaDrive);
				break;
			}
			mbaDrive->RPDS &= ~RPDS_DRY;
			ts10_SetTimer(mbaDrive->Timer);
//			rp2_Process(mbaDrive);
			break;

		default:
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("%s: Unit %d - Unknown function code %02o\n", diskName,
					mbaDrive->idDrive, (mbaDrive->RPCS1 & (RPCS1_FUNC|RPCS1_GO)));
#endif /* DEBUG */
			mbaDrive->RPCS1 &= ~RPCS1_GO;
			mbaDrive->RPER1 |= RPER1_ILF;
			rp2_SetAttention(mbaDrive);
			break;
	}
}

void rp2_ReadIO(MBA_DRIVE *mbaDrive, int reg, uint16 *data)
{
	char *diskName = mbaDrive->dType->Name;

	if (rp2_regValid[reg]) {

		if (reg == nRPLA) {
			// Increase sector count by one
			mbaDrive->RPLA += (1 << RPLA_P_SC);
			if ((mbaDrive->RPLA >> RPLA_P_SC) == mbaDrive->dType->Sectors)
				mbaDrive->RPLA = 0;
		}
		if (reg == nRPAS) {
			MBA_DEVICE *mba = mbaDrive->wDevice;
			int idx, ata = 0;

			for (idx = 0; idx < MBA_MAXDRIVES; idx++) {
				if (mba->Drive[idx].MBA_REG(nRPDS) & RPDS_ATA)
					ata |= (1 << idx);
			}
			mbaDrive->RPAS = ata;
		}
		*data = mbaDrive->MBA_REG(reg);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("%s: %s (%02o) => %06o\n",
			diskName, regNames[reg], reg, *data);
#endif /* DEBUG */
	} else {
#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("%s: (R) Unknown MASSBUS Register %02o\n",
				diskName, reg);
#endif /* DEBUG */
		mbaDrive->RPER1 |= RPER1_ILR;
		mbaDrive->RPDS  |= RPDS_ERR;
		rp2_SetAttention(mbaDrive);
	}
}

void rp2_WriteIO(MBA_DRIVE *mbaDrive, int reg, uint16 data)
{
	char *diskName = mbaDrive->dType->Name;

	if (rp2_regValid[reg]) {

		// Update registers
		if (reg == nRPCS1) {
			mbaDrive->RPCS1 = (mbaDrive->RPCS1 & ~RPCS1_WR) | (data & RPCS1_WR);
			if (mbaDrive->RPCS1 & RPCS1_GO) {
				mbaDrive->RPDS &= ~RPDS_ATA;
				rp2_Execute(mbaDrive);
			}
		} else if (reg == nRPAS) {
			MBA_DEVICE *mba = mbaDrive->wDevice;
			int idx;

			for (idx = 0; idx < MBA_MAXDRIVES; idx++) {
				if (data & 1)
					mba->Drive[idx].MBA_REG(nRPDS) &= ~RPDS_ATA;
				data >>= 1;
			}
		} else if (rp2_regWrite[reg])
			mbaDrive->MBA_REG(reg) = data;

#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("%s: %s (%02o) <= %06o\n",
				diskName, regNames[reg], reg, data);
#endif /* DEBUG */

	} else {
#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("%s: (W) Unknown MASSBUS Register %o\n",
				diskName, reg);
#endif /* DEBUG */
		mbaDrive->RPER1 |= RPER1_ILR;
	}

	if (mbaDrive->RPER1 || mbaDrive->RPER2 || mbaDrive->RPER3) {
		mbaDrive->RPDS |= RPDS_ERR;
		if (mbaDrive->RPER2 || mbaDrive->RPER3)
			mbaDrive->RPER1 |= RPER1_UNS;
		rp2_SetAttention(mbaDrive);
	}
}

// ****************************************************************

// Usage: attach <device> <filename>
int rp2_CmdAttach(void *dptr, int argc, char **argv)
{
	if (argc != 1) {
		printf("Usage: attach <device> <filename>\n");
		return EMU_OK;
	}

	// Attach a file to that drive slot.
	return rp2_Attach((MBA_DRIVE *)dptr, argv[0]);
}

// Usage: detach <device>
int rp2_CmdDetach(void *dptr, int argc, char **argv)
{
	return rp2_Detach((MBA_DRIVE *)dptr);
}

COMMAND rp2_Commands[] =
{
	{ "attach",    "<filename>",   rp2_CmdAttach },
	{ "detach",    NULL,           rp2_CmdDetach },
	{ "dismount",  NULL,           rp2_CmdDetach },
	{ "mount",     "<filename>",   rp2_CmdAttach },
	{ NULL }
};

DEVINFO rp2_DeviceInfo = 
{
	"RP",                     // Device Name
	"Disk Pack Drive Series", // Description
	"0.2 (Alpha)",            // Version
	rp2_dTypes,               // Listing of RP devices
	DF_USE,                   // Device Flags
	DT_DEVICE,                // Device Type

	rp2_Commands,             // Command Table
	NULL,                     // Set Command Table
	NULL,                     // Show Command Table

	// Command Table
	rp2_CmdAttach,   // Mount/Attach Command Routine
	rp2_CmdDetach,   // Dismount/Detach Command Routine
	NULL,            // Info Command Routine
};

MBA_DEVINFO rp2_DriveInfo =
{
	"RP",                     // Device Name
	"Disk Pack Drive Series", // Description
	"0.2 (Alpha)",            // Version
	rp2_dTypes,               // Listing of RP devices

	rp2_InitDrive,  // InitDrive Routine
	rp2_ResetDrive, // ResetDrive Routine
	rp2_Attach,     // Mount/Attach Routine
	rp2_Detach,     // Dismount/Detach Routine
	rp2_ReadIO,     // Read I/O Routine
	rp2_WriteIO,    // Write I/O Routine
};
