// dev_dte.c - DTE20 Ten-Eleven Interface routines
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS10 Emulator.
// See ReadMe for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// Device Name: DTE (4 Channels)
// Device Code: 200,204,210,214

#include "pdp10/defs.h"
#include "emu/proto.h"
#include "emu/socket.h"
#include "pdp10/kl10.h"
#include "pdp10/dte20.h"
#include "pdp10/proto.h"

#ifdef HAVE_SIGACTION
#include <sys/file.h>
#include <unistd.h>
#include <stropts.h>
#include <signal.h>
#endif /* HAVE_SIGACTION */

#include <time.h>
#include <sys/time.h>

#if defined(linux)
#include <endian.h>
#endif /* linux */

//****************************************************************

// Get EPT base address from KL10 Processor or so.
extern int KL10_eptAddr;

// Send initialization codes to make sure that telnet behave correctly.
static int  n_telnetInit = 15;
static char telnetInit[] =
{
	255, 251,  34, // IAC WILL LINEMODE
	255, 251,   3, // IAC WILL SGA
	255, 251,   1, // IAC WILL ECHO
	255, 251,   0, // IAC WILL BINARY
	255, 253,   0, // IAC DO BINARY
};

void dte20_Process(DTE20_DEVICE *);
static int (*fncExecute[])(DTE20_DEVICE *);

#ifdef DEBUG
static char *fncNames[] =
{
	"*Unknown Function",             // 000
	"TO-11 Initial Message",         // 001
	"TO-10 Initial Message",         // 002
	"String Data",                   // 003
	"Line-Char, Line-Char",          // 004
	"Request Device Status",         // 005
	"*Unknown Function",             // 006
	"Here is Device Status",         // 007
	"*Unknown Function",             // 010
	"Request Date/Time",             // 011
	"Here is Date/Time",             // 012
	"Flush Output",                  // 013
	"Send All TTYs",                 // 014
	"Dataset Connected",             // 015
	"Hang Up Dataset",               // 016
	"Acknowledge",                   // 017
	"XOFF TTY Line",                 // 020
	"XON TTY Line",                  // 021
	"Here are Line Speeds for TTYs", // 022
	"Here are Allocations",          // 023
	"TO10 Reboot Information",       // 024
	"Acknowledge All",               // 025
	"Turn Device (On/Off)",          // 026
	"Enable/Disable Remotes",        // 027
	"Load Line Printer RAM",         // 030
	"Load Line Printer VFU",         // 031
	"DAS60 Data",                    // 032
	"KLINIK Parameter Status",       // 033
	"Enable/Disable Auto-XOFF",      // 034
	"TO11 Break Through Char Data",  // 035
	"TO11 Turn On Debug Mode",       // 036
	"TO11 Turn Off Debug Mode",      // 037
};
#endif /* DEBUG */

void dte20_Initialize(DTE20_DEVICE *);
void dte20_ctyAccept(SOCKET *);
void dte20_ctyEof(SOCKET *, int, int);
void dte20_ctyInput(SOCKET *, char *, int);

void *dte20_Create(void *dptr, char *devName, int argc, char **argv)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;
	CLK_QUEUE    *newTimer;

	if (dte20 == NULL) {
		int idDevice = argc;

		dte20 = (DTE20_DEVICE *)calloc(1, sizeof(DTE20_DEVICE));
		dte20->idDevice  = idDevice;
		dte20->idChannel = idDevice + 010;
		dte20->Protocol  = PRIMARY;
		dte20->srFlags   = 0;
		dte20->mtdFlags  = 0;

		newTimer = (CLK_QUEUE *)calloc(1, sizeof(CLK_QUEUE));
		newTimer->Next     = NULL;
		newTimer->Flags    = 0;
		newTimer->outTimer = 20;
		newTimer->nxtTimer = 20;
		newTimer->Device   = (void *)dte20;
		newTimer->Execute  = dte20_Process;
		dte20->Timer       = newTimer;

		dte20_Initialize(dte20);
	} else {
		printf("DTE(%d): Already created.\n", dte20->idDevice);
	}

	return (void *)dte20;
}

void dte20_Initialize(DTE20_DEVICE *dte20)
{
	SOCKET *newServer;

	// Set up the listing socket for CTY device.
	newServer = sock_Open(CTY_PORT, NET_SERVER);
	if (newServer != NULL) {
		newServer->maxConns = 1;
		newServer->nConns   = 0;
		newServer->Accept   = dte20_ctyAccept;
		newServer->Eof      = NULL;
		newServer->Process  = NULL;
		newServer->Device   = (void *)dte20;
		dte20->ctyServer    = newServer;

		// Now accept incoming connections;
		sock_Listen(newServer, 5);

		// Tell operator that.
		printf("DTE20: Activating console on TCP port %d.\n", CTY_PORT);
		printf("DTE20: Enter 'telnet localhost %d' to log in\n", CTY_PORT);
	}

}

void dte20_Reset(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	dte20->Protocol = PRIMARY;
	dte20->srFlags  = 0;
	dte20->mtdFlags = 0;
}

void dte20_Cleanup(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	sock_Close(dte20->ctySocket);
	sock_Close(dte20->ctyServer);

	dte20->ctySocket = NULL;
	dte20->ctyServer = NULL;
}

void dte20_ctyAccept(SOCKET *srvSocket)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)srvSocket->Device;
	SOCKET *newSocket;

	if (newSocket = sock_Accept(srvSocket)) {
		// First, check if CTY connection already was taken.
		// If so, tell operator that.
		if (dte20->ctySocket != NULL) {

			sock_Print(newSocket, "Console (CTY) connection already was taken.\r\n", 0);
			sock_Print(newSocket, "Check other terminal which has that connection.\r\n", 0);
			sock_Print(newSocket, "\r\nTerminated.\r\n", 0);
			sock_Close(newSocket);

			return;
		}
 
		// Set up the CTY socket connection.
		newSocket->Accept  = NULL;
		newSocket->Eof     = dte20_ctyEof;
		newSocket->Process = dte20_ctyInput;
		newSocket->Device  = (void *)dte20;

		// Reset all buffer for this new connection.
		dte20->idxInQueue   = 0;
		dte20->idxOutQueue  = 0;
		dte20->idxOutBuffer = 0;
		dte20->ctySocket    = newSocket;

		// Send initialization codes and welcome messages
		sock_Print(newSocket, telnetInit, n_telnetInit);
		sock_Print(newSocket, "Welcome to KL10 Emulator\r\n\r\n", 0);
	}
}

void dte20_ctyEof(SOCKET *Socket, int rc, int nError)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)Socket->Device;

	sock_Close(dte20->ctySocket);
	dte20->ctySocket = NULL;
}

void dte20_ctyInput(SOCKET *Socket, char *keyBuffer, int len)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)Socket->Device;
	boolean okSend = FALSE;
	int     idx;

	// Process telnet codes and filter them out of data stream.
	if (len > 1) {
		if ((len = sock_ProcessTelnet(keyBuffer, len)) == 0)
			return;
	}

	for (idx = 0; idx < len; idx++) {
		uchar ch = keyBuffer[idx];

		// Press ^\ twice to disconnect.
		if (ch == CTY_ESCAPE) {
			if (dte20->lastSeen == CTY_ESCAPE) {
				dte20_ctyEof(dte20->ctySocket, 0, 0);
				return;
			}
			dte20->lastSeen = ch;
			continue;
		}

		// Convert CR NL to CR line.
		if ((ch == 012) && (dte20->lastSeen == 015))
			continue;
		dte20->lastSeen = ch;

		okSend = TRUE;
		dte20->inBuffer[dte20->idxInQueue] = ch;
		if (++dte20->idxInQueue == 4096)
			dte20->idxInQueue = 0;
		if (dte20->idxInQueue == dte20->idxOutQueue) {
#ifdef DEBUG
			dbg_Printf("CTY: Error - Overrun!!\n");
#endif /* DEBUG */
			break;
		}
	}

	if (okSend && (dte20->Protocol == SECONDARY) &&
		 (p10_pRead(KL10_eptAddr + DTEMTI, 0) == 0)) {
		p10_pWrite(KL10_eptAddr + DTEF11, dte20->inBuffer[dte20->idxOutQueue], 0);
		p10_pWrite(KL10_eptAddr + DTEMTI, -1, 0);
		if (++dte20->idxOutQueue == 4096)
			dte20->idxOutQueue = 0;
	}
}

/*
void dte20_ctyCheckQueue(void)
{
	static int count = 0;
	int   pi;

	if (count++ == 1000) {
		count = 0;

		if ((idxOutQueue != idxInQueue) &&
		    (p10_pRead(KL10_eptAddr + DTEMTI, 0) == 0)) {
			p10_pWrite(KL10_eptAddr + DTEF11, inBuffer[idxOutQueue], 0);
			p10_pWrite(KL10_eptAddr + DTEMTI, -1, 0);
			if (++idxOutQueue == 4096)
				idxOutQueue = 0;
		}
	}
}
*/

void dte20_ctyOutput(DTE20_DEVICE *dte20, uchar ch)
{
	if (dte20->ctySocket) {
		ch &= 0177;
		sock_Print(dte20->ctySocket, &ch, 1);

		// Log a character into a log file.
		if (ch == '\n') {
			dte20->outBuffer[dte20->idxOutBuffer++] = ch;
			dte20->outBuffer[dte20->idxOutBuffer++] = '\0';
#ifdef DEBUG
			if (dbg_Check(DBG_CONSOLE))
				dbg_Printf("CTY: %s", dte20->outBuffer);
#endif /* DEBUG */
			if (emu_logFile >= 0)
				write(emu_logFile, dte20->outBuffer, dte20->idxOutBuffer);
			dte20->idxOutBuffer = 0;
		} else {
			if (ch == '\b' || ch == 127) {
				if (dte20->idxOutBuffer > 0)
					dte20->idxOutBuffer--;
			} else if (ch != '\r' && ch != '\0') {
				if (dte20->idxOutBuffer < 4095)
					dte20->outBuffer[dte20->idxOutBuffer++] = ch;
			}
		}
	}
}

// Ring Doorbell to PDP-10.
void dte20_Ring10(DTE20_DEVICE *dte20, uint16 srFlag)
{
	int36 iop = IRQ_VEC | (dte20->idChannel << IRQ_P_DEV) | DTEDII;

#ifdef DEBUG
//	if (dbg_Check(DBG_INTERRUPT))
		dbg_Printf("DTE(%d): Ring Doorbell to PDP-10\n", dte20->idDevice);
#endif /* DEBUG */

	dte20->srFlags |= srFlag;
	KL10pi_RequestIO(dte20->srFlags & DTE_PIA, iop);

//	dbg_SetMode(DBG_TRACE|DBG_DATA);
//	printf("[Trace/Data On]\n");
}

inline void dte20_comWrite(DTE20_DEVICE *dte20, int30 comIndex, int36 data)
{
	int30 comAddr, comLimit;

	if (comAddr = p10_pRead(KL10_eptAddr + DTEDRW, 0)) {
		comLimit = p10_pRead(KL10_eptAddr + DTEDPW, 0);
#ifdef DEBUG
//		if (dbg_Check(IOREGS)) {
			dbg_Printf("DTE(%d): DRW %o,,%o  DPW %o,,%o\n",
				dte20->idDevice, LH18(comAddr), RH18(comAddr),
				LH18(comLimit), RH18(comLimit));

			if (comIndex <= comLimit) {
				int30 Addr = comAddr + comIndex;
				dbg_Printf("DTE(%d): Index %o (%o,,%o) <= %o,,%o\n",
					dte20->idDevice, comIndex, LH18(Addr), RH18(Addr), 
					LH18(data), RH18(data));
			} else {
				dbg_Printf("DTE(%d): Out of range - No action.\n",
					dte20->idDevice);
			}
//		}
#endif /* DEBUG */
		if (comIndex <= comLimit) {
			p10_eWrite(comAddr + comIndex, data);
		}
	}
}

inline int36 dte20_comRead(DTE20_DEVICE *dte20, int30 comIndex)
{
	int30 comAddr, comLimit;
	int36 data = 0;

	if (comAddr = p10_pRead(KL10_eptAddr + DTEERW, 0)) {
		comLimit = p10_pRead(KL10_eptAddr + DTEEPW, 0);
		comIndex += 2;
		if (comIndex <= comLimit)
			data = p10_eRead(comAddr + comIndex);
#ifdef DEBUG
//		if (dbg_Check(IOREGS)) {
			dbg_Printf("DTE(%d): ERW %o,,%o  EPW %o,,%o\n",
				dte20->idDevice, LH18(comAddr), RH18(comAddr),
				LH18(comLimit), RH18(comLimit));

			if (comIndex <= comLimit) {
				int30 Addr = comAddr + comIndex;
				dbg_Printf("DTE(%d): Index %o (%o,,%o) => %o,,%o\n",
					dte20->idDevice, comIndex, LH18(Addr), RH18(Addr), 
					LH18(data), RH18(data));
			} else {
				dbg_Printf("DTE(%d): Out of range - No action.\n",
					dte20->idDevice);
			}
//		}
#endif /* DEBUG */
	}
	return data;
}

// Get Time/Date Command (Secondary Protocol)
void dte20_GetTOD(int32 dteCmd)
{
	uint30 eptAddr = KL10_eptAddr + (dteCmd >> 16);
	uint32 tod0, tod1, tod2;
	struct tm *tm;
	struct timezone tz;
	time_t now;

	// Get time information from native system.
	now = time(NULL);
	tm = localtime(&now);
	gettimeofday(NULL, &tz);

	// Packing time information into three TOD words.
	tod0 = TOD_VALID | tm->tm_year;
	tod1 = (tm->tm_mon << 24) | ((tm->tm_mday - 1) << 16) |
	  (tm->tm_wday << 8) | (tm->tm_isdst > 0 ? TOD_DST : 0);
	tod2 = ((((tm->tm_hour * 60) + tm->tm_min) * 60 + tm->tm_sec) >> 1) << 16;

	// Put three TOD words into EPT area.
	p10_pWrite(eptAddr,   tod0, 0);
	p10_pWrite(eptAddr+1, tod1, 0);
	p10_pWrite(eptAddr+2, tod2, 0);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		dbg_Printf("DTE: %o,,%06o <= TOD0 %06o,,%06o (%08X)\n",
			LH18(eptAddr), RH18(eptAddr), LH18(tod0), RH18(tod0), tod0);
		dbg_Printf("DTE: %o,,%06o <= TOD1 %06o,,%06o (%08X)\n",
			LH18(eptAddr+1), RH18(eptAddr+1), LH18(tod1), RH18(tod1), tod1);
		dbg_Printf("DTE: %o,,%06o <= TOD2 %06o,,%06o (%08X)\n",
			LH18(eptAddr+2), RH18(eptAddr+2), LH18(tod2), RH18(tod2), tod2);
	}
#endif /* DEBUG */
}

void dte20_EnterQueue(DTE20_DEVICE *dte20, uint16 fnc, uint16 dev, uint16 line)
{
	uint16 *t10msg = (uint16 *)&dte20->t10Data[0];

	*t10msg++ = 8;
	*t10msg++ = fnc;
	*t10msg++ = dev;
	*t10msg++ = line;

	dte20->t10Length = 8;
}

int dte20_fncInitialMessage(DTE20_DEVICE *dte20)
{
	// Tell 10 that CTY configuration information
	dte20_EnterQueue(dte20, FNC_2TI, 1, 0);
}

void dte20_GetMessage(DTE20_DEVICE *dte20)
{
	int36   ebp;
	int30   ebpAddr;
	int36   qSize;
	int36   msgData;
	uint16  bCount;
	int     pos, size, mask;
	int     idx;

	// If DTEEBP is zero, do nothing and return.
	if ((ebp = p10_pRead(KL10_eptAddr + DTEEBP, 0)) == 0) {
		dte20->t11Length = 0;
		return;
	}

	ebpAddr = RH18(ebp);
	pos = BP_GETPOS(ebp);
	size = BP_GETSIZE(ebp);

	// Get queue size in bytes
	qSize = dte20_comRead(dte20, nQSIZE);
	bCount = qSize;

#ifdef DEBUG
//	if (dbg_Check(DBG_IODATA)) {
		dbg_Printf("DTE(%d): DTEEBP (TO-11 Byte Pointer) => %06o,,%06o\n",
			dte20->idDevice, LH18(ebp), RH18(ebp));
		dbg_Printf("DTE(%d):   (Position: %d  Size: %d  Address: %o)\n",
			dte20->idDevice, pos, size, ebpAddr);
		dbg_Printf("DTE(%d): QSIZE => %06o,,%06o = %o (%d) bytes\n",
			dte20->idDevice, LH18(qSize), RH18(qSize), bCount, bCount);
//	}
#endif /* DEBUG */

	// Transfer TO11 Message
	dte20->t11Length = bCount;
	msgData = p10_pRead(ebpAddr, 0);
	mask = (1 << size) - 1;
	for (idx = 0; idx < bCount; idx++) {
		if ((pos -= size) < 0) {
			pos = 36 - size;
			msgData = p10_pRead(++ebpAddr, 0);
		}
		dte20->t11Data[idx] = (uint8)(msgData >> pos) & mask;
	}

	// Update TO-11 Byte Pointer
	ebp = BP_PUTPOS(pos) | BP_PUTSIZE(size) | ebpAddr;
	p10_pWrite(KL10_eptAddr + DTEEBP, ebp, 0);

#ifdef DEBUG
//	if (dbg_Check(DBG_IODATA)) {
	{
		char msgData[1024];
		char *msg = msgData;
		int  idx;

		for (idx = 0; idx < dte20->t11Length; idx++) {
			if (idx > 0)
				*msg++ = ',';
			msg = ToRadix(msg, dte20->t11Data[idx], 8, 0);
		}
		*msg = '\0';

		// Log the TO11 message dump
		dbg_Printf("DTE(%d): TO11 Message: %s\n", dte20->idDevice, msgData);
	}
#endif /* DEBUG */
}

void dte20_SendMessage(DTE20_DEVICE *dte20)
{
	int36   tbp;
	int30   tbpAddr;
	int36   qSize;
	int36   msgData;
	uint16  bCount;
	int     pos, size, mask;
	int     idx;

	// If DTEEBP is zero, do nothing and return.
	tbp = p10_pRead(KL10_eptAddr + DTETBP, 0);

	tbpAddr = RH18(tbp);
	pos = BP_GETPOS(tbp);
	size = BP_GETSIZE(tbp);

	// Get queue size in bytes
	qSize = dte20_comRead(dte20, nQSIZE);
	bCount = qSize;

#ifdef DEBUG
//	if (dbg_Check(DBG_IODATA)) {
		dbg_Printf("DTE(%d): DTETBP (TO-10 Byte Pointer) => %06o,,%06o\n",
			dte20->idDevice, LH18(tbp), RH18(tbp));
		dbg_Printf("DTE(%d):   (Position: %d  Size: %d  Address: %o)\n",
			dte20->idDevice, pos, size, tbpAddr);
		dbg_Printf("DTE(%d): QSIZE => %06o,,%06o = %o (%d) bytes\n",
			dte20->idDevice, LH18(qSize), RH18(qSize), bCount, bCount);
//	}
#endif /* DEBUG */

	// Transfer TO10 Message
	bCount = 010000 - (dte20->t10bCount & TO10BC);
	mask = (1 << size) - 1;
	msgData = 0;
	for (idx = 0; idx < bCount; idx++) {
		if ((pos -= size) < 0) {
			pos = 36 - size;
			p10_pWrite(tbpAddr++, msgData, 0);
			msgData = 0;
		}
		msgData |= (int36)(dte20->t10Data[idx] & mask) << pos;
	}
	p10_pWrite(tbpAddr, msgData, 0);

	// Update TO-10 Byte Pointer
	tbp = BP_PUTPOS(pos) | BP_PUTSIZE(size) | tbpAddr;
	p10_pWrite(KL10_eptAddr + DTETBP, tbp, 0);

#ifdef DEBUG
//	if (dbg_Check(DBG_IODATA)) {
	{
		char msgData[1024];
		char *msg = msgData;
		int  idx;

		for (idx = 0; idx < dte20->t10Length; idx++) {
			if (idx > 0)
				*msg++ = ',';
			msg = ToRadix(msg, dte20->t10Data[idx], 8, 0);
		}
		*msg = '\0';

		// Log the TO11 message dump
		dbg_Printf("DTE(%d): TO10 Message: %s\n", dte20->idDevice, msgData);
	}
#endif /* DEBUG */

	dte20->t10Length = 0;
}

void dte20_ProcessMessage(DTE20_DEVICE *dte20)
{
	int fnc = dte20->t11Data[3];
	int ring = 0;
	
#ifdef DEBUG
//	if (dbg_Check(DBG_IODATA))
		dbg_Printf("DTE(%d): Function: %s\n",
			dte20->idDevice, fnc < 040 ? fncNames[fnc] : "Illegal Function");
#endif /* DEBUG */

	if ((fnc < 040) && fncExecute[fnc])
		fncExecute[fnc](dte20);

	if (dte20->t10Length)
		ts10_SetTimer(dte20->Timer);

	// Tell 10 that it is done.
	dte20_Ring10(dte20, DTE_TO11DN);
}

// Callback from timer routine
void dte20_Process(DTE20_DEVICE *dte20)
{
	int36 stWord;

	if (dte20->t10Length) {
		stWord = dte20_comRead(dte20, nSTATUS);
		stWord = (stWord & ~ST_TO11IC) | ((stWord + 1) & ST_M_IC);
		dte20_comWrite(dte20, nQSIZE, dte20->t10Length);
		dte20_comWrite(dte20, nSTATUS, stWord);

		// Now tell 10 that TO10Q message is here.
		dte20_Ring10(dte20, DTE_TO10DB);
	}
}

// Primary Protocol Process
boolean dte20_ppcProcess(DTE20_DEVICE *dte20)
{
	int36  stWord;
	uint16 to10ic, to11ic;	

	// Check examine valid bit first or go to secondary protocol
	if (((stWord = dte20_comRead(dte20, nSTATUS)) == 0) &&
	    ((stWord & ST_VALID) == 0)) {
#ifdef DEBUG
//		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("DTE(%d): Not valid examine - Enter secondary protocol\n",
				dte20->idDevice);
#endif /* DEBUG */
		dte20->Protocol = SECONDARY;
		return SECONDARY;
	}

	// When STATUS is valid, clear TO-11 doorbell.
	dte20->srFlags &= ~DTE_TO11DB;

	// Check and Update Interrupt Count (STATUS)
	to10ic = (uint16)(stWord >> ST_P_TO10IC) & ST_M_IC;
	to11ic = (uint16)stWord & ST_M_IC;
	if (to10ic == to11ic)
		return PRIMARY;
	stWord = (stWord & ~ST_TO10IC) |
		(int36)(((to10ic + 1) & ST_M_IC) << ST_P_TO10IC);
	dte20_comWrite(dte20, nSTATUS, stWord);

	// Process TO11 Message.
	dte20_GetMessage(dte20);
	dte20_ProcessMessage(dte20);

	return PRIMARY;
}

// Enter primary protocol and
// initialize communication region area
void dte20_InitPrimary(DTE20_DEVICE *dte20)
{
#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS)) {
		dbg_Printf("DTE(%d): Enter Primary Protocol\n", dte20->idDevice);
		dbg_Printf("DTE(%d): Initializing Communication Region Area\n",
			dte20->idDevice);
//	}
#endif /* DEBUG */
	dte20->Protocol = PRIMARY;

	dte20_ppcProcess(dte20);
}

// Secondary Protocol Process
void dte20_spcProcess(DTE20_DEVICE *dte20)
{
	int32 dteCmd = p10_pRead(KL10_eptAddr + DTECMD, 0);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("DTE: DTECMD => %06llo,,%06llo\n",
			LHSR(dteCmd), RH(dteCmd));
#endif /* DEBUG */

	switch ((dteCmd >> CMD_P_CODE) & CMD_M_CODE) {
		case CMD_MTO:
			// Monitor Output Command
			dte20_ctyOutput(dte20, dteCmd & MTO_CHAR);
			p10_pWrite(KL10_eptAddr + DTEMTD, dte20->mtdFlags | MTD_ODN, 0);
			if (dte20->srFlags & DTE_PIOENB)
				dte20_Ring10(dte20, DTE_TO10DB);
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS)) {
				uint8 ch = dteCmd & 0277;
				dbg_Printf("DTE: Output Character = %c (%02X)\n",
					((ch >= ' ' && ch < 127) ? ch : '.'), ch);
			}
#endif /* DEBUG */
			break;

		case CMD_ESP:
			dte20->Protocol = SECONDARY;
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("DTE(%d): Enter Secondary Protocol\n",
					dte20->idDevice);
#endif /* DEBUG */
			break;

		case CMD_LSP:
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("DTE(%d): Leave Secondary Protocol\n",
					dte20->idDevice);
#endif /* DEBUG */
			dte20->Protocol = PRIMARY;
			if (dteCmd & LSP_RCOM)
				dte20_InitPrimary(dte20);
			break;

		case CMD_GDT:
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("DTE(%d): Get Date/Time.\n",
					dte20->idDevice);
#endif /* DEBUG */
			dte20_GetTOD(dteCmd);
			break;

		default:
#ifdef DEBUG
			if (dbg_Check(DBG_IOREGS))
				dbg_Printf("DTE(%d): Unknown command code %02o\n", 
					dte20->idDevice, (dteCmd >> 8) & 017);
#endif /* DEBUG */
	}

	// Tell KL10 that DTE command was done.
	p10_pWrite(KL10_eptAddr + DTEFLG, -1, 0);

	// Clear TO-11 doorbell.
	dte20->srFlags &= ~DTE_TO11DB;
}

static int (*fncExecute[])(DTE20_DEVICE *) =
{
	NULL,                       // 000 *Unknown Function
	dte20_fncInitialMessage,    // 001 TO11 Initial Message
	NULL,                       // 002 TO10 Initial Message
	NULL,                       // 003 String Data
	NULL,                       // 004 Line-Char, Line-Char
	NULL,                       // 005 Request Device Status
	NULL,                       // 006 *Unknown Function
	NULL,                       // 007 Here is Device Status
	NULL,                       // 010 *Unknown Function
	NULL,                       // 011 Request Date/Time
	NULL,                       // 012 Here is Date/Time
	NULL,                       // 013 Flush Output (^O)
	NULL,                       // 014 Send All
	NULL,                       // 015 Dataset Connected
	NULL,                       // 016 Hang Up Dataset
	NULL,                       // 017 Acknowledge (Device Line Alloc Free)
	NULL,                       // 020 X-OFF TTY Line
	NULL,                       // 021 X-ON TTY Line
	NULL,                       // 022 Here are Speed Lines for TTYs
	NULL,                       // 023 Here are Allocations for TTYs
	NULL,                       // 024 TO10 Reboot Information
	NULL,                       // 025 Acknowledge All
	NULL,                       // 026 Turn Device (On/Off)
	NULL,                       // 027 Enable/Disable Remotes
	NULL,                       // 030 Load Line Printer RAM
	NULL,                       // 031 Load Line Printer VFU
	NULL,                       // 032 DAS60 Data
	NULL,                       // 033 KLINIK Parameter Status
	NULL,                       // 034 Enable/Disable Auto-XOFF
	NULL,                       // 035 TO11 Break Through Char Data
	NULL,                       // 036 TO11 Turn On Debug Mode
	NULL,                       // 037 TO11 Turn Off Debug Mode
};

//****************************************************************

void dte20_ioOpcode_DODTE(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	dte20->t10bCount = p10_vRead(eAddr, 0);

#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS)) {
	{
		int32 bCount = dte20->t10bCount & TO10BC;
		dbg_Printf("DTE(%d): TO10 Byte Count <= %o (%d)  Interrupt: %s\n",
			dte20->idDevice, bCount, bCount,
			(dte20->t10bCount & TO10IB ? "Yes" : "No"));
	}
#endif /* DEBUG */

	// Transfer TO10 message to 10's memory.
	dte20_SendMessage(dte20);
	if (dte20->t10bCount & TO10IB)
		dte20_Ring10(dte20, DTE_TO10DN);
}

void dte20_ioOpcode_WRDTE(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	// Clear done/doorbell/error flags
	dte20->srFlags &= ~(eAddr & DTE_CLR);
	if (eAddr & DTE_CLTO11)
		dte20->srFlags &= ~DTE_TO11ER;

	// Load PI assignment.
	if (eAddr & DTE_PILDEN) {
		dte20->srFlags = (dte20->srFlags & ~(DTE_PIOENB|DTE_PIA)) |
			(eAddr & (DTE_PIOENB|DTE_PIA));
	}

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("DTE: (WRDTE) SR %06o <= %06o\n",
			dte20->srFlags, (uint16)eAddr);
#endif /* DEBUG */

	// Ring doorbell to PDP11.
	if (eAddr & DTE_TO11DB) {
#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("DTE(0): Rings Doorbell at %o,,%o\n",
				LH18(pager_PC), RH18(pager_PC));
#endif /* DEBUG */

		// Set TO-11 Doorbell
		dte20->srFlags |= DTE_TO11DB;

		// Process Communication
		if (dte20->Protocol == PRIMARY) {
			if (dte20_ppcProcess(dte20))
				dte20_spcProcess(dte20);
		} else
			dte20_spcProcess(dte20);
	}
}

void dte20_ioOpcode_RDDTE(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	p10_vWrite(eAddr, dte20->srFlags, 0);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("DTE: (RDDTE) SR => %06o\n", dte20->srFlags);
#endif /* DEBUG */
}

void dte20_ioOpcode_CZDTE(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	if ((dte20->srFlags & eAddr) == 0)
		DO_SKIP;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		int32 result = dte20->srFlags & eAddr;
		dbg_Printf("DTE: (CZDTE) SR => %06o & %06o = %06o : %s\n",
			dte20->srFlags, (int18)eAddr, result,
			((result == 0) ? "Skip" : "Continue"));
	}
#endif /* DEBUG */
}

void dte20_ioOpcode_CODTE(void *dptr)
{
	DTE20_DEVICE *dte20 = (DTE20_DEVICE *)dptr;

	if (dte20->srFlags & eAddr)
		DO_SKIP;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		int32 result = dte20->srFlags & eAddr;
		dbg_Printf("DTE: (CODTE) SR => %06o & %06o = %06o : %s\n",
			dte20->srFlags, (int18)eAddr, result,
			(result ? "Skip" : "Continue"));
	}
#endif /* DEBUG */
}

P10DEVICE KX10_Device_DTE20 = {
	"DTE20",
	dte20_Create,
	dte20_Cleanup,
	dte20_Initialize,
	dte20_Reset,
	{
		NULL, // (00) Block In
		NULL, // (04) Data In
		NULL, // (10) Block Out
		dte20_ioOpcode_DODTE, // (14) Data Out
		dte20_ioOpcode_WRDTE, // (20) Conditions Out
		dte20_ioOpcode_RDDTE, // (24) Conditions In
		dte20_ioOpcode_CZDTE, // (30) Conditions In and Skip if zero
		dte20_ioOpcode_CODTE, // (34) Conditions In and Skip if one
	}
};
