// system.c - VAX System Configurations
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS10 Emulator.
// See ReadMe for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// Device Name: RH20 (8 Channels)
// Device Code: 540,544,550,554,560,564,570,574

#include "pdp10/defs.h"
#include "pdp10/kl10.h"
#include "pdp10/proto.h"
#include "dev/mba/mba.h"

// Channel Logout Area in Executive Process Table.
extern int30 KL10_eptAddr;

MBA_CALLBACK rh20_Callback;

// DATAI/DATAO Register definitions

// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// | Register  |L|               |Drive|                                   |
// +-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-^-+-+-+
//  0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3
//  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5

// Common Data for All Registers (left halfword)
#define D_REG   0770000 // (R/W) Register Select Field
#define D_LDR   0004000 // (R/W) Load Register
#define D_DRE   0000400 // (W)   Disable Register Access Error
#define D_DRV   0000007 // (R/W) Drive Select Field

#define D_M_REG  077
#define D_P_REG  30
#define D_M_DRV  07
#define D_P_DRV  18

// External Register Data
#define ER_CPE  0001000000000 // (R)   Control Bus Parity Error
#define ER_TRA  0000200000000 // (R)   Transfer Received
#define ER_CEP  0000000400000 // (W)   Control Bus Even Parity
#define ER_CPA  0000000200000 // (R)   Control Bus Parity Bit
#define ER_DATA 0000000177777 // (R/W) External Register Data

// Primary/Secondary Block Address Register
#define BA_ADR  0000000177777 // (R/W) Block Address Register

// Primary/Secondary Transfer Control Register
#define TC_RCP  0002000000000 // (R/W) Reset Command List Pointer
#define TC_SCS  0000200000000 // (R/W) Store Channel Status
#define TC_DTE  0000000200000 // (R/W) Disable Transfer Error Stop
#define TC_CNT  0000000177700 // (R/W) Negative Block Count
#define TC_FNC  0000000000077 // (R/W) Massbus Function Code

#define TC_MAXCNT 02000  // Maximum Block Count
#define TC_M_CNT  01777  // Mask of Block Count
#define TC_P_CNT  6      // Position of Block Count
#define TC_M_FNC  077    // Mask of Massbus Function

// Interrupt Vector Index Data Register
#define IVI_ADR 0000000000777 // (R/W) Interrupt Vector Address

// Specific Register definitions
#define RH20_XREG 040 // External Data Register (00 - 37)
#define RH20_SBA  070 // Secondary Block Address Register
#define RH20_STC  071 // Secondary Transfer Control Register
#define RH20_PBA  072 // Primary Block Address Register
#define RH20_PTC  073 // Primary Transfer Control Register
#define RH20_IVI  074 // Interrupt Vector Index Register

// CONI/CONO/CONSZ/CONSO Register definitions

#define CI_DBP  0400000 // Data Bus Parity Error
#define CI_EXC  0200000 // Exception
#define CI_LWC  0100000 // Long Word Count
#define CI_SWC  0040000 // Short Word Count
#define CI_CER  0020000 // Channel Error
#define CI_DRE  0010000 // Drive Response Error
#define CI_RAE  0004000 // Register Access Error
#define CI_RDY  0002000 // Channel Ready
#define CI_OVR  0001000 // Data Overun
#define CI_MBE  0000400 // Massbus Enable
#define CI_ATN  0000200 // Attention
#define CI_SCR  0000100 // SCR Full
#define CI_AIE  0000040 // Attention Interrupt Enable
#define CI_BSY  0000020 // PCR Full (Busy)
#define CI_DON  0000010 // Command Done
#define CI_PIA  0000007 // Priority Interrupt Assignment

#define CO_RAE  0004000 // Clear Register Access Error
#define CO_MBI  0002000 // Massbus Initialization
#define CO_TEC  0001000 // Transfer Error Clear
#define CO_MBE  0000400 // Massbus Enable
#define CO_RCP  0000200 // Reset Command List Pointer
#define CO_SCR  0000100 // Delete SCR
#define CO_AIE  0000040 // Attention Interrupt Enable
#define CO_STP  0000020 // Stop Transfer
#define CO_CCD  0000010 // Clear Command Done
#define CO_PIA  0000007 // Priority Interrupt Assignment

#define CO_CMASK (CO_RAE|CO_TEC|CO_CCD)  // Clear bits mask
#define CO_WMASK (CO_MBE|CO_AIE|CO_PIA)  // Write bits mask

// Channel Logout Area - Defintions

#define CSICW  0   // Initial Control Word
#define CSCLP  1   // Status/Command List Pointer
#define CSDBA  2   // Current CCW
#define CSIVI  3   // Interrupt Vector Instruction

// CSICW - Initial Control Word
#define ICW_OPC  0700000000000 // Opcode
#define ICW_XFR  0400000000000 // Opcode - Data Transfer if Set
#define ICW_HLT  0200000000000 // Opcode - Halt Bit for Data Transfers
#define ICW_REV  0100000000000 // Opcode - Reverse Bit for Data Transfers
#define ICW_WDC  0077760000000 // Word Count
#define ICW_ADR  0000017777777 // Address

#define CWM_OPC 07        // Opcode Mask
#define CWP_OPC 33        // Position of Opcode Field
#define CWB_XFR 04        // Data Tranfer if Set
#define CWB_HLT 02        // Halt Bit for Data Transfers
#define CWB_REV 01        // Reverse Bit for Data Transfers
#define CWM_WDC 03777     // Word Count Mask
#define CWP_WDC 22        // Position of Word Count Field
#define CWM_ADR 017777777 // Address Mask

#define CWOP_HALT 00      // HALT Opcode
#define CWOP_JUMP 02      // JUMP Opcode
#define CWOP_RDF  04      // Read Data Forward (Do not halt)
#define CWOP_RDR  05      // Read Data Reverse (Do not halt)
#define CWOP_HRDF 06      // Read Data Forward (Halt)
#define CWOP_HRDR 07      // Read Data Reverse (Halt)

// CSCLP - Status/Command List Pointer
#define CLP_MPE  0200000000000LL // Memory Parity Error
#define CLP_NAE  0100000000000LL // Not Address Parity Error
#define CLP_NW0  0040000000000LL // Not Zero Word Count
#define CLP_NXM  0020000000000LL // Non-existant Memory
#define CLP_LTE  0000400000000LL // Late Transfer Error
#define CLP_RHE  0000200000000LL // RH20 Error
#define CLP_LWC  0000100000000LL // Long Word Count
#define CLP_SWC  0000040000000LL // Short Word Count
#define CLP_OVR  0000020000000LL // Overrun
#define CLP_ADR  0000017777777LL // Command List Pointer

// CSDBA - Current Command Word
#define DBA_OPC  0700000000000 // Opcode
#define DBA_CWC  0077760000000 // Word Count
#define DBA_ADR  0000017777777 // Data Buffer Address

// ***************************************************************

typedef struct {
	int    idChannel;   // Channel Number Identification
	MBA_DEVICE *Device; // Massbus Structure Pointer
	int18  srFlags;     // Status Register
	int18  erFlags;     // External Register
	int32  piLevel;     // Priority Interrupt Level
	int30  secCmdAddr;  // Secondary: Command List Pointer
	int30  secCount;    // Secondary: Block Count
	int30  blkAddr;     // Block Address Register
	int30  iviAddr;     // Interrupt Vector Index
	uint16 Register;    // Current Register Number
	uint16 Drive;       // Current Drive Number
	uint16 Data;        // Current Data
} RH20_DEVICE;

void *rh20_Create(void *dptr, char *devName, int argc, char **argv)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	if (rh20 == NULL) {
		int nChannel = argc;

		// Create a individual device for each channel.
		rh20 = (RH20_DEVICE *)calloc(1, sizeof(RH20_DEVICE));
		rh20->idChannel = nChannel;
		rh20->Device = mba_Create(rh20, &rh20_Callback, nChannel, argc, argv);
	} else {
		return mba_CreateDrive(rh20->Device, devName, argc, argv);
	}
	
	return (void *)rh20;
}

void rh20_Initialize(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	rh20->srFlags = CI_RDY;
}

void rh20_Reset(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	rh20->srFlags = CI_RDY;
}

void rh20_Cleanup(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;
}

void rh20_DoInterrupt(RH20_DEVICE *rh20)
{
	int36 iopWord = IRQ_VEC | (rh20->idChannel << IRQ_P_DEV) |
		rh20->iviAddr;

#ifdef DEBUG
	if (dbg_Check(DBG_INTERRUPT))
		dbg_Printf("RH20(%d): Interrupt Word: %06o,,%06o at PI level %d\n",
			rh20->idChannel, (int18)SR(iopWord), (int18)RH(iopWord),
			rh20->piLevel);
#endif /* DEBUG */

	// Tell KL10 Processor to process it.
	KL10pi_RequestIO(rh20->piLevel, iopWord);
}

void rh20_SetAttention(RH20_DEVICE *rh20)
{
	rh20->srFlags |= CI_ATN;
	if (rh20->srFlags & CI_AIE) {
#ifdef DEBUG
		if (dbg_Check(DBG_INTERRUPT))
			dbg_Printf("RH20(%d): Attention Interrupt\n", rh20->idChannel);
#endif /* DEBUG */
		rh20_DoInterrupt(rh20);
	}
}

void rh20_SetReady(RH20_DEVICE *rh20)
{
	rh20->srFlags |= CI_DON;
#ifdef DEBUG
	if (dbg_Check(DBG_INTERRUPT))
		dbg_Printf("RH20(%d): Command Done Interrupt\n", rh20->idChannel);
#endif /* DEBUG */
	rh20_DoInterrupt(rh20);
}

int36 rh20_GetNextCCW(RH20_DEVICE *rh20,
	int32 *ccwOpcode, int32 *ccwCount, int30 *ccwAddr)
{
	int36   ccw;       // Command Control Word
	boolean nxtFlag;   // Get Next CCW for Jump, etc.

	do {
		ccw = p10_pRead(rh20->secCmdAddr, 0);

		*ccwOpcode = (ccw >> CWP_OPC) & CWM_OPC;
		*ccwCount  = (ccw >> CWP_WDC) & CWM_WDC;
		*ccwAddr   = ccw & CWM_ADR;

#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("RH20(%d): %08o - Opcode: %o  Word Count: %04o  Address: %08o\n",
				rh20->idChannel, rh20->secCmdAddr, *ccwOpcode, *ccwCount, *ccwAddr);
#endif /* DEBUG */

		nxtFlag = FALSE;
		if (*ccwOpcode == CWOP_JUMP) {
			rh20->secCmdAddr = *ccwAddr;
			nxtFlag = TRUE;
		} else
			rh20->secCmdAddr++;

	} while (nxtFlag);

	// Save for block address register.
	rh20->blkAddr = *ccwAddr;

	return ccw;
}

int rh20_ReadBlock(RH20_DEVICE *rh20, int32 *blkData, int blkSize)
{
	int36   data36;        // Data buffer
	int36   ccw;           // Command Control Word
	int32   ccwOpcode = 0; // Opcode Field
	int32   ccwCount = 0;  // Word Count Field
	int30   ccwAddr;       // Data Buffer Address
	int     idx;

	for (idx = 0; idx < blkSize; idx += 2) {
		if (ccwCount == 0) {
			// If opcode contains HALT bit, stop transfer.
			if (ccwOpcode & CWB_HLT)
				break;

			// Get next control word to store data on desired location.
			ccw = rh20_GetNextCCW(rh20, &ccwOpcode, &ccwCount, &ccwAddr);

			// if opcode is HALT, stop transfer.
			if (ccwOpcode == 0)
				break;

			// Throw-away CCW
			if (ccwAddr == 0) {
				idx += (ccwCount << 1);
#ifdef DEBUG
				if (dbg_Check(DBG_IODATA)) {
					dbg_Printf("RH20(%d): Throw-Away CCW - %04o (%d) words skipped.\n",
						rh20->idChannel, ccwCount, ccwCount);
				}
#endif /* DEBUG */
				ccwCount = 0;
				continue;
			}
		}

		data36 = p10_pRead(ccwAddr, 0);

		if (ccwOpcode & CWB_REV) {
			// Reverse Data
			blkData[idx+1] = LHSR(data36);
			blkData[idx] = RH(data36);
		} else {
			// Forward Data
			blkData[idx] = LHSR(data36);
			blkData[idx+1] = RH(data36);
		}

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			if (ccwOpcode & CWB_REV) {
				// Reverse Data
				dbg_Printf("RH20(%d): (R) %08o (%04o) => %06o,,%06o\n",
					rh20->idChannel, ccwAddr, idx,
					blkData[blkSize - (idx + 1)], blkData[blkSize + idx]);
			} else {
				// Forward Data
				dbg_Printf("RH20(%d): (R) %08o (%04o) => %06o,,%06o\n",
					rh20->idChannel, ccwAddr, idx, blkData[idx], blkData[idx+1]);
			}
		}
#endif /* DEBUG */

		ccwAddr++;
		ccwCount--;
	}

	// When HALT is reached, store status channel...
	if ((ccwOpcode == 0) || (ccwOpcode & CWB_HLT)) {
		int30 clpAddr; // Command List Pointer Address
		int30 dbaAddr; // Data Buffer Address
		int36 clp;     // Status/CLP Word

		// Update status/command list pointer
		clp = rh20->secCmdAddr | CLP_NAE;
		if (ccwCount)
			clp |= (CLP_NW0|CLP_LWC);
		if (idx < blkSize)
			clp |= CLP_SWC;

		// Update current control word
		ccw = ((int36)(ccwCount & CWM_WDC) << CWP_WDC) |
		      (int36)(ccwAddr & CWM_ADR) | (ccw & ICW_OPC);

		clpAddr = KL10_eptAddr + CSCLP + (rh20->idChannel << 4);
		dbaAddr = KL10_eptAddr + CSDBA + (rh20->idChannel << 4);
		p10_pWrite(clpAddr, clp, 0);
		p10_pWrite(dbaAddr, ccw, 0);

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			dbg_Printf("RH20(%d): CSCLP (%08o) <= %06o,,%06o\n",
				rh20->idChannel, clpAddr, (int18)SR(clp), (int18)RH(clp));
			dbg_Printf("RH20(%d): CSDBA (%08o) <= %06o,,%06o\n",
				rh20->idChannel, dbaAddr, (int18)SR(ccw), (int18)RH(ccw));
		}
#endif /* DEBUG */
	}

	return EMU_OK;
}

// Transfer a 18-bit block to the PDP-10 main memory directly
int rh20_WriteBlock(RH20_DEVICE *rh20, int32 *blkData, int blkSize)
{
	int36   data36;        // Data buffer
	int36   ccw;           // Command Control Word
	int32   ccwOpcode = 0; // Opcode Field
	int32   ccwCount = 0;  // Word Count Field
	int30   ccwAddr;        // Data Buffer Address
	int     idx;

	for (idx = 0; idx < blkSize; idx += 2) {
		if (ccwCount == 0) {
			// If opcode contains HALT bit, stop transfer.
			if (ccwOpcode & CWB_HLT)
				break;

			// Get next control word to store data on desired location.
			ccw = rh20_GetNextCCW(rh20, &ccwOpcode, &ccwCount, &ccwAddr);

			// if opcode is HALT, stop transfer.
			if (ccwOpcode == 0)
				break;

			// Throw-away CCW
			if (ccwAddr == 0) {
				idx += (ccwCount << 1);
#ifdef DEBUG
				if (dbg_Check(DBG_IODATA)) {
					dbg_Printf("RH20(%d): Throw-Away CCW - %04o (%d) words skipped.\n",
						rh20->idChannel, ccwCount, ccwCount);
				}
#endif /* DEBUG */
				ccwCount = 0;
				continue;
			}
		}

		if (ccwOpcode & CWB_REV) {
			// Reverse Data
			data36 = blkData[blkSize - idx] |
				SL((int36)blkData[blkSize - (idx + 1)]);
		} else {
			// Forward Data
			data36 = SL((int36)blkData[idx]) | blkData[idx+1];
		}

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			if (ccwOpcode & CWB_REV) {
				// Reverse Data
				dbg_Printf("RH20(%d): (W) %08o (%04o) <= %06o,,%06o\n",
					rh20->idChannel, ccwAddr, idx,
					blkData[blkSize - (idx + 1)], blkData[blkSize + idx]);
			} else {
				// Forward Data
				dbg_Printf("RH20(%d): (W) %08o (%04o) <= %06o,,%06o\n",
					rh20->idChannel, ccwAddr, idx, blkData[idx], blkData[idx+1]);
			}
		}
#endif /* DEBUG */

		p10_pWrite(ccwAddr++, SXT36(data36), 0);
		ccwCount--;
	}

	// When HALT is reached, store status channel...
	if ((ccwOpcode == 0) || (ccwOpcode & CWB_HLT)) {
		int30 clpAddr; // Command List Pointer Address
		int30 dbaAddr; // Data Buffer Address
		int36 clp;     // Status/CLP Word

		// Update status/command list pointer
		clp = rh20->secCmdAddr | CLP_NAE;
		if (ccwCount)
			clp |= (CLP_NW0|CLP_LWC);
		if (idx < blkSize)
			clp |= CLP_SWC;

		// Update current control word
		ccw = ((int36)(ccwCount & CWM_WDC) << CWP_WDC) |
		      (int36)(ccwAddr & CWM_ADR) | (ccw & ICW_OPC);

		clpAddr = KL10_eptAddr + CSCLP + (rh20->idChannel << 4);
		dbaAddr = KL10_eptAddr + CSDBA + (rh20->idChannel << 4);
		p10_pWrite(clpAddr, clp, 0);
		p10_pWrite(dbaAddr, ccw, 0);

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			dbg_Printf("RH20(%d): CSCLP (%08o) <= %06o,,%06o\n",
				rh20->idChannel, clpAddr, (int18)SR(clp), (int18)RH(clp));
			dbg_Printf("RH20(%d): CSDBA (%08o) <= %06o,,%06o\n",
				rh20->idChannel, dbaAddr, (int18)SR(ccw), (int18)RH(ccw));
		}
#endif /* DEBUG */
	}

	return EMU_OK;
}

// Secondary Transfer Control
void rh20_secTransfer(RH20_DEVICE *rh20, int36 tc)
{
	int   blkCount = TC_MAXCNT - (tc >> TC_P_CNT) & TC_M_CNT;
	int   fncDrive = tc & TC_M_FNC;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		dbg_Printf("RH20(%d): Secondary Transfer Control\n", rh20->idChannel);
		dbg_Printf("RH20(%d): Block Count = %04o (%d) Massbus Function = %02o\n",
			rh20->idChannel, blkCount, blkCount, fncDrive);
	}
#endif /* DEBUG */

	// Reset Command List Pointer
	// Get Initial/Current Control Word
	rh20->secCmdAddr = KL10_eptAddr + (rh20->idChannel << 2)
		+ ((tc & TC_RCP) ? CSICW : CSDBA);

	// Now execute Massbus function...
//	mba_SetBlockCount(rh20->Device, rh20->Drive, blkCount);
	mba_WriteIO(rh20->Device, rh20->Drive, 0, fncDrive);
}

// ***************************************************************

void rh20_ioOpcode_DATAI(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;
	int36 data;
	int rc;

	if (rh20->Register < RH20_XREG) {
		mba_ReadIO(rh20->Device, rh20->Drive, rh20->Register, &rh20->Data);
	} else {
		switch(rh20->Register) {
			case RH20_PTC:
			case RH20_STC:
				rh20->Data = 0;
				break;

			case RH20_PBA:
			case RH20_SBA:
				rh20->Data = rh20->blkAddr;
				break;

			case RH20_IVI:
				// Load Interrupt Vector Index Register
				rh20->Data = rh20->iviAddr;
#ifdef DEBUG
				if (dbg_Check(DBG_IOREGS))
					dbg_Printf("RH20(%d): Interrupt Vector Address %03o\n",
						rh20->idChannel, rh20->iviAddr);
#endif /* DEBUG */
				break;

			default:
				// Unknown or Illegal Register
				rh20->srFlags |= CI_RAE;
#ifdef DEBUG
				if (dbg_Check(DBG_IOREGS))
					dbg_Printf("RH20(%d): Unknown Register %02o\n",
						rh20->idChannel, rh20->Register);
#endif /* DEBUG */
		}
	}
	data = SL((int36)(rh20->erFlags & ~D_DRE)) | rh20->Data;

	p10_vWrite(eAddr, data, 0);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("RH20(%d): (DI) ER => %06o,,%06o\n",
			rh20->idChannel, (int18)LHSR(data), (int18)RH(data));
#endif /* DEBUG */
}

void rh20_ioOpcode_DATAO(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;
	int36 data = p10_vRead(eAddr, 0);
	int rc;

	rh20->erFlags  = (int18)LHSR(data);
	rh20->Register = (data >> D_P_REG) & D_M_REG;
	rh20->Drive    = (data >> D_P_DRV) & D_M_DRV;
	rh20->Data     = data & ER_DATA;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("RH20(%d): (DO) ER <= %06o,,%06o\n",
			rh20->idChannel, (int18)LHSR(data), (int18)RH(data));
#endif /* DEBUG */

	// Load Register if LDR bit set
	if (rh20->erFlags & D_LDR) {
		if (rh20->Register < RH20_XREG) {
			mba_WriteIO(rh20->Device, rh20->Drive, rh20->Register, rh20->Data);
		} else {
			switch(rh20->Register) {
				case RH20_PTC:
				case RH20_STC:
					rh20_secTransfer(rh20, data);
					break;

				case RH20_IVI:
					// Load Interrupt Vector Index Register
					rh20->iviAddr = data & IVI_ADR;
#ifdef DEBUG
					if (dbg_Check(DBG_IOREGS))
						dbg_Printf("RH20(%d): Interrupt Vector Address %03o\n",
							rh20->idChannel, rh20->iviAddr);
#endif /* DEBUG */
					break;

				default:
					// Unknown or Illegal Register
					rh20->srFlags |= CI_RAE;
#ifdef DEBUG
					if (dbg_Check(DBG_IOREGS))
						dbg_Printf("RH20(%d): Unknown Register %02o\n",
							rh20->idChannel, rh20->Register);
#endif /* DEBUG */
			}
		}
	}
}

void rh20_ioOpcode_CONO(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	if (eAddr & CO_MBI) {
		// Massbus Initialization
		rh20_Reset(dptr);
#ifdef DEBUG
		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("RH20(%d): (CO) Massbus Initialization\n",
				rh20->idChannel);
#endif /* DEBUG */
	}
 
	// Write bits to clear bits
	rh20->srFlags &= ~(eAddr & CO_CMASK);

	// Write bits to enable/clear bits
	// Also, Change Priority Interrupt Assignment
	rh20->srFlags = (rh20->srFlags & ~CO_WMASK) | (eAddr & CO_WMASK);
	rh20->piLevel = (eAddr & CO_PIA);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("RH20(%d): (CO) SR %06o <= %06o\n",
			rh20->idChannel, rh20->srFlags, (int18)RH(eAddr));
#endif /* DEBUG */
}

void rh20_ioOpcode_CONI(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	p10_vWrite(eAddr, rh20->srFlags, 0);

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("RH20(%d): (CI) SR => %06o\n",
			rh20->idChannel, rh20->srFlags);
#endif /* DEBUG */
}

void rh20_ioOpcode_CONSZ(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	if ((rh20->srFlags & RH(eAddr)) == 0)
		DO_SKIP;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		int18 result = rh20->srFlags & RH(eAddr);
		dbg_Printf("RH20(%d): (SZ) SR => %06o & %06o = %06o : %s\n",
			rh20->idChannel, rh20->srFlags, (int18)RH(eAddr), result,
			(result == 0 ? "Skip" : "Continue"));
	}
#endif /* DEBUG */
}

void rh20_ioOpcode_CONSO(void *dptr)
{
	RH20_DEVICE *rh20 = (RH20_DEVICE *)dptr;

	if (rh20->srFlags & RH(eAddr))
		DO_SKIP;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS)) {
		int18 result = rh20->srFlags & RH(eAddr);
		dbg_Printf("RH20(%d): (SO) SR => %06o & %06o = %06o : %s\n",
			rh20->idChannel, rh20->srFlags, (int18)RH(eAddr), result,
			(result ? "Skip" : "Continue"));
	}
#endif /* DEBUG */
}

P10DEVICE KX10_Device_RH20 = {
	"RH20",
	rh20_Create,
	rh20_Cleanup,
	rh20_Initialize,
	rh20_Reset,
	{
		NULL,                // (00) Block In
		rh20_ioOpcode_DATAI, // (04) Data In
		NULL,                // (10) Block Out
		rh20_ioOpcode_DATAO, // (14) Data Out
		rh20_ioOpcode_CONO,  // (20) Conditions Out
		rh20_ioOpcode_CONI,  // (24) Conditions In
		rh20_ioOpcode_CONSZ, // (30) Conditions In and Skip if zero
		rh20_ioOpcode_CONSO, // (34) Conditions In and Skip if one
	}
};

MBA_CALLBACK rh20_Callback = {
	rh20_SetAttention,  // Attention Signal
	rh20_SetReady,      // Ready/Done Signal
	rh20_ReadBlock,     // Read Block
	rh20_WriteBlock     // Write Block
};
