// instruction.c - PDP-11 Instruction Table
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS10 Emulator.
// See ReadMe for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// List of PDP-11 Instruction
//
// SINGLE-OPERAND INSTRUCTION
// --------------------------
//
// General
//
//   Mnemonic Instruction                Op Code
//   -------- -----------                -------
//   CLR(B)   Clear Destination          B050DD
//   COM(B)   Complement Destination     B051DD
//   INC(B)   Increment Destination      B052DD
//   DEC(B)   Decrement Destination      B053DD
//   NEG(B)   Negate Destination         B054DD
//   TST(B)   Test Destination           B057DD
//
// Shift and Rotate
//
//   Mnemonic Instruction                Op Code
//   -------- -----------                -------
//   ASR(B)   Arithmetic Shift Right     B062DD
//   ASL(B)   Arithmetic Shift Left      B063DD
//   ROR(B)   Rotate Right               B060DD
//   ROL(B)   Rotate Left                B061DD
//   SWAB     Swap Bytes                 0003DD
//
// Multiple-Precision
//
//   Mnemonic Instruction                Op Code
//   -------- -----------                -------
//   ADC(B)   Add Carry                  B055DD
//   SBC(B)   Subtract Carry             B056DD
//   SXT      Sign Extend                0067DD
//
// Processor Status Word Operators
//
//   Mnemonic Instruction                Op Code
//   -------- -----------                -------
//   MFPS     Move Byte from PS          1067DD
//   MTPS     Move Byte to PS            1064SS
//
// DOUBLE-OPERAND INSTRUCTION
// --------------------------
//
// General
//
//   Mnemonic Instruction                    Op Code
//   -------- -----------                    -------
//   MOV(B)   Move Source to Destination     B1SSDD
//   CMP(B)   Compare Source to Destination  B2SSDD
//   ADD      Add Source to Destination      06SSDD
//   SUB      Subtract Source to Destination 16SSDD
//
// Logical
//
//   Mnemonic Instruction                    Op Code
//   -------- -----------                    -------
//   BIT(B)   Bit Test                       B3SSDD
//   BIC(B)   Bit Clear                      B4SSDD
//   BIS(B)   Bit Set                        B5SSDD
//   XOR      Exclusive OR                   074RDD
//
// PROGRAM CONTROL INSTRUCTION
// ---------------------------
//
// Branch
//
//   Mnemonic Instruction                    Op Code
//   -------- -----------                    -------
//   BR       Branch (Unconditional)         000400
//   BNE      Branch if Not Equal            001000
//   BEQ      Branch if Equal                001400
//   BPL      Branch if Plus                 100000
//   BMI      Branch if Minus                100400
//   BVC      Branch if Overflow Clear       102000
//   BVS      Branch if Overflow Set         102400
//   BCC      Branch if Carry Clear          103000
//   BCS      Branch if Carry Set            103400
//
// Signed Conditional Branch
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   BGE      Branch if Greater Than or Equal 002000
//   BLT      Branch if Less Than             002400
//   BGT      Branch if Greater Than          003000
//   BLE      Branch if Less Than or Equal    003400
//
// Unsigned Conditional Branch
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   BHI      Branch if Higher                101000
//   BLOS     Branch if Lower or Same         101400
//   BHIS     Branch if Higher or Same        103000
//   BLO      Branch if Lower                 103400
//
// Jump and Subroutine
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   JMP      Jump                            101000
//   JSR      Jump to Subroutine              004RDD
//   RTS      Return from Subroutine          00020R
//   SOB      Subtract One and Branch         077R00
//
// Trap and Interrupt
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   EMT      Emulator Trap                   104000-104377
//   TRAP     Trap                            104400-104777
//   BPT      Breakpoint Trap                 000003
//   IOT      Input/Output Trap               000004
//   RTI      Return from Interrupt           000002
//   RTT      Return from Interrupt           000006
//
// MISCELLANEOUS INSTRUCTION
// -------------------------
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   HALT     Halt                            000000
//   WAIT     Wait                            000001
//   RESET    Reset External Bus              000005
//   MFPT     Move Processor Type             000007
//
// CONDITION CODE OPERATORS
// ------------------------
//
//   Mnemonic Instruction                     Op Code
//   -------- -----------                     -------
//   CLC      Clear C (Carry)                 000241
//   CLV      Clear V (Overflow)              000242
//   CLZ      Clear Z (Zero)                  000244
//   CLN      Clear N (Negate)                000250
//   CCC      Clear All CC Bits               000257
//   SEC      Set C (Carry)                   000261
//   SEV      Set V (Overflow)                000262
//   SEZ      Set Z (Zero)                    000264
//   SEN      Set N (Negate)                  000270
//   SCC      Set All CC Bits                 000277
//   NOP      No Operation                    000240

// INSTRUCTION FORMATS
// -------------------
//
// Single Operand Instruction
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |         Opcode              |     Src/Dst     |
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// Double Operand Instruction
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |  Opcode   |       Src       |       Dst       |
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// Branch Instruction
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |       Opcode          |        Offset         |
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// Jump to Subroutine (JSR) Instruction
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |         004        |   Reg  |       Dst       |
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// Operate Instruction
//
// HALT, MFPT, RESET, RIT, RTT, and WAIT
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |                      Opcode                   | 
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// Condition Code Operator Instruction
//
// NOP, CCC, CLC, CLN, CLV, CLZ, SCC, SEC, SEN, SEV, and SEZ
//
// +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
// |              00024             |S |N |Z |V |C | 
// +--^--+--+--^--+--+--^--+--+--^--+--+--^--+--+--+
//  15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
//
// S = (Set = 1, Clear = 0)
// N = Negate Bit
// Z = Zero Bit
// V = Overflow Bit
// C = Carry Bit

PDP11_INSTRUCTION pdp11_Instructions[] =
{
	{
		"ADC", "Add Carry Word", "0055DD",
		0005500, // Op Code (Octal)
	},

	{
		"ADCB", "Add Carry Byte", "1055DD",
		0105500, // Op Code (Octal)
	},

	{
		"ADD", "Add Word", "06SSDD",
		0060000, // Op Code (Octal)
	},

	{
		"ASL", "Arithmetic Shift Left Word", "0063DD",
		0006300, // Op Code (Octal)
	},

	{
		"ASLB", "Arithmetic Shift Left Byte", "1063DD",
		0106300, // Op Code (Octal)
	},

	{
		"ASR", "Arithmetic Shift Right Word", "0062DD",
		0006200, // Op Code (Octal)
	},

	{
		"ASRB", "Arithmetic Shift Right Byte", "1062DD",
		0106200, // Op Code (Octal)
	},

	{
		"BCC", "Branch if Carry Clear", "103000",
		0103000, // Op Code (Octal)
	},

	{
		"BCS", "Branch if Carry Set", "103400",
		0103400, // Op Code (Octal)
	},

	{
		"BEQ", "Branch if Not Equal", "001400",
		0001400, // Op Code (Octal)
	},

	{
		"BGE", "Branch if Greater Than or Equal", "002000",
		0002000, // Op Code (Octal)
	},

	{
		"BGT", "Branch if Greater Than", "003000",
		0003000, // Op Code (Octal)
	},

	{
		"BHI", "Branch if Higher", "101000",
		0101000, // Op Code (Octal)
	},

	{
		"BHIS", "Branch if Higher or Same", "103000",
		0103000, // Op Code (Octal)
	},

	{
		"BIC", "Bit Clear Word", "04SSDD",
		0040000, // Op Code (Octal)
	},

	{
		"BICB", "Bit Clear Byte", "14SSDD",
		0140000, // Op Code (Octal)
	},

	{
		"BIS", "Bit Set Word", "05SSDD",
		0050000, // Op Code (Octal)
	},

	{
		"BISB", "Bit Set Byte", "15SSDD",
		0150000, // Op Code (Octal)
	},

	{
		"BIT", "Bit Test Word", "03SSDD",
		0030000, // Op Code (Octal)
	},

	{
		"BITB", "Bit Test Byte", "13SSDD",
		0130000, // Op Code (Octal)
	},

	{
		"BLE", "Branch if Less Than or Equal", "003400",
		0003400, // Op Code (Octal)
	},

	{
		"BLO", "Branch if Lower", "103400",
		0103400, // Op Code (Octal)
	},

	{
		"BLOS", "Branch if Lower or Same", "101400",
		0101400, // Op Code (Octal)
	},

	{
		"BLT", "Branch if Less Than", "002400",
		0002400, // Op Code (Octal)
	},

	{
		"BMI", "Branch if Minus", "100400",
		0100400, // Op Code (Octal)
	},

	{
		"BNE", "Branch if Not Equal", "001000",
		0001000, // Op Code (Octal)
	},

	{
		"BPL", "Branch if Plus", "100000",
		0100000, // Op Code (Octal)
	},

	{
		"BPT", "Breakpoint Trap", "000003",
		0000003, // Op Code (Octal)
	},

	{
		"BR", "Branch", "000400",
		0000400, // Op Ocde (Octal)
	},

	{
		"BVC", "Branch if Overflow Clear", "102000",
		0102000, // Op Code (Octal)
	},

	{
		"BVS", "Branch if Overflow Set", "102400",
		0102400, // Op Code (Octal)
	},

	{
		"CCC", "Clear All CC Bits", "000257",
		0000257, // Op Code (Octal)
	},

	{
		"CLC", "Clear Carry Bit (C)", "000241",
		0000241, // Op Code (Octal)
	},

	{
		"CLN", "Clear Negate Bit (N)", "000250",
		0000250, // Op Code (Octal)
	},

	{
		"CLR", "Clear Word", "0050DD",
		0005000, // Op Code (Octal)
	},

	{
		"CLRB", "Clear Byte", "1050DD",
		0105000, // Op Code (Octal)
	},

	{
		"CLV", "Clear Overflow Bit (V)", "000242",
		0000242, // Op Code (Octal)
	},

	{
		"CLZ", "Clear Zero Bit (Z)", "000244",
		0000244, // Op Code (Octal)
	},

	{
		"CMP", "Compare Word", "02SSDD",
		0020000, // Op Code (Octal)
	},

	{
		"CMPB", "Compare Byte", "12SSDD",
		0120000, // Op Code (Octal)
	},

	{
		"COM", "Complement Word", "0051DD",
		0005100, // Op Code (Octal)
	},

	{
		"COMB", "Complement Byte", "1051DD",
		0105100, // Op Code (Octal)
	},

	{
		"DEC", "Decrement Word", "0053DD",
		0005300, // Op Code (Octal)
	},

	{
		"DECB", "Decrement Byte", "1053DD",
		0105300, // Op Code (Octal)
	},

	{
		"EMT", "Emulator Trap", "104000-104377",
		0104000, // Op Code (Octal)
	},

	{
		"INC", "Increment Word", "0052DD",
		0005200, // Op Code (Octal)
	},

	{
		"INCB", "Increment Byte", "1052DD",
		0105200, // Op Code (Octal)
	},

	{
		"IOT", "Input/Output Trap", "000004",
		0000004, // Op Code (Octal)
	},

	{
		"JMP", "Jump", "0001DD",
		0000100, // Op Code (Octal)
	},

	{
		"JSR", "Jump to Subroutine", "004RDD",
		0004000, // Op Code (Octal)
	},

	{
		"HALT", "Halt Processor", "000000",
		0000000, // Op Code (Octal)
	},

	{
		"MFPS", "Move Byte from PS", "1067DD",
		0106700, // Op Code (Octal)
	},

	{
		"MFPT", "Move Processor Type", "000007",
		0000007, // Op Code (Octal)
	},

	{
		"MOV", "Move Word", "01SSDD",
		0010000, // Op Code (Octal)
	},

	{
		"MOVB", "Move Byte", "11SSDD",
		0110000, // Op Code (Octal)
	},

	{
		"MTPS", "Move Byte to PS", "1064SS",
		0106400, // Op Code (Octal)
	},

	{
		"NEG", "Negate Word", "0054DD",
		0005400, // Op Code (Octal)
	},

	{
		"NEGB", "Negate Byte", "1054DD",
		0105400, // Op Code (Octal)
	},

	{
		"NOP", "No Operation", "000240",
		0000240, // Op Code (Octal)
	},

	{
		"RESET", "Reset External Bus", "000005",
		0000005, // Op Code (Octal)
	},

	{
		"ROL", "Rotate Left Word", "0061DD",
		0006000, // Op Code (Octal)
	},

	{
		"ROLB", "Rotate Left Byte", "1061DD",
		0106000, // Op Code (Octal)
	},

	{
		"ROR", "Rotate Right Word", "0060DD",
		0006000, // Op Code (Octal)
	},

	{
		"RORB", "Rotate Right Byte", "1060DD",
		0106000, // Op Code (Octal)
	},

	{
		"RTI", "Return from Interrupt", "000002",
		0000002, // Op Code (Octal)
	},

	{
		"RTT", "Return from Trap", "000006",
		0000006, // Op Code (Octal)
	},

	{
		"RTS", "Return from Subroutine", "00020R",
		0000200, // Op Code (Octal)
	},

	{
		"SCC", "Set All CC Bits", "000277",
		0000277, // Op Code (Octal)
	},

	{
		"SEC", "Set Carry Bit (C)", "000261",
		0000261, // Op Code (Octal)
	},

	{
		"SEN", "Set Negate Bit (N)", "000270",
		0000270, // Op Code (Octal)
	},

	{
		"SEV", "Set Overflow Bit (V)", "000262",
		0000262, // Op Code (Octal)
	},

	{
		"SEZ", "Set Zero Bit (Z)", "000264",
		0000264, // Op Code (Octal)
	},

	{
		"SOB", "Subtract One and Branch", "077R00",
		0077000, // Op Code (Octal)
	},

	{
		"SUB", "Subtract Word", "16SSDD",
		0160000, // Op Code (Octal)
	},

	{
		"SWAB", "Swap Bytes", "0003DD",
		0000300, // Op Code (Octal)
	},

	{
		"SXT", "Sign Extend", "0067DD",
		0006700, // Op Code (Octal)
	},

	{
		"TRAP", "Trap", "104400-104777",
		0104400, // Op Code (Octal)
	},

	{
		"TST", "Test Word", "0057DD",
		0005700, // Op Code (Octal)
	},

	{
		"TSTB", "Test Byte", "1057DD",
		0105700, // Op Code (Octal)
	},

	{
		"WAIT", "Wait", "000001",
		0000001, // Op Code (Octal)
	},

	{
		"XOR", "Exclusive OR", "074RDD",
		0074000, // Op Code (Octal)
	},
};

