// dev_u2qba.c - Q22 Bus Interface for MicroVAX II series
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS10 Emulator.
// See ReadMe for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "vax/defs.h"
#include "dev/defs.h"

#define QBA_KEY     "U2QBA"
#define QBA_NAME    "Q22-Bus Interface for MicroVAX II Series"
#define QBA_VERSION "v0.5 (Alpha)"

#define QBA_N_IPCR   4    // Number of IPC Registers
#define QBA_N_IOREG  4096 // Number of I/O Registers
#define QBA_N_MAP    8192 // Number of Map Registers

#define BYTE OP_BYTE
#define WORD OP_WORD
#define LONG OP_LONG

// 1F40 IPCR - Interprocess Communication Registers
// (Up to 4 IPCR registers - 1F40 to 1F48)
// Note: It supports up to 4 processors on the same Q22 bus.

#define IPCR_ADDR0 0x1F40 // IPCR Register #0
#define IPCR_ADDR1 0x1F42 // IPCR Register #1
#define IPCR_ADDR2 0x1F44 // IPCR Register #2
#define IPCR_ADDR3 0x1F46 // IPCR Register #3

#define IPCR_CSR   017777500 // Base address of IPCR registers
#define IPCR_NREGS 4         // Number of Registers
#define IPCR_INT   24        // IPL 14
#define IPCR_VEC   0x204     // Interrupt Vector

#define IPCR_QPE   0x8000 // (R/?) DMA Q22-Bus Address Space Parity Error
#define IPCR_HLT   0x0100 // (R/?) Auxiliary Halt
#define IPCR_IE    0x0040 // (R/W) Doorbell Interrupt Enable
#define IPCR_LME   0x0020 // (R/W) Local Memory External Access Enable
#define IPCR_IRQ   0x0001 // (R/C) Doorbell Interrupt Request

#define IPCR_WMASK 0x0060 // Write Mask
#define IPCR_CMASK 0x0001 // Clear Mask
#define IPCR_ZMASK 0x7E9E // MBZ Mask

// Map Registers

#define MAP_V     0x80000000 // Valid bit
#define MAP_ADR   0x00007FFF // Page Frame Number (512 bytes a page)
#define MAP_WMASK 0x80007FFF // Write Mask

#define MAP_N_OFF 9
#define MAP_OFF   ((1u << MAP_N_OFF) - 1)

typedef struct {
	char        *devName;  // Device Name (QBA:)
	int         idDevice;  // Device Identification

	void        *System;   // System (VAX, PDP11, etc.)
	QBA_CALL    *Callback; // Callback Functions

	// I/O Page Table
	MAP_IO      *ioMap[QBA_N_IOREG];
	MAP_IO      *ioList;

	// Interrupt for IPCR registers
	uint32      intLevel;   // Interrupt Level (IPL)
	uint32      intMask;    // Interrupt Level Mask
	uint32      intVector;  // Interrupt Vector

	// Q22 Bus Registers
	uint16      ipcReg[QBA_N_IPCR];
	uint32      mapReg[QBA_N_MAP];
} QBA_DEVICE;

// *************************************************************

int qba_ReadIPCR(void *dptr, uint32 pAddr, uint16 *data, uint32 size)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)dptr;

	if (size == BYTE)
		*data = qba->ipcReg[0] >> ((pAddr & 1) ? 8 : 0) & BMASK;
	else
		*data = qba->ipcReg[0];

#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("%s: (R) IPCR (%08X) %04X => %04X (Size: %d)\n",
			qba->devName, pAddr, qba->ipcReg[0], *data, size);
#endif /* DEBUG */

	return QBA_OK;
}

int qba_WriteIPCR(void *dptr, uint32 pAddr, uint16 data, uint32 size)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)dptr;

	if ((size == BYTE) && (pAddr & 1))
		data <<= 8;

	qba->ipcReg[0] = (qba->ipcReg[0] & ~IPCR_WMASK) | (data & IPCR_WMASK);
	qba->ipcReg[0] &= ~(data & IPCR_CMASK);
	if ((data & (IPCR_IE|IPCR_IRQ)) == (IPCR_IE|IPCR_IRQ)) {
		qba->ipcReg[0] |= IPCR_IRQ;
		HIRQ |= qba->intMask;
		SET_IRQ;
	}

#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("%s: (W) IPCR (%08X) %04X <= %04X (Size: %d)\n",
			qba->devName, pAddr, qba->ipcReg[0], data, size);
#endif /* DEBUG */

	return QBA_OK;
}

// *************************************************************
 
// Read Word Aligned (Q22 I/O Space)
int qba_ReadIO(QBA_DEVICE *qba, uint32 ioAddr, uint16 *data, uint32 size)
{
	MAP_IO *io;

	// Access a desired I/O device
	if (io = qba->ioMap[ioAddr >> 1])
		return io->ReadIO(io->Device, ioAddr, data, size);

	return QBA_NXM;
}

// Write Word Aligned (Q22 I/O Space)
int qba_WriteIO(QBA_DEVICE *qba, uint32 ioAddr, uint32 data, uint32 size)
{
	MAP_IO *io;

	// Access a desired I/O device
	if (io = qba->ioMap[ioAddr >> 1])
		return io->WriteIO(io->Device, ioAddr, data, size);

	return QBA_NXM;
}

// Set Software Interrupt Vector
void qba_SetVector(MAP_IO *io, int newVector)
{
	INTVEC[io->intLevel] = newVector | 0x200;
	io->intVector        = newVector | 0x200;
}

// Send Interrupt to Host
void qba_SendInterrupt(MAP_IO *io)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)io->Device;

//	if (qba->ipcReg[0] & IPCR_IE) {
//		qba->ipcReg[0] |= IPCR_IRQ;

		HIRQ |= io->intMask;
		SET_IRQ;

#ifdef DEBUG
//		if (dbg_Check(DBG_INTERRUPT))
			dbg_Printf("QBA: Interrupt - IPL %02X (%d) Vector %03X (%03o)\n",
				io->intLevel, io->intLevel, io->intVector, io->intVector);
#endif /* DEBUG */
//	}
}

// *************************************************************

DEVINFO qba_Device;

int qba_SetMap(void *dptr, MAP_IO *io)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)dptr;
	MAP_IO     *cptr, *pptr = NULL;
	int        csrAddr = (io->csrAddr - 0x3FE000) >> 1;
	int        idx;

	// Check free area first. If any reserved area,
	// Report that any or all area is already reserved.
	for (idx = 0; idx < io->nRegs; idx++)
		if (qba->ioMap[csrAddr + idx])
			return QBA_RESERVED;

	// Set up callback functions for this Q22 interface.
	io->SetVector     = qba_SetVector;
	io->SendInterrupt = qba_SendInterrupt;

	// Now assign device registers to call.
	for (idx = 0; idx < io->nRegs; idx++)
		qba->ioMap[csrAddr + idx] = io;

	// Insert a I/O node into a device list.
	if (qba->ioList == NULL) {
		io->Next    = NULL;
		qba->ioList = io;
	} else {
		for (cptr = qba->ioList; cptr; cptr = cptr->Next) {
			if (io->csrAddr < cptr->csrAddr)
				break;
			pptr = cptr;
		}
		if (cptr == qba->ioList) {
			io->Next = qba->ioList;
			qba->ioList = io;
		} else {
			io->Next = pptr->Next;
			pptr->Next = io;
		}
	}

	return QBA_OK;
}

// Usage: create <device> qba [<ipl> <vector>]
void *qba_Create(MAP_DEVICE *nmap, MAP_DEVICE *map, int argc, char **argv)
{
	QBA_DEVICE  *qba;
	DEVINFO     *devInfo;
	DEVINFO     *iptr, *cptr;
	MAP_IO      *ipcr;
	void        *ioptr;
	uint32      nDevices;
	uint32      ipl, vec;
	uint32      idx;

	if (map->Device == NULL) {
		// Create a dynamic interface device.
		if ((qba = (QBA_DEVICE *)calloc(1, sizeof(QBA_DEVICE))) == NULL)
			return NULL;

		// Create a dynamic I/O device.
		if ((ipcr = (MAP_IO *)calloc(1, sizeof(MAP_IO))) == NULL) {
			free(qba);
			return NULL;
		}

		map->dtName     = qba_Device.dtName;
		map->emuName    = qba_Device.emuName;
		map->emuVersion = qba_Device.emuVersion;
		map->devInfo    = &qba_Device;
		map->Device     = qba;

		qba->devName    = map->devName;
		qba->idDevice   = 0;
		qba->intLevel   = IPCR_INT;
		qba->intMask    = (1u << IPCR_INT);
		qba->intVector  = IPCR_VEC;

		ipcr->Next      = NULL;
		ipcr->Descrip   = "IPCR Registers";
		ipcr->Device    = qba;
		ipcr->csrAddr   = IPCR_CSR;
		ipcr->nRegs     = IPCR_NREGS;
		ipcr->intLevel  = IPCR_INT;
		ipcr->intMask   = (1u << IPCR_INT);
		ipcr->intVector = IPCR_VEC;
		ipcr->ReadIO    = qba_ReadIPCR;
		ipcr->WriteIO   = qba_WriteIPCR;
//		qba_SetMap(qba, ipcr);

		// Assign IPCRs to I/O Page Area (Temp.)
		qba->ioMap[IPCR_ADDR0 >> 1] = ipcr;
//		qba->ioMap[IPCR_ADDR1 >> 1] = ipcr;
//		qba->ioMap[IPCR_ADDR2 >> 1] = ipcr;
//		qba->ioMap[IPCR_ADDR3 >> 1] = ipcr;
		qba->ioList = ipcr;

		printf("Created %s on %s: - %s %s\n",
			map->dtName, map->devName, map->emuName, map->emuVersion);
			
		return qba;
	} else {
		QBA_DEVICE *qba  = (QBA_DEVICE *)map->Device;
		DEVINFO    *iptr = map->devInfo;
		DEVINFO    *cptr = NULL;

		for (idx = 0; cptr = iptr->Devices[idx]; idx++)
			if (!strcasecmp(argv[2], cptr->dtName))
				break;

		if (cptr == NULL) {
			printf("%s: No Such Device on %s\n", 
				argv[2], map->devName);
			return NULL;
		}

		ioptr = cptr->Create(NULL, nmap, argc, argv);
		cptr->Setup(ioptr, qba, qba->System, qba->Callback);
		nmap->sysDevice = map->Device;
		nmap->sysMap    = map->sysMap;

		return ioptr;
	}
}

void qba_Setup(void *dptr, void *sys, QBA_CALL *call)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)dptr;

	qba->System    = sys;
	qba->Callback  = call;
}

void qba_Info(MAP_DEVICE *map, int argc, char **argv)
{
	QBA_DEVICE *qba = (QBA_DEVICE *)map->Device;
	MAP_IO     *io;

	// Display a listing of I/O device.
	if (qba->ioList) {
		for (io = qba->ioList; io; io = io->Next) {
			printf("%-20s %08o (%06X) %3d\n",
				io->Descrip, io->csrAddr, io->csrAddr, io->nRegs);	
		}
	} else
		printf("\nNo I/O devices.\n\n");
}

DEVINFO qba_Device =
{
	QBA_KEY,       // Key Name
	QBA_NAME,      // Emulator Name
	QBA_VERSION,   // Emulator Version
	uq_Devices,    // Listing of Unibus/Q22-Bus Devices
	DF_USE,        // Device Flags
	DT_DEVICE,     // Device Type

	NULL,          // Commands
	NULL,          // Set Commands
	NULL,          // Show Commands

	qba_Create,      // Create Routine
	NULL,            // Delete Routine
	qba_Setup,       // Setup Routine
	NULL,            // Reset Routine
	NULL,            // Attach Routine
	NULL,            // Detach Routine
	qba_Info,        // Info Routine
	NULL,            // Boot Routine    (Not Used)
	NULL,            // Execute Routine (Not Used)
};
