// ka630.c - KA630 Bus System Configuration (MicroVAX II series)
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS-10 Emulator.
// See README for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// KA630 - MicroVAX II Memory Map
//
// 00000000 +-------------------------------------+
//          |           Main Memory               |
//          |- - - - - - - - - - - - - - - - - - -|
//          |        Up to 16 MB Memory           |
// 01000000 +-------------------------------------+
//          |             Reserved                |
// 20000000 +-------------------------------------+
//          |         Q22-Bus I/O space           |
// 20002000 +-------------------------------------+
//          |             Reserved                |
// 20040000 +-------------------------------------+
//          |               ROM                   |
// 20080000 +-------------------------------------+
//          |      Local I/O Register Area        |
// 200C0000 +-------------------------------------+
//          |             Reserved                |
// 30000000 +-------------------------------------+
//          |        Q22-Bus Memory Space         |
// 303FFFFF +-------------------------------------+
//          |             Reserved                |
// 3FFFFFFF +-------------------------------------+
//
// Local Memory Area:
//
//   Main Memory       - 00000000-00FFFFFF (up to 16 MB memory)
//
// Local I/O Register Area:
//
//   CPU Registers     - 20080000-2008000F (4 16-bit CPU Registers)
//   Clock Registers   - 200B8000-2008007F (64 8-bit Clock Registers)
//
//   Q22 I/O Page      - 20000000-20001FFF (8 KB I/O Space)
//   Q22 Map Registers - 20088000-2008FFFF (8096 32-bit Map Registers)
//   Q22 Memory Space  - 30000000-3003FFFF (4 MB Memory Space)

#include "vax/defs.h"
#include "dev/defs.h"

#define KA630_DTNAME  "KA630"
#define KA630_NAME    "KA630 (MicroVAX II) Emulator"
#define KA630_VERSION "v0.7 (Alpha)"

// KA630's Q22-Bus Inferface
extern DEVINFO qba_Device;

// This extends to VAX_CPU structure.
typedef struct {
	char        *devName; // Device Name
	char        *dtName;  // Device Type Name

	VAX_CPU    *Processor; // VAX Processor
	CLK_QUEUE  Timer;      // Clock Timer
	int32      TickCount;  // Count each tick
	void       *qba;       // QBA Interface for KA630.

#ifdef TEST_PARITY
	// That is stub for required test #3 on KA630 firmware
	uint32     cpuAddr;    // CPU Parity Check Test
	uint32     cpuState;   // CPU Test State

	uint32     q22Addr1;   // Q22 Parity Check Test #1
	uint32     q22Addr2;   // Q22 Parity Check Test #2
	uint32     q22State;   // Q22 Test State
#endif /* TEST_PARITY */

	// Local I/O Registers Area (in bytes)
	uint16 cpuRegs[4];   // KA630 CPU Registers
	uint8  clkRegs[128]; // KA630 Clock Registers (TOY) w/scratch pad RAM
	uint32 clkZero;

	// Map Registers for Q22-Bus Interface
	uint32 mapRegs[8192]; // 8192 Map Registers
} KA630_DEVICE;

// ***************************************
// ******** KA630 CPU Registers **********
// ***************************************

#define CPUREG(reg) ka630->cpuRegs[reg]

#define nCPUREG 4 // Number of CPU Registers

#define nBDR    0 // Boot Diagnostics Register
#define nMSER   1
#define nCEAR   2
#define nDEAR   3

#define BDR  CPUREG(nBDR)
#define MSER CPUREG(nMSER)
#define CEAR CPUREG(nCEAR)
#define DEAR CPUREG(nDEAR)

#ifdef DEBUG
static char *cpuNames[] = { "BDR", "MSER", "CEAR", "DEAR" };
#endif /* DEBUG */

// 20080000 - KA630_BDR - Boot Diagnostics Register

#define BDR_PWROK  0x8000 // (R)   Power Ok
#define BDR_HLTENB 0x4000 // (R)   Halt Enable Switch
#define BDR_CPU    0x0C00 // (R)   CPU Arbitration ID
#define BDR_BDG    0x0300 // (R)   Language, Test, and Loopback Switch
#define BDR_DSPL   0x000F // (R/W) LED Display (0-F)

#define BDR_WMASK  0x000F // Write Mask

// 20080004 - KA630_MSER - Memory System Error Register

#define MSER_CD    0x0300 // (R)   Memory Code
#define MSER_NXM   0x0080 // (R/C) CPU Non-Existant Memory
#define MSER_LPE   0x0040 // (R/C) CPU Local Address Space Parity Error
#define MSER_QPE   0x0020 // (R/C) CPU Q22-Bus Address Space Parity Error
#define MSER_DQPE  0x0010 // (R/C) DMA Q22-Bus Address Space Parity Error
#define MSER_LEB   0x0008 // (R/C) Memory System Lost Error Bit
#define MSER_WWP   0x0002 // (R/W) Write Wrong Parity (1 = Bad, 0 = Good)
#define MSER_PEN   0x0001 // (R/W) Parity Enable

#define MSER_CMASK 0x00F8 // Clear Mask
#define MSER_WMASK 0x0003 // Write Mask

// Memory Code for MSER_CD field

#define MCD_Q22   0x00 // Q-22 Memory or Device
#define MCD_MEM   0x01 // KA630 Memory On-Board
#define MCD_ME1   0x02 // Memory Expansion Module 1
#define MCD_ME2   0x03 // Memory Expansion Module 2

// 20080008 - KA630_CEAR - CPU Error Address Register
// Valid if CPU LPE or CPU QPE are set in MSER register.

#define CEAR_PAGE  0x7FFF // (R) Local Memory Address Bits <23:9>

// 2008000C - KA630_DEAR - DMA Error Address Register
// Valid if DMA QPE is set in MSER register.

#define DEAR_PAGE  0x7FFF // (R) Local Memory Address Bits <23:9>

// ***********************************************
// ******** KA630 TOY (Clock) Registers **********
// ***********************************************

#define CLKREG(reg) ka630->clkRegs[reg << 1]

#define nCLKREG 64 // Number of Clock registers in bytes

#define nSEC       0 // Second
#define nSECLARM   1 // Second Alarm
#define nMINUTE    2 // Minute
#define nMINALRM   3 // Minute Alarm
#define nHOUR      4 // Hour
#define nHOURALRM  5 // Hour Alarm
#define nDAYWK     6 // Day of Week
#define nDAYMON    7 // Day of Month
#define nMONTH     8 // Month
#define nYEAR      9 // Year
#define nCSRA     10 // Control and Status Register #0
#define nCSRB     11 // Control and Status Register #1
#define nCSRC     12 // Control and Status Register #2
#define nCSRD     13 // Control and Status Register #3
#define nCPMBX    14 // Control Panel Mailbox

#define SECOND   CLKREG(nSECOND)
#define SECALRM  CLKREG(nSECALRM)
#define MINUTE   CLKREG(nMINUTE)
#define MINALRN  CLKREG(nMINALRM)
#define HOUR     CLKREG(nHOUR)
#define HOURALRM CLKREG(nHOURALRM)
#define DAYWK    CLKREG(nDAYWK)
#define DAYMON   CLKREG(nDAYMON)
#define MONTH    CLKREG(nMONTH)
#define YEAR     CLKREG(nYEAR)
#define CSRA     CLKREG(nCSRA)
#define CSRB     CLKREG(nCSRB)
#define CSRC     CLKREG(nCSRC)
#define CSRD     CLKREG(nCSRD)
#define CPMBX    CLKREG(nCPMBX)

// 200B8000 - TOY Clock Chip (64 8-bit Registers with scratch pad RAM)
//
//                     Address
// Reg Function        Offset   Comment
//  0  Second          00       0-59 Seconds
//  1  Second Alarm    02       0-59 Seconds
//  2  Minute          04       0-59 Minutes
//  3  Minute Alarm    06       0-59 Minutes
//  4  Hour            08       0-23 Hours
//  5  Hour Alarm      0A       0-23 Hours
//  6  Day of Week     0C       1-7  Day of Week
//  7  Day of Month    0E       1-31 Day of Month
//  8  Month           10       1-12 Months
//  9  Year            12       0-99 Years
// 10  CSRA            14
// 11  CSRB            16
// 12  CSRC            18
// 13  CSRD            1A
// 14  RAM Byte 1      1C       Scratch Pad RAM  (CPMBX)
// ...............     ..       ...........
// 63  RAM Byte 50     7E       Last byte

// TOY Register 10 - CSRA   Control and Status Register #0

#define CSRA_UIP    0x80 // (R)   Update in Progress Flag
#define CSRA_DV     0x70 // (R/W) DV0-DV2 ??
#define CSRA_RS     0x0F // (R/W) Not used??

// TOY Register 11 - CSRB   Control and Status Register #1

#define CSRB_SET    0x80 // Set Time (0 = Normal, 1 = Stop to update)
#define CSRB_PIE    0x40 // Peroidic Interrupt Enable
#define CSRB_AIE    0x20 // Alarm Interrupt Enable
#define CSRB_UIE    0x10 // Update Interrupt Enable
#define CSRB_SQWE   0x08 // Square-Wave Enable
#define CSRB_DM     0x04 // Data Mode  (0 = Binary, 1 = BCD)
#define CSRB_24     0x02 // 12/24 Hour (0 = 12, 1 = 24)
#define CSRB_DS     0x01 // Daylight Saving Enable

// TOY Register 12 - CSRC   Control and Status Register #2

// TOY Register 13 - CSRD   Control and Status Register #3

// TOY Register 14 - CPMBX  Control Panel Mailbox
// (First byte in scratch pad RAM)

#define CPMBX_LANG  0xF0 // Language
#define CPMBX_RIP   0x08 // Restart in Progress Flag
#define CPMBX_BIP   0x04 // Bootstrap in Progress Flag
#define CPMBX_HALT  0x03 // Halt Actions

// Console Message          Halt Actions
// Text Language
//
// 7 6 5 4  Langauge        1 0  Halt Action
// -------  --------        ---  -----------
// 0 0 0 1  German          0 0  Use Halt Enable (BDR 14 bit)
// 0 0 1 0  English         0 1  Restart, if that fails, Halt
// 0 0 1 1  Spanish         1 0  Reboot, if that fails, Halt
// 0 1 0 0  French          1 1  Halt
// 0 1 0 1  Italian
// 0 1 1 0  Danish
// 0 1 1 1  Dutch
// 1 0 0 0  Finish
// 1 0 0 1  Norwegian
// 1 0 1 0  Swedish
// 1 0 1 1  Portuguese 

// *******************************************
// ******* KA630 Machine Check Code **********
// *******************************************

#define MCHK_FSD 0x01 // Impossible Microcode State (FSD)
#define MCHK_SSD 0x02 // Impossible Microcode State (SSD)
#define MCHK_FP0 0x03 // Undefined FPU Error Code 0
#define MCHK_FP7 0x04 // Undefined FPU Error Code 7
#define MCHK_TBM 0x05 // Undefined Memory Management Status (TB Miss)
#define MCHK_MZ  0x06 // Undefined Memory Management Status (M = 0)
#define MCHK_P0  0x07 // Process PTE in P0 Space
#define MCHK_P1  0x08 // Process PTE in P1 Space
#define MCHK_IPL 0x09 // Undefined Interrupt ID Code
#define MCHK_RBV 0x80 // Read Bus Error, Address Parameter is Virtual
#define MCHK_RBP 0x81 // Read Bus Error, Address Parameter is Physical
#define MCHK_WBV 0x82 // Write Bus Error, Address Parameter is Virtual
#define MCHK_WBP 0x83 // Wirte Bus Error, Address Parameter is Physical

// *******************************************
// ************* KA630 Halt Code *************
// *******************************************

#define HLT_SWITCH 0x02 // HALT L Asserted (Halt Button)
#define HLT_PON    0x03 // Initial Power On
#define HLT_ISNV   0x04 // Interrupt Stack Not Valid During Exception
#define HLT_MCHK   0x05 // Machine-Check During Machine-Check/KSNV
#define HLT_INST   0x06 // HALT Instruction Executed in Kernel Mode
#define HLT_SCB11  0x07 // SCB Vector <1:0> = 11
#define HLT_SCB10  0x08 // SCB Vector <1:0> = 10
#define HLT_CHM    0x0A // CHMx Executed While On Interrupt Stack
#define HLT_MMCHK  0x10 // ACV/TNV During Machine-Check Exception
#define HLT_MKSNV  0x11 // ACV/TNV During Kernel-Stack-Not-Valid Exception

// *******************************************
// ******** KA630 Q22-Bus Registers **********
// *******************************************

// 1F40 IPCR - Interprocess Communication Registers
// (Up to 4 IPCR registers - 1F40 to 1F48)
// Note: It supports up to 4 CPUs on the same Q22 bus.

#define IPCR_QPE   0x8000 // (R/?) DMA Q22-Bus Address Space Parity Error
#define IPCR_HLT   0x0100 // (R/?) Auxiliary Halt
#define IPCR_IE    0x0040 // (R/W) Doorbell Interrupt Enable
#define IPCR_LME   0x0020 // (R/W) Local Memory External Access Enable
#define IPCR_IRQ   0x0001 // (R/C) Doorbell Interrupt Request

#define IPCR_WMASK 0x0060 // Write Mask
#define IPCR_CMASK 0x0001 // Clear Mask
#define IPCR_ZMASK 0x7E9E // MBZ Mask

#define IPCR_ADDR  0x3FFF40 // Q22 Bus Address
#define nIPCR_INT  24       // IPL 14
#define IPCR_INT   (1u << nIPCR_INT)
#define IPCR_VEC   0x204    // Interrupt Vector

// Map Registers

#define Q22MAP_V     0x80000000 // Valid bit
#define Q22MAP_ADR   0x00007FFF // Page Frame Number (512 bytes a page)
#define Q22MAP_N_OFF 9
#define Q22MAP_OFF   ((1u << Q22MAP_N_OFF) - 1)
#define Q22MAP_PAGSZ (1u << Q22MAP_N_OFF)
#define Q22MAP_WMASK 0x80007FFF // Write Mask

// **********************************************************
// ***************** Privileged Registers *******************
// **********************************************************

#ifdef DEBUG
static int  regLength = 0x40;
static char *regNames[] = {
	"KSP",     // (R/W) 00 Kernel Stack Pointer
	"ESP",     // (R/W) 01 Executive Stack Pointer
	"SSP",     // (R/W) 02 Supervisor Stack Pointer
	"USP",     // (R/W) 03 User Stack Pointer
	"ISP",     // (R/W) 04 Interrupt Stack Pointer
	NULL,      //       05 Unknown
	NULL,      //       06 Unknown
	NULL,      //       07 Unknown
	"P0BR",    // (R/W) 08 P0 Base Register
	"P0LR",    // (R/W) 09 P0 Length Register
	"P1BR",    // (R/W) 0A P1 Base Register
	"P1LR",    // (R/W) 0B P1 Length Register
	"SBR",     // (R/W) 0C System Base Register
	"SLR",     // (R/W) 0D System Length Register
	NULL,      //       0E Unknown
	NULL,      //       0F Unknown
	"PCBB",    // (R/W) 10 Process Control Block Base 
	"SCBB",    // (R/W) 11 System Control Block Base
	"IPL",     // (R/W) 12 Interrupt Priority Level
	"ASTLVL",  // (R/W) 13 AST Level
	"SIRR",    // (W)   14 Software Interrupt Request
	"SISR",    // (R/W) 15 Software Interrupt Summary
	NULL,      //       16 Unknown
	NULL,      //       17 Unknown
	"ICCS",    // (R/W) 18 Interval Clock Control
	"NICR",    // (W)   19 Next Interval Count
	"ICR",     // (R)   1A Interval Count
	"TODR",    // (R/W) 1B Time of Year
	NULL,      //       1C Unknown
	NULL,      //       1D Unknown
	NULL,      //       1E Unknown
	NULL,      //       1F Unknown
	"RXCS",    // (R/W) 20 Console Receiver Status
	"RXDB",    // (R)   21 Console Receiver Data Buffer
	"TXCS",    // (R/W) 22 Console Transmit Status
	"TXDB",    // (W)   23 Console Transmit Data Buffer
	NULL,      //       24 Unknown
	NULL,      //       25 Unknown
	NULL,      //       26 Unknown
	NULL,      //       27 Unknown
	"ACCS",    //       28 Accelerator Control and Status
	"SAVISP",  //       29 Console Saved ISP (Interrupt Stack Pointer)
	"SAVPC",   //       2A Console Saved PC  (Program Counter)
	"SAVPSL",  //       2B Console Saved PSL (Processor Status Longword)
	NULL,      //       2C Unknown
	NULL,      //       2D Unknown
	NULL,      //       2E Unknown
	NULL,      //       2F Unknown
	"SBIFS",   //       30 SBI Fault Status
	"SBIS",    //       31 SBI silo
	"SBISC",   //       32 SBI silo comparator
	"SBIMT",   //       33 SBI Maintenance
	"SBIER",   //       34 SBI Error
	"SBITA",   //       35 SBI Timeout Address
	"SBIQC",   //       36 SBI Quadword Clear
	"IORESET", //       37 I/O Reset
	"MAPEN",   // (R/W) 38 Memory Management Enable
	"TBIA",    // (W)   39 Translation Buffer Invalidate All
	"TBIS",    // (W)   3A Translation Buffer Invalidate Single
	NULL,      //       3B Unknown
	"MBRK",    //       3C Microprogram Breakpoint
	"PME",     // (R/W) 3D Performance Monitor Enable
	"SID",     // (R)   3E System Identification
	"TBCHK",   // (W)   3F Translation Buffer Check
};
#endif /* DEBUG */

// System ID Longword
//
//  31   24                                          0
// +-------+------------------------------------------+
// |   8   |               Reserved                   |
// +-------+------------------------------------------+

#define KA630_SYSID    0x08000000
//#define KA630_SYSID    0x0A000000

// Privileged Register for MicroVAX II system

#define nKSP     0x00 // Kernel Stack Pointer
#define nESP     0x01 // Executive Stack Pointer
#define nSSP     0x02 // Supervisor Stack Pointer
#define nUSP     0x03 // User Stack Pointer
#define nISP     0x04 // Interrupt Stack Pointer

#define nP0BR    0x08 // P0 Base Register
#define nP0LR    0x09 // P0 Length Register
#define nP1BR    0x0A // P1 Base Register
#define nP1LR    0x0B // P1 Length Register
#define nSBR     0x0C // System Base Register
#define nSLR     0x0D // System Length Register

#define nPCBB    0x10 // Process Control Block Base
#define nSCBB    0x11 // System Control Block Base
#define nIPL     0x12 // Interrupt Priority Level
#define nASTLVL  0x13 // AST Level
#define nSIRR    0x14 // Software Interrupt Request
#define nSISR    0x15 // Software Interrupt Summary

#define nICCS    0x18 // Interval Clock Control Status Register

#define nRXCS    0x20 // Console Receive Control and Status Register
#define nRXDB    0x21 // Console Receive Data Buffer Register
#define nTXCS    0x22 // Console Transmit Control and Status Register
#define nTXDB    0x23 // Console Transmit Data Buffer Register

#define nSAVISP  0x29 // Console Saved ISP
#define nSAVPC   0x2A // Console Saved PC
#define nSAVPSL  0x2B // Console Saved PSL

#define nIORESET 0x37 // I/O Reset
#define nMAPEN   0x38 // Map Enable
#define nTBIA    0x39 // TB Clear All Entries
#define nTBIS    0x3A // TB Clear Single Entry
#define nSID     0x3E // System Identification
#define nTBCHK   0x3F // TB Check

#define SCBB_MBZ   0x000001FF // Must be zeros.
#define BR_MASK    0xFFFFFFFC // xxBR Mask
#define LR_MASK    0x003FFFFF // xxLR Mask

// Independent Implmentations for KA630 Processor

#define ACCS     PRN(nACCS)
#define ICCS     PRN(nICCS)
#define SAVISP   PRN(nSAVISP)
#define SAVPC    PRN(nSAVPC)
#define SAVPSL   PRN(nSAVPSL)
#define IORESET  PRN(nIORESET)

// Interval Clock Control Status Registers (Subset)

#define ICCS_IE     0x00000040  // Interrupt Enable
#define ICCS_WMASK  0x00000040  // Write Mask
#define ICCS_TICK   10000       // Each tick in usecs (10 msecs).
#define ICCS_SECOND 100         // Each second tick

#define nICCS_INT   0
#define ICCS_INT    (1u << nICCS_INT)

// Console Terminal Registers

// RXCS - Console Receive Control and Status Register
#define RXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define RXCS_RDY   0x00000080 // (R)   Ready
#define RXCS_IE    0x00000040 // (R/W) Interrupt Enable

// RXDB - Console Receive Data Buffer Register
#define RXDB_ERROR 0x00008000 // (R)   Error
#define RXDB_ID    0x00000F00 // (R)   Identification
#define RXDB_DATA  0x000000FF // (R)   Data Received

// TXCS - Console Terminal Control and Status Register
#define TXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define TXCS_RDY   0x00000080 // (R)   Ready
#define TXCS_IE    0x00000040 // (R/W) Interrupt Enable

// TXDB - Console Terminal Data Buffer Register
#define TXDB_MBZ   0xFFFFF000 // Must be zeros
#define TXDB_ID    0x00000F00 // (R/W) Identification
#define TXDB_DATA  0x000000FF // (R/W) Data Transmitted

// Read Privileged Register
int32 ka630_ReadRegister(int32 pReg)
{
	int32 data;

	switch (pReg) {
		case nKSP:
			data = (PSL & PSL_IS) ? KSP : SP;
			break;

		case nISP:
			data = (PSL & PSL_IS) ? SP : ISP;
			break;

		case nIPL:
			IPL = PSB.ipl;
			data = PSB.ipl;
			break;

		case nRXCS:
			data = vax_ReadRXCS(vax->Console);
			break;

		case nRXDB:
			data = vax_ReadRXDB(vax->Console);
			break;

		case nTXCS:
			data = vax_ReadTXCS(vax->Console);
			break;

//		case nTXDB:
//			data = vax_ReadTXDB(vax->Console);
//			break;

		default:
			data = PRN(pReg);
			break;
	}

#ifdef DEBUG
	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA630: (R) %s (%02X) => %08X\n", name, pReg, data);
	}
#endif /* DEBUG */

	return data;
}

void ka630_WriteRegister(int32 pReg, int32 data)
{
	switch (pReg) {
		case nKSP: // Kernel Stack Pointer
			((PSL & PSL_IS) ? KSP : SP) = data;
			break;

		case nESP: // Executive Stack Pointer
		case nSSP: // Supervisor Stack Pointer
		case nUSP: // User Stack Pointer
			PRN(pReg) = data;
			break;

		case nISP: // Interrupt Stack Pointer
			((PSL & PSL_IS) ? SP : ISP) = data;
			break;

		case nP0BR:
		case nP1BR:
		case nSBR:
			PRN(pReg) = data & BR_MASK;
			vax_ClearTBTable(pReg == nSBR);
			break;

		case nP0LR:
		case nP1LR:
		case nSLR:
			PRN(pReg) = data & LR_MASK;
			vax_ClearTBTable(pReg == nSLR);
			break;
		
		case nPCBB:
		case nSCBB:
			PRN(pReg) = data & LALIGN;
			break;

		case nIPL:
			PSB.ipl = data;
			IPL = data & PSL_M_IPL;
			break;

		case nSIRR:
			if ((data > 0xF) || (data == 0))
				RSVD_OPND_FAULT;
			SISR |= (1 << data);
			break;

		case nSISR:
			SISR = data & SISR_MASK;
			break;

		case nASTLVL:
			if ((uint32)data > AST_MAX)
				RSVD_OPND_FAULT;
			ASTLVL = data;
			break;

		case nICCS:
			// Subset implementation in MicroVAX series
			ICCS = data & ICCS_WMASK;
			break;

		case nRXCS:
//			if (data & RXCS_MBZ)
//				RSVD_OPND_FAULT;
			vax_WriteRXCS(vax->Console, data);
			break;

		case nRXDB:
//			vax_WriteRXDB(vax->Console, data);
			break;

		case nTXCS:
//			if (data & TXCS_MBZ)
//				RSVD_OPND_FAULT;
			vax_WriteTXCS(vax->Console, data);
			break;

		case nTXDB:
			vax_WriteTXDB(vax->Console, data);
			break;

		case nSAVISP:
		case nSAVPC:
		case nSAVPSL:
			PRN(pReg) = data;
			break;

		case nMAPEN:
			MAPEN = data & 1;
			break;

		case nTBIA:
			vax_ClearTBTable(1);
			break;

		case nTBIS:
			vax_ClearTBEntry(data);
			break;

		case nTBCHK:
			if (vax_CheckTBEntry(data))
				CC |= CC_V;
			break;
	}

#ifdef DEBUG
	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA630: (W) %s (%02X) <= %08X\n", name, pReg, data);
	}
#endif /* DEBUG */
}

// ***********************************************************
// *************** KA630 System Configuration ****************
// ***********************************************************

// *********** Clock Timer System *************

// Tick each 10 milliseconds
// for ICCS register and RTC (TOY) chip
void ka630_ClockTimer(void *device)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)device;
	VAX_CPU *vax = ka630->Processor;

	// Generate IPL 22 signal each 10 milliseconds
	if (ICCS & ICCS_IE) {
		HIRQ |= ICCS_INT;
		SET_IRQ;
	}

	if (ka630->TickCount++ == ICCS_SECOND) {
		ka630->TickCount = 0;

#ifdef DEBUG
		// Instructions Per Second Meter
		if (dbg_Check(DBG_IPS))
			dbg_Printf("KA630: %d ips\n", vax->ips);
#else
		printf("KA630: %d ips\n", vax->ips);
#endif /* DEBUG */
		vax->ips = 0;
	}
}

// Enable Clock Timer
void ka630_StartTimer(void *device)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)device;

	ka630->TickCount = 0;
	ka630->Timer.Flags |= CLK_ENABLE;
}

// Disable Clock Timer
void ka630_StopTimer(void *device)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)device;

	ka630->Timer.Flags &= ~CLK_ENABLE;
}

int32 ka630_CheckTimer(uint32 ipl)
{
//	if (ipl > ICCS_IPL)
//		return -1;
//	if (HIRQ & (1u << ICCS_IPL))
//		return ICCS_IPL;
//	return 0;
}

// ******************************************

// Note: It only works for little endian machines at this time.

#define IN_RAM(addr) \
	((addr) < vax->sizeRAM)

#define IN_ROM(addr) \
	(((addr) >= 0x20040000) && ((addr) < 0x20080000))
#define IN_CPUREG(addr) \
	(((addr) >= 0x20080000) && ((addr) < 0x20088000))
#define IN_CLKREG(addr) \
	(((addr) >= 0x200B8000) && ((addr) < 0x200C0000))

#define IN_Q22IO(addr) \
	(((addr) >= 0x20000000) && ((addr) < 0x20001FFF))
#define IN_Q22MAP(addr) \
	(((addr) >= 0x20008800) && ((addr) < 0x2008FFFF))
#define IN_Q22MEM(addr) \
	(((addr) >= 0x30000000) && ((addr) < 0x303FFFFF))

static const align[]  = { 0x00FFFFFF, 0x00FFFFFF, 0x0000FFFF, 0x000000FF };
static const insert[] = { 0x00000000, 0x000000FF, 0x0000FFFF, 0x00FFFFFF };

// Aligned RAM Access
#define BMEM(addr)  ((uint8 *)vax->RAM)[addr]
#define WMEM(addr)  ((uint16 *)vax->RAM)[addr]
#define LMEM(addr)  ((uint32 *)vax->RAM)[addr]

// Unaligned RAM Access
#define BMEMU(addr) *(uint8 *)(&vax->RAM[addr])
#define WMEMU(addr) *(uint16 *)(&vax->RAM[addr])
#define LMEMU(addr) *(uint32 *)(&vax->RAM[addr])

// Aligned ROM Access
#define BROM(addr) ((uint8 *)vax->ROM)[addr]
#define WROM(addr) ((uint16 *)vax->ROM)[addr]
#define LROM(addr) ((uint32 *)vax->ROM)[addr]

// Unaligned ROM Access
#define BROMU(addr) *(uint8 *)(&vax->ROM[addr])
#define WROMU(addr) *(uint16 *)(&vax->ROM[addr])
#define LROMU(addr) *(uint32 *)(&vax->ROM[addr])

// Aligned Register Access
#define BREG(x, addr) ((uint8 *)(x))[addr]
#define WREG(x, addr) ((uint16 *)(x))[addr]
#define LREG(x, addr) ((uint32 *)(x))[addr]

// Unaligned Register Access
#define BREGU(x, addr) *(uint8 *)(&(x)[addr])
#define WREGU(x, addr) *(uint16 *)(&(x)[addr])
#define LREGU(x, addr) *(uint32 *)(&(x)[addr])

#define LONG OP_LONG
#define WORD OP_WORD
#define BYTE OP_BYTE

// Halt Action
inline void ka630_HaltAction(uint32 code)
{
	// Save ISP, PC, and PSL into SAVISP, SAVPC, SAVPSL.
	SAVISP = ISP;
	SAVPC  = PC;
	SAVPSL = PSL | CC | (code << 8) | (MAPEN ? 0x8000 : 0);

#ifdef DEBUG
	dbg_Printf("VAX: Halt Action - Reason %d\n", code);
#endif /* DEBUG */

	// Reset Processor Registers
	PC     = 0x20040000;
	PSL    = 0x041F0000;
	CC     = 0;
	MAPEN  = 0;
	SISR   = 0; 
	ASTLVL = 4;

	// Go Next Instruction
	FLUSH_ISTR;
	ABORT(NEXT);
}

// Machine-Check Exception
void ka630_MachineCheck(void *device)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)device;

	vax_Exception(SCB_MCHK, 0, IE_SVE);
	IN_IE = 1;
	vax_Write(SP - 16, 0xC, LONG, WA);
	vax_Write(SP - 12, P3, LONG, WA);
	vax_Write(SP - 8,  P2, LONG, WA);
	vax_Write(SP - 4,  P1, LONG, WA);
	SP -= 16;
	IN_IE = 0;

#ifdef DEBUG
	DumpHistory();
#endif /* DEBUG */
}

// Read CPU Register (Aligned Longword)
inline uint32 ka630_ReadCPU(KA630_DEVICE *ka630, uint32 pAddr, uint32 size)
{
	uint32 reg, data;

	reg = (pAddr & 0x7FFF) >> 2;
	if (reg < nCPUREG) {
		if ((size <= WORD) && (pAddr & 2))
			data = 0;
		else if (size == BYTE)
			data = (CPUREG(reg) >> ((pAddr & 1) ? 8 : 0)) & BMASK;
		else
			data = CPUREG(reg);
	} else {
		// Non-Existant Memory Exception
		return 0;
	}

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("KA630: (R) %s (%08X) %08X => %08X (Size: %d)\n",
			cpuNames[reg], pAddr, CPUREG(reg), data, size);
#endif /* DEBUG */

	return data;
}

void ka630_TestParity(VAX_CPU *, uint32);
void ka630_WriteParity(VAX_CPU *, uint32);

// Write CPU Register (Aligned Longword)
inline void ka630_WriteCPU(KA630_DEVICE *ka630, uint32 pAddr, uint32 data, uint32 size)
{
	uint32 reg = (pAddr & 0x7FFF) >> 2;

	switch (reg) {
		case nBDR: // Boot and Diagnostics Register
			if ((size <= WORD) && (pAddr & 2))
				break;
			if ((size == BYTE) && (pAddr & 1))
				break;
			BDR = (BDR & ~BDR_WMASK) | (data & BDR_WMASK);	

#ifdef TEST_PARITY
			if ((BDR & BDR_DSPL) == 0) {
				vax->TestParity  = NULL;
				vax->WriteParity = NULL;
			} else {
				vax->TestParity  = ka630_TestParity;
				vax->WriteParity = ka630_WriteParity;
			}
#endif /* TEST_PARITY */

//				printf("[Trace On]\n");
//				dbg_SetMode(DBG_TRACE|DBG_DATA);
			break;

		case nMSER: // Memory System Error Register
			if ((size <= WORD) && (pAddr & 2))
				break;
			if ((size == BYTE) && (pAddr & 1))
				break;
			MSER = (MSER & ~MSER_WMASK) | (data & MSER_WMASK);
			MSER &= ~(data & MSER_CMASK);
#ifdef TEST_PARITY
			if ((data & MSER_PEN) == 0) {
				ka630->cpuState = ka630->q22State = 0;
				ka630->cpuAddr  = -1;
				ka630->q22Addr1 = ka630->q22Addr2 = -1;
			}
#endif /* TEST_PARITY */
			break;

		case nCEAR: // CPU Error Address Register
		case nDEAR: // DMA Error Address Register
			// Read-Only Registers - Do nothing
			break;

		default:
			// Non-Existant Memory Exception
			return;
	}

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("KA630: (W) %s (%08X) %08X <= %08X (Size: %d)\n",
			cpuNames[reg], pAddr, CPUREG(reg), data, size);
#endif /* DEBUG */
}

inline uint32 ka630_ReadClock(KA630_DEVICE *ka630, uint32 pAddr, uint32 size)
{
	uint32 reg = pAddr & 0x7FFF;
	uint32 data;

	if (reg < (nCLKREG << 1)) {
		if (size >= LONG)
			data = LREGU(ka630->clkRegs, reg);
		else if (size == WORD)
			data = WREGU(ka630->clkRegs, reg);
		else
			data = BREGU(ka630->clkRegs, reg);
	} else
		return 0;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("KA630: (R) Clock Register %d => %04X (Size %d) at PC %08X\n",
			reg >> 1, data, size, faultPC);
#endif /* DEBUG */

	return data;
}

inline void ka630_WriteClock(KA630_DEVICE *ka630, uint32 pAddr, uint32 data, uint32 size)
{
	uint32 reg = pAddr & 0x7FFF;

	if (reg < (nCLKREG << 1)) {
		data &= (pAddr & 1) ? 0xFF00FF00 : 0x00FF00FF;
		if (size >= LONG)
			LREGU(ka630->clkRegs, reg) = data;
		else if (size == WORD)
			WREGU(ka630->clkRegs, reg) = data;
		else
			BREGU(ka630->clkRegs, reg) = data;
	} else
		return;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("KA630: (W) Clock Register %d <= %04X (Size %d) at PC %08X\n",
			reg >> 1, data, size, faultPC);
#endif /* DEBUG */
}

// *******************************************************
// **************** Q22-Bus Interface ********************
// *******************************************************

// *********** Local I/O Map Registers ************

// Read Longword Aligned (Q22-Bus Map Registers)
inline uint32 ka630_ReadMap(KA630_DEVICE *ka630, uint32 pAddr)
{
	uint32 reg  = (pAddr & 0x7FFF) >> 2;
	uint32 data = ka630->mapRegs[reg];

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (R) Map Register %d (%04X) => %08X at PC %08X\n",
			reg, pAddr & 0x7FFF, data, faultPC);
#endif /* DEBUG */

	return data;
}

// Write Longword Aligned (Map Registers)
inline void ka630_WriteMap(KA630_DEVICE *ka630, uint32 pAddr, uint32 data)
{
	uint32 reg = (pAddr & 0x7FFF) >> 2;

	ka630->mapRegs[reg] = data & Q22MAP_WMASK;

#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (W) Map Register %d (%04X) <= %08X at PC %08X\n",
			reg, pAddr & 0x7FFF, data, faultPC);
#endif /* DEBUG */
}

// ********* Local Memory Accesses for Q22 I/O Devices **********

void *ka630_GetHostAddr(void *dptr, uint32 ioAddr)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	uint32 mAddr;

	if (ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] & Q22MAP_V) {
		mAddr = (ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] << Q22MAP_N_OFF) |
			(ioAddr & Q22MAP_OFF);
		return &vax->RAM[mAddr];
	}

	return NULL;
}

// Read data aligned to a Q22 I/O device.
uint32 ka630_ReadDataIO(void *dptr, uint32 pAddr, uint32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	uint32 mAddr;

	// Translate to I/O Address from low/high CPU address
	// High CPU Address - 30000000 to 303FFFFF
	pAddr &= 0x3FFFFF;

	if (ka630->mapRegs[pAddr >> Q22MAP_N_OFF] & Q22MAP_V) {
		mAddr = (ka630->mapRegs[pAddr >> Q22MAP_N_OFF] << Q22MAP_N_OFF) |
			(pAddr & Q22MAP_OFF);

#ifdef TEST_PARITY
		// "This memory location is bad" Test
		if ((MSER & MSER_PEN) &&
	 	    ((mAddr == ka630->q22Addr1) || (mAddr == ka630->q22Addr2))) {
#ifdef DEBUG
			dbg_Printf("PAR: Parity Error Trap on loc %08X at PC %08X\n",
				mAddr, faultPC);
#endif /* DEBUG */
			MSER |= (MSER_CD|MSER_QPE);
			if (ka630->q22State++ < 1)
				CEAR = pAddr >> 9;
			else
				MSER |= MSER_LEB;

			// Go to machine check exception
			P1 = 0;
			P2 = pAddr;
			P3 = MCHK_RBP;
			ABORT(-SCB_MCHK);
		}
#endif /* TEST_PARITY */

		if (size >= LONG)
			return LMEM(mAddr >> 2);
		else if (size == WORD)
			return WMEM(mAddr >> 1);
		else
			return BMEM(mAddr);

	}

	// Non-Existant Memory Error
#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (R) Non-Existant Memory at PC %08X\n", faultPC);
#endif /* DEBUG */

	MSER |= (MSER_CD|MSER_NXM);

	// Go to machine check exception
	P1 = 0;
	P2 = pAddr;
	P3 = MCHK_RBP;
	ABORT(-SCB_MCHK);
}

// Write data aligned from a Q22 I/O Device
void ka630_WriteDataIO(void *dptr, uint32 pAddr, uint32 data, uint32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	uint32 mAddr;

	// Translate to I/O Address from CPU address
	// CPU Address - 30000000 to 303FFFFF
	pAddr &= 0x3FFFFF;

	if (ka630->mapRegs[pAddr >> Q22MAP_N_OFF] & Q22MAP_V) {
		mAddr = (ka630->mapRegs[pAddr >> Q22MAP_N_OFF] << Q22MAP_N_OFF) |
			(pAddr & Q22MAP_OFF);

#ifdef TEST_PARITY
		if (MSER & MSER_PEN) {
			if (MSER & MSER_WWP) {
				ka630->cpuAddr = mAddr;
				if (ka630->q22Addr1 == -1)
					ka630->q22Addr1 = mAddr;
				else
					ka630->q22Addr2 = mAddr;
#ifdef DEBUG
				dbg_Printf("PAR: Set Wrong Parity on loc %08X at PC %08X\n",
					mAddr, faultPC);
#endif /* DEBUG */
			} else if (ka630->cpuAddr == mAddr) {
				ka630->cpuAddr = -1;
				if (ka630->q22Addr1 == mAddr)
					ka630->q22Addr1 = -1;
				else if (ka630->q22Addr2 == mAddr)
					ka630->q22Addr2 = -1;
#ifdef DEBUG
				dbg_Printf("PAR: Clear Wrong Parity on loc %08X at PC %08X\n",
					mAddr, faultPC);
#endif /* DEBUG */
			}
		}
#endif /* TEST_PARITY */

		if (size >= LONG)
			LMEM(mAddr >> 2) = data;
		else if (size == WORD)
			WMEM(mAddr >> 1) = data;
		else
			BMEM(mAddr) = data;
		return;
	}

	// Non-Existant Memory Error
#ifdef DEBUG
	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (W) Non-Existant Memory at PC %08X\n", faultPC);
#endif /* DEBUG */

	MSER |= (MSER_CD|MSER_NXM);

	// Go to machine check exception
	P1 = 0;
	P2 = pAddr;
	P3 = MCHK_WBP;
	ABORT(-SCB_MCHK);
}

// Read block to a Q22 I/O device
uint32 ka630_ReadBlockIO(void *dptr, uint32 ioAddr, uint8 *data, uint32 szBytes)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	VAX_CPU      *vax   = ka630->Processor;
	uint32       mapAddr, idxAddr = 0;
	uint32       cntBytes;

	cntBytes = Q22MAP_PAGSZ - (ioAddr & Q22MAP_OFF);
	if (szBytes < cntBytes)
		cntBytes = szBytes;

	while (szBytes) {
		if ((ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] & Q22MAP_V) == 0) {
			// Non-Existant Memory Error
			MSER |= (MSER_CD|MSER_NXM);

// 		Go to machine check exception
//			P1 = 0;
//			P2 = ioAddr;
//			P3 = MCHK_RBP;
//			ABORT(-SCB_MCHK);

			return szBytes;
		}
			
		mapAddr = (ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] << Q22MAP_N_OFF) |
			(ioAddr & Q22MAP_OFF);

		memcpy(&data[idxAddr], &vax->RAM[mapAddr], cntBytes);

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			dbg_Printf("I/O Address: %08X  Host Address: %08X  Size: %d bytes\n",
				ioAddr, mapAddr, cntBytes);
			PrintDump(idxAddr, &data[idxAddr], cntBytes);
		}
#endif /* DEBUG */

		// Increment them by 512-bytes page.
		ioAddr   += cntBytes;
		idxAddr  += cntBytes;
		szBytes  -= cntBytes;
		cntBytes =  (szBytes > Q22MAP_PAGSZ) ? Q22MAP_PAGSZ : szBytes;
	}

	return szBytes;
}

// Write block from a Q22 I/O device
uint32 ka630_WriteBlockIO(void *dptr, uint32 ioAddr, uint8 *data, uint32 szBytes)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	VAX_CPU      *vax   = ka630->Processor;
	uint32       mapAddr, idxAddr = 0;
	uint32       cntBytes;

	cntBytes = Q22MAP_PAGSZ - (ioAddr & Q22MAP_OFF);
	if (szBytes < cntBytes)
		cntBytes = szBytes;

	while (szBytes) {
		if ((ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] & Q22MAP_V) == 0) {
			// Non-Existant Memory Error
			MSER |= (MSER_CD|MSER_NXM);

// 		Go to machine check exception
//			P1 = 0;
//			P2 = ioAddr;
//			P3 = MCHK_RBP;
//			ABORT(-SCB_MCHK);

			return szBytes;
		}
			
		mapAddr = (ka630->mapRegs[ioAddr >> Q22MAP_N_OFF] << Q22MAP_N_OFF) |
			(ioAddr & Q22MAP_OFF);

		memcpy(&vax->RAM[mapAddr], &data[idxAddr], cntBytes);

#ifdef DEBUG
		if (dbg_Check(DBG_IODATA)) {
			dbg_Printf("I/O Address: %08X  Host Address: %08X  Size: %d bytes\n",
				ioAddr, mapAddr, cntBytes);
			PrintDump(idxAddr, &data[idxAddr], cntBytes);
		}
#endif /* DEBUG */

		// Increment them by 512-bytes page.
		ioAddr   += cntBytes;
		idxAddr  += cntBytes;
		szBytes  -= cntBytes;
		cntBytes =  (szBytes > Q22MAP_PAGSZ) ? Q22MAP_PAGSZ : szBytes;
	}

	return szBytes;
}

// *********** I/O Register Access ************

// Read Word Aligned (Q22 I/O Space)
uint32 ka630_ReadIO(KA630_DEVICE *ka630, uint32 pAddr, uint32 size)
{
	uint32 ioAddr = pAddr & 0x1FFF;
	uint16 data;
	void   *qba;

#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS)) {
	{
		uint32 addr = ioAddr + 0x3FE000;
		dbg_Printf("QBA: (R) %08X => %06X (%08o) at PC %08X\n",
			pAddr, addr, addr, faultPC);
	}
#endif /* DEBUG */

	// Read I/O Register from Q22-Bus I/O Device...
	qba = ka630->qba;
	if (qba && (qba_ReadIO(qba, ioAddr, &data, size) == QBA_OK))
		return data;

	// Non-Existant Device Exception
#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (R) Non-Existant Device at PC %08X\n", faultPC);
#endif /* DEBUG */

	MSER = (MSER & ~MSER_CD) | MSER_NXM;

	// Go to machine check exception
	P1 = 0;
	P2 = pAddr;
	P3 = MCHK_RBP;
	ABORT(-SCB_MCHK);
}

// Write Word Aligned (Q22 I/O Space)
void ka630_WriteIO(KA630_DEVICE *ka630, uint32 pAddr, uint32 data, uint32 size)
{
	uint32 ioAddr = pAddr & 0x1FFF;
	void   *qba;

#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS)) {
	{
		int addr = ioAddr + 0x3FE000;
		dbg_Printf("QBA: (W) %08X => %06X (%08o) at PC %08X\n",
			pAddr, addr, addr, faultPC);
	}
#endif /* DEBUG */

	// Write I/O Register to Q22-Bus I/O Device...
	qba = ka630->qba;
	if (qba && (qba_WriteIO(qba, ioAddr, data, size) == QBA_OK))
		return;

	// Non-Existant Device Exception
#ifdef DEBUG
//	if (dbg_Check(DBG_IOREGS))
		dbg_Printf("QBA: (W) Non-Existant Device at PC %08X\n", faultPC);
#endif /* DEBUG */

	MSER = (MSER & ~MSER_CD) | MSER_NXM;

	// Go to machine check exception
	P1 = 0;
	P2 = pAddr;
	P3 = MCHK_WBP;
	ABORT(-SCB_MCHK);
}

// *******************************************************
// *******************************************************
// *******************************************************

#ifdef TEST_PARITY

// Test Parity
void ka630_TestParity(VAX_CPU *vax, uint32 pAddr)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	// This routine is stub for test #3
	if ((pAddr == ka630->cpuAddr) && (MSER & MSER_PEN)) {
#ifdef DEBUG
		dbg_Printf("PAR: Parity Error Trap on loc %08X at PC %08X\n",
			pAddr, faultPC);
#endif /* DEBUG */
		MSER |= 0x300 | MSER_LPE;
		if (ka630->cpuState++ < 1)
			CEAR = pAddr >> 9;
		else
			MSER |= MSER_LEB;
	
		// Go to machine check exception
		P1 = 0;
		P2 = pAddr;
		P3 = MCHK_RBP;
		ABORT(-SCB_MCHK);
	}
}

// Write Parity
void ka630_WriteParity(VAX_CPU *vax, uint32 pAddr)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (MSER & MSER_PEN) {
		if (MSER & MSER_WWP) {
			ka630->cpuAddr = pAddr;
#ifdef DEBUG
			dbg_Printf("PAR: Set Wrong Parity on loc %08X at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		} else if (ka630->cpuAddr == pAddr) {
			ka630->cpuAddr = -1;
#ifdef DEBUG
			dbg_Printf("PAR: Clear Wrong Parity on loc %08X at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		}
	}
}

#endif /* TEST_PARITY */

// Physical Read Aligned
uint32 ka630_ReadAligned(VAX_CPU *vax, int32 pAddr, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_ROM(pAddr)) {
		pAddr &= vax->maskROM;
		if (size >= LONG)
			return LROM(pAddr >> 2);
		else if (size == WORD)
			return WROM(pAddr >> 1);
		else
			return BROM(pAddr);
	} else if (IN_CPUREG(pAddr))
			return ka630_ReadCPU(ka630, pAddr, size);
	else if (IN_CLKREG(pAddr))
		return ka630_ReadClock(ka630, pAddr, size);
	else if (IN_Q22MAP(pAddr))
		return ka630_ReadMap(ka630, pAddr);
	else if (IN_Q22IO(pAddr))
		return ka630_ReadIO(ka630, pAddr, size);
	else if (IN_Q22MEM(pAddr))
		return ka630_ReadDataIO(ka630, pAddr, size);
	else {
		// Non-Existant Memory Exception
#ifdef DEBUG
//		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("KA630: (R) CPU Non-Existant Memory (%08X) at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		MSER |= MSER_NXM;

		// Go to machine check exception
		P1 = 0;
		P2 = pAddr;
		P3 = MCHK_RBP;
		ABORT(-SCB_MCHK);
	}

	return 0;
}

// Physical Write Aligned
void ka630_WriteAligned(VAX_CPU *vax, int32 pAddr, uint32 data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_CPUREG(pAddr))
		ka630_WriteCPU(ka630, pAddr, data, size);
	else if (IN_CLKREG(pAddr))
		ka630_WriteClock(ka630, pAddr, data, size);
	else if (IN_Q22MAP(pAddr))
		ka630_WriteMap(ka630, pAddr, data);
	else if (IN_Q22IO(pAddr))
		ka630_WriteIO(ka630, pAddr, data, size);
	else if (IN_Q22MEM(pAddr))
		ka630_WriteDataIO(ka630, pAddr, data, size);
	else {
		// Non-Existant Memory Exception
#ifdef DEBUG
//		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("KA630: (W) CPU Non-Existant Memory (%08X) at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		MSER |= MSER_NXM;

		// Go to machine check exception
		P1 = 0;
		P2 = pAddr;
		P3 = MCHK_WBP;
		ABORT(-SCB_MCHK);
	}
}

// Physical Read Unaligned
uint32 ka630_ReadUnaligned(VAX_CPU *vax, int32 pAddr, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_ROM(pAddr)) {
		pAddr &= vax->maskROM;
		if (size >= LONG)
			return LROMU(pAddr);
		else if (size == WORD)
			return WROMU(pAddr);
		else
			return BROMU(pAddr);
	} else if (IN_CPUREG(pAddr))
		return ka630_ReadCPU(ka630, pAddr, size);
	else if (IN_CLKREG(pAddr))
		return ka630_ReadClock(ka630, pAddr, size);
	else if (IN_Q22MAP(pAddr))
		return ka630_ReadMap(ka630, pAddr);
	else if (IN_Q22IO(pAddr))
		return ka630_ReadIO(ka630, pAddr, size);
	else if (IN_Q22MEM(pAddr))
		return ka630_ReadDataIO(ka630, pAddr, size);
	else { 
		// Non-Existant Memory Exception
#ifdef DEBUG
//		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("KA630: (RU) CPU Non-Existant Memory (%08X) at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		MSER |= MSER_NXM;

		// Go to machine check exception
		P1 = 0;
		P2 = pAddr;
		P3 = MCHK_RBP;
		ABORT(-SCB_MCHK);
	}

	return 0;
}

// Physical Write Unaligned
void ka630_WriteUnaligned(VAX_CPU *vax, int32 pAddr, uint32 data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_CPUREG(pAddr))
		ka630_WriteCPU(ka630, pAddr, data, size);
	else if (IN_CLKREG(pAddr))
		ka630_WriteClock(ka630, pAddr, data, size);
	else if (IN_Q22MAP(pAddr))
		ka630_WriteMap(ka630, pAddr, data);
	else if (IN_Q22IO(pAddr))
		ka630_WriteIO(ka630, pAddr, data, size);
	else if (IN_Q22MEM(pAddr))
		ka630_WriteDataIO(ka630, pAddr, data, size);
	else {
		// Non-Existant Memory Exception
#ifdef DEBUG
//		if (dbg_Check(DBG_IOREGS))
			dbg_Printf("KA630: (WU) CPU Non-Existant Memory (%08X) at PC %08X\n",
				pAddr, faultPC);
#endif /* DEBUG */
		MSER |= MSER_NXM;

		// Go to machine check exception
		P1 = 0;
		P2 = pAddr;
		P3 = MCHK_WBP;
		ABORT(-SCB_MCHK);
	}
}

// *******************************************************
// ************* Console Read/Write Access ***************
// *******************************************************

// Console Read Aligned
int ka630_ReadCA(VAX_CPU *vax, int32 pAddr, uint32 *data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_RAM(pAddr)) {
		if (size >= LONG)
			*data = LMEM(pAddr >> 2);
		else if (size == WORD)
			*data = WMEM(pAddr >> 1);
		else
			*data = BMEM(pAddr);
		return MM_OK;
	} else if (IN_ROM(pAddr)) {
		pAddr &= vax->maskROM;
		if (size >= LONG)
			*data = LROM(pAddr >> 2);
		else if (size == WORD)
			*data = WROM(pAddr >> 1);
		else
			*data = BROM(pAddr);
		return MM_OK;
	} else if (IN_CPUREG(pAddr))
		return ka630_ReadCPU(ka630, pAddr, size);
	else if (IN_CLKREG(pAddr))
		return ka630_ReadClock(ka630, pAddr, size);
	else if (IN_Q22MAP(pAddr))
		return ka630_ReadMap(ka630, pAddr);
	else if (IN_Q22IO(pAddr))
		return ka630_ReadIO(ka630, pAddr, size);
	else if (IN_Q22MEM(pAddr))
		return ka630_ReadDataIO(ka630, pAddr, size);

	return MM_NXM;
}

// Console Write Aligned
int ka630_WriteCA(register VAX_CPU *vax, int32 pAddr, uint32 data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_RAM(pAddr)) {
		if (size >= LONG)
			LMEM(pAddr >> 2) = data;
		else if (size == WORD)
			WMEM(pAddr >> 1) = data;
		else
			BMEM(pAddr) = data;
	} else if (IN_CPUREG(pAddr))
		ka630_WriteCPU(ka630, pAddr, data, size);
	else if (IN_CLKREG(pAddr))
		ka630_WriteClock(ka630, pAddr, data, size);
	else if (IN_Q22MAP(pAddr))
		ka630_WriteMap(ka630, pAddr, data);
	else if (IN_Q22IO(pAddr))
		ka630_WriteIO(ka630, pAddr, data, size);
	else if (IN_Q22MEM(pAddr))
		ka630_WriteDataIO(ka630, pAddr, data, size);
	else
		return MM_NXM;

	return MM_OK;
}

// Console Read Unaligned
int ka630_ReadCU(register VAX_CPU *vax, int32 pAddr, int32 *data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_RAM(pAddr)) {
		if (size >= LONG)
			*data = LMEMU(pAddr);
		else if (size == WORD)
			*data = WMEMU(pAddr);
		else
			*data = BMEMU(pAddr);
		return MM_OK;
	} else if (IN_ROM(pAddr)) {
		pAddr &= vax->maskROM;
		if (size >= LONG)
			*data = LROMU(pAddr);
		else if (size == WORD)
			*data = WROMU(pAddr);
		else
			*data = BROMU(pAddr);
		return MM_OK;
	} else if (IN_CPUREG(pAddr))
		return ka630_ReadCPU(ka630, pAddr, size);
	else if (IN_CLKREG(pAddr))
		return ka630_ReadClock(ka630, pAddr, size);
	else if (IN_Q22MAP(pAddr))
		return ka630_ReadMap(ka630, pAddr);
	else if (IN_Q22IO(pAddr))
		return ka630_ReadIO(ka630, pAddr, size);
	else if (IN_Q22MEM(pAddr))
		return ka630_ReadDataIO(ka630, pAddr, size);

	return MM_NXM;
}

// Console Write Unaligned
int ka630_WriteCU(register VAX_CPU *vax, int32 pAddr, uint32 data, int32 size)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)vax->Processor;

	if (IN_RAM(pAddr)) {
		if (size >= LONG)
			LMEMU(pAddr) = data;
		else if (size == WORD)
			WMEMU(pAddr) = data;
		else
			BMEMU(pAddr) = data;
	} else if (IN_CPUREG(pAddr))
		ka630_WriteCPU(ka630, pAddr, data, size);
	else if (IN_CLKREG(pAddr))
		ka630_WriteClock(ka630, pAddr, data, size);
	else if (IN_Q22MAP(pAddr))
		ka630_WriteMap(ka630, pAddr, data);
	else if (IN_Q22IO(pAddr))
		ka630_WriteIO(ka630, pAddr, data, size);
	else if (IN_Q22MEM(pAddr))
		ka630_WriteDataIO(ka630, pAddr, data, size);
	else
		return MM_NXM;

	return MM_OK;
}

// **********************************************************
// ************ MicroVAX II System Configuration ************
// **********************************************************

DEVINFO  vax_System_KA630;
QBA_CALL ka630_Callback;

void ka630_Build(KA630_DEVICE *ka630)
{
	VAX_CPU *cpu = ka630->Processor;
	int32   idxInst, idxOpnd;

	// Clear all instruction tables
	for (idxInst = 0; idxInst < NUM_INST; idxInst++)
		cpu->tblOpcode[idxInst] = vax_Opcode_Illegal;
	memset(&cpu->tblOperand, 0, sizeof(uint32) * (NUM_INST * (MAX_SPEC+1)));

	// Build instruction table
	for (idxInst = 0; vax_Instruction[idxInst].Name; idxInst++) {
		void   (*Execute)() = vax_Instruction[idxInst].Execute;
		uint32 opExtend     = vax_Instruction[idxInst].Extended;
		uint32 opCode       = vax_Instruction[idxInst].Opcode;
		int32  nOpnds       = vax_Instruction[idxInst].nOperands;

		if (opExtend >= INST_EXTEND)
			opCode |= ((opExtend - (INST_EXTEND - 1)) << 8);

		// Build opcode entry
		if (Execute == NULL)
			cpu->tblOpcode[opCode] = vax_Opcode_Unimplemented;
		else if (Execute == EMULATE)
			cpu->tblOpcode[opCode] = vax_Opcode_Emulate;
		else
			cpu->tblOpcode[opCode] = Execute;

		// Build operand entry
		cpu->tblOperand[opCode][0] = nOpnds;
		for (idxOpnd = 0; idxOpnd < nOpnds; idxOpnd++)
			cpu->tblOperand[opCode][idxOpnd+1] =
				vax_Instruction[idxInst].opMode[idxOpnd];
	}
}

void *ka630_Create(MAP_DEVICE *nmap, MAP_DEVICE *map, int argc, char **argv)
{
	KA630_DEVICE *ka630;
	DEVINFO      *iptr, *cptr;
	int          idx;

	if (nmap == NULL) {
		// Set up KA630 Processor
		if (ka630 = (KA630_DEVICE *)calloc(1, sizeof(KA630_DEVICE))) {
			ka630->devName  = map->devName;
			ka630->dtName   = KA630_DTNAME;

			map->dtName     = KA630_DTNAME;
			map->emuName    = KA630_NAME;
			map->emuVersion = KA630_VERSION;
			map->devInfo    = &vax_System_KA630;
			map->Device     = ka630;

			printf("Created %s on %s: - %s %s\n",
				map->dtName, map->devName, map->emuName, map->emuVersion);
		}

		return ka630;
	} else {
		ka630 = (KA630_DEVICE *)map->Device;
		iptr  = map->devInfo;
		cptr  = NULL;

		for (idx = 0; cptr = iptr->Devices[idx]; idx++)
			if (!strcasecmp(argv[2], cptr->dtName))
				break;

		if (cptr == NULL) {
			printf("%s: No Such Device on %s:\n",
				argv[2], map->devName);
			return NULL;
		}

		ka630->qba = cptr->Create(NULL, nmap, argc, argv);
		cptr->Setup(ka630->qba, ka630, &ka630_Callback);
		nmap->sysDevice = map->sysDevice;
		nmap->sysMap    = map->sysMap;

		return ka630->qba;
	}
}

// Delete CPU Device
int ka630_Delete(void *dptr)
{
}

int ka630_Reset(void *dptr)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	VAX_CPU      *vax   = ka630->Processor;
	int32        idx;

#ifdef TEST_PARITY
	ka630->cpuState = ka630->q22State = 0;
	ka630->cpuAddr  = -1;
	ka630->q22Addr1 = ka630->q22Addr2 = -1;
#endif /* TEST_PARITY */

	// Show my identification that I am MicroVAX II system.

	SID = KA630_SYSID;

	// Clear all CPU and clock registers first
	for (idx = 0; idx < nCPUREG; idx++)
		ka630->cpuRegs[idx] = 0;
	for (idx = 0; idx < nCLKREG; idx++)
		ka630->clkRegs[idx] = 0;
	for (idx = 0; idx < 8096; idx++)
		ka630->mapRegs[idx] = 0;
}

int ka630_Setup(void *dptr, VAX_CPU *vax)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)dptr;
	CLK_QUEUE    *Clock;
	
	ka630->Processor = vax;
	ka630_Reset(ka630);

	// Build instruction table
	ka630_Build(ka630);

	// Reload current time/date into clock registers
	// Load saved NVRAM data into TOY chip.
	// ACTION: Implement that.

	// Set up clock timer
	Clock = &ka630->Timer;
	Clock->Next     = NULL;
	Clock->Flags    = CLK_REACTIVE;
	Clock->outTimer = ICCS_TICK;
	Clock->nxtTimer = ICCS_TICK;
	Clock->Device   = (void *)ka630;
	Clock->Execute  = ka630_ClockTimer;

	// Now enable clock timer (on host timer)
	INTVEC[nICCS_INT] = SCB_TIMER;
	ts10_SetRealTimer(Clock);

	INTVEC[nIPCR_INT] = IPCR_VEC;

	// Set up functions for KA630 CPU (MicroVAX II Series)
	vax->ReadRegister   = ka630_ReadRegister;
	vax->WriteRegister  = ka630_WriteRegister;
	vax->ReadAligned    = ka630_ReadAligned;
	vax->WriteAligned   = ka630_WriteAligned;
	vax->ReadUnaligned  = ka630_ReadUnaligned;
	vax->WriteUnaligned = ka630_WriteUnaligned;
	vax->ReadCA         = ka630_ReadCA;
	vax->ReadCU         = ka630_ReadCU;
	vax->WriteCA        = ka630_WriteCA;
	vax->WriteCU        = ka630_WriteCU;
	vax->TestParity     = ka630_TestParity;
	vax->WriteParity    = ka630_WriteParity;
	vax->MachineCheck   = ka630_MachineCheck;
	vax->HaltAction     = ka630_HaltAction;
	vax->CheckTimer     = ka630_CheckTimer;
	vax->StartTimer     = ka630_StartTimer;
	vax->StopTimer      = ka630_StopTimer;

	return VAX_OK;
}

int ka630_Boot(MAP_DEVICE *map, int argc, char **argv)
{
	KA630_DEVICE *ka630 = (KA630_DEVICE *)map->Device;
	VAX_CPU      *vax   = ka630->Processor;

	// Tell operator that processor is being booted.
	printf("Booting %s: (%s)... ", ka630->devName, ka630->dtName);

	// Initialize KA630 Processor
	vax->State = VAX_HALT;
	ka630_Reset(ka630);

	// Set BDR register as power-up state
	BDR = BDR_PWROK | BDR_HLTENB | 0xF;

	// Must load 0x300 (Restart Code - Power On) into SAVPSL in order that
	// the console program perform properly.

	// Initial Power On
	MAPEN  = 0;
	SAVPSL = HLT_PON << 8;
	PC     = 0x20040000;
	PSL    = 0x041F0000;
	CC     = 0;
	SISR   = 0; 
	ASTLVL = 4;

	// Enable KA630 processor to run.
	vax->State = VAX_RUN;
	emu_State  = VAX_RUN;

	// Tell operator that it was done.
	printf("done.\n");
}

int qba_SetMap(void *, MAP_IO *);

// Callback Functions for Q22-Bus interface device
QBA_CALL ka630_Callback =
{
	ka630_ReadDataIO,     // Read Data I/O
	ka630_WriteDataIO,    // Write Data I/O
	ka630_ReadBlockIO,    // Read Block I/O
	ka630_WriteBlockIO,   // Write Block I/O
	ka630_GetHostAddr,    // Get Host Address

	qba_SetMap,           // SetMap Routine
};

DEVINFO *ka630_Devices[] =
{
	&qba_Device, // Q22-Bus Interface
	NULL         // Terminator
};

// MicroVAX II Device Information Table
DEVINFO vax_System_KA630 =
{
	KA630_DTNAME,   // Device Type Name
	KA630_NAME,     // Emulator Name
	KA630_VERSION,  // Emulator Version
	ka630_Devices,  // Listing of Devices
	DF_USE,         // Device Flags
	DT_PROCESSOR,   // Device Type

	NULL, NULL, NULL,

	ka630_Create,   // Create Routine
	NULL,           // Delete Routine
	ka630_Setup,    // Setup Routine
	ka630_Reset,    // Reset Routine
	NULL,           // Attach Routine
	NULL,           // Detach Routine
	NULL,           // Info Routine
	ka630_Boot,     // Boot Routine
	NULL,           // Execute Routine
	NULL,           // Debug Routine
};

