// ka650.c - CVAX Series System Configurations and Routines
//
// Copyright (c) 2001, Timothy M. Stark
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// TIMOTHY M STARK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Timothy M Stark shall not
// be used in advertising or otherwise to promote the sale, use or other 
// dealings in this Software without prior written authorization from
// Timothy M Stark.

// CVAX Series - KA650/KA655 Processors
//   KA655 Processor - MicroVAX 3800/3900

#include "vax/defs.h"

#define KA650_DTNAME  "KA650" // Device Type Name
#define KA650_NAME    "KA650 (CVAX Series) Emulator"

#define KA655_DTNAME  "KA655" // Device Type Name
#define KA655_NAME    "MicroVAX 3800/3900 Emulator"

#define KA65X_VERSION "v0.1 (Pre-Alpha)"

// KA650/KA655 - CVAX Memory Map
//
// 00000000 +-------------------------------------+
//          |           Main Memory               |
//          |- - - - - - - - - - - - - - - - - - -|
//          |        Up to 64 MB Memory           |
// 01000000 +-------------------------------------+
//          |             Reserved                |
// 20000000 +-------------------------------------+
//          |         Q22-Bus I/O space           |
// 20002000 +-------------------------------------+
//          |             Reserved                |
// 20040000 +-------------------------------------+
//          |               ROM                   |
// 20080000 +-------------------------------------+
//          |      Local I/O Register Area        |
// 200C0000 +-------------------------------------+
//          |             Reserved                |
// 30000000 +-------------------------------------+
//          |        Q22-Bus Memory Space         |
// 303FFFFF +-------------------------------------+
//          |             Reserved                |
// 3FFFFFFF +-------------------------------------+
//
// Local I/O Register Area:
//
//   CPU Registers     - 20080000-2008000F (4 16-bit registers)
//   Clock Registers   - 200B8000-2008007F (64 8-bit registers)
//
//   Q22 I/O Page      - 20000000-20001FFF (Last page in memory space)
//   Q22 Map Registers - 20088000-2008FFFF (8096 32-bit registers)
//   Q22 Memory Space  - 30000000-3003FFFF

typedef struct {
	char *sysName; // System Name
} KA650_DEVICE;

#ifdef DEBUG
static int  regLength = 0x40;
static char *regNames[] = {
	"KSP",     // (R/W) 00 Kernel Stack Pointer
	"ESP",     // (R/W) 01 Executive Stack Pointer
	"SSP",     // (R/W) 02 Supervisor Stack Pointer
	"USP",     // (R/W) 03 User Stack Pointer
	"ISP",     // (R/W) 04 Interrupt Stack Pointer
	NULL,      //       05 Unknown
	NULL,      //       06 Unknown
	NULL,      //       07 Unknown
	"P0BR",    // (R/W) 08 P0 Base Register
	"P0LR",    // (R/W) 09 P0 Length Register
	"P1BR",    // (R/W) 0A P1 Base Register
	"P1LR",    // (R/W) 0B P1 Length Register
	"SBR",     // (R/W) 0C System Base Register
	"SLR",     // (R/W) 0D System Length Register
	NULL,      //       0E Unknown
	NULL,      //       0F Unknown
	"PCBB",    // (R/W) 10 Process Control Block Base 
	"SCBB",    // (R/W) 11 System Control Block Base
	"IPL",     // (R/W) 12 Interrupt Priority Level
	"ASTLVL",  // (R/W) 13 AST Level
	"SIRR",    // (W)   14 Software Interrupt Request
	"SISR",    // (R/W) 15 Software Interrupt Summary
	NULL,      //       16 Unknown
	NULL,      //       17 Unknown
	"ICCS",    // (R/W) 18 Interval Clock Control
	"NICR",    // (W)   19 Next Interval Count
	"ICR",     // (R)   1A Interval Count
	"TODR",    // (R/W) 1B Time of Year
	NULL,      //       1C Unknown
	NULL,      //       1D Unknown
	NULL,      //       1E Unknown
	NULL,      //       1F Unknown
	"RXCS",    // (R/W) 20 Console Receiver Status
	"RXDB",    // (R)   21 Console Receiver Data Buffer
	"TXCS",    // (R/W) 22 Console Transmit Status
	"TXDB",    // (W)   23 Console Transmit Data Buffer
	NULL,      //       24 Unknown
	NULL,      //       25 Unknown
	NULL,      //       26 Unknown
	NULL,      //       27 Unknown
	"ACCS",    //       28 *Accelerator Control and Status
	"ACCR",    //       29 *Accelerator Maintenance
	NULL,      //       2A Unknown
	NULL,      //       2B Unknown
	"WCSA",    //       2C *Writable-Control-Store Address
	"WCSD",    //       2D *Writable-Control-Store Data
	NULL,      //       2E Unknown
	NULL,      //       2F Unknown
	"SBIFS",   //       30 *SBI Fault Status
	"SBIS",    //       31 *SBI silo
	"SBISC",   //       32 *SBI silo comparator
	"SBIMT",   //       33 *SBI Maintenance
	"SBIER",   //       34 *SBI Error
	"SBITA",   //       35 *SBI Timeout Address
	"SBIQC",   //       36 *SBI Quadword Clear
	NULL,      //       37 Unknown
	"MAPEN",   // (R/W) 38 Memory Management Enable
	"TBIA",    // (W)   39 Translation Buffer Invalidate All
	"TBIS",    // (W)   3A Translation Buffer Invalidate Single
	NULL,      //       3B Unknown
	"MBRK",    //       3C Microprogram Breakpoint
	"PME",     // (R/W) 3D Performance Monitor Enable
	"SID",     // (R)   3E System Identification
	"TBCHK",   // (W)   3F Translation Buffer Check
};
#endif /* DEBUG */

// System ID Longword
//
//  31   24 23 22          15 14   12 11             0
// +-------+--+--------------+-------+----------------+
// |   A   |                                          |
// +-------+--+--------------+-------+----------------+

#define KA650_SYSID    0x0A000000

// Privileged Register for CVAX Series System
#define nKSP    0x00 // Kernel Stack Pointer
#define nESP    0x01 // Executive Stack Pointer
#define nSSP    0x02 // Supervisor Stack Pointer
#define nUSP    0x03 // User Stack Pointer
#define nISP    0x04 // Interrupt Stack Pointer

#define nP0BR   0x08 // P0 Base Register
#define nP0LR   0x09 // P0 Length Register
#define nP1BR   0x0A // P1 Base Register
#define nP1LR   0x0B // P1 Length Register
#define nSBR    0x0C // System Base Register
#define nSLR    0x0D // System Length Register

#define nPCBB   0x10 // Process Control Block Base
#define nSCBB   0x11 // System Control Block Base
#define nIPL    0x12 // Interrupt Priority Level
#define nASTLVL 0x13 // AST Level
#define nSIRR   0x14 // Software Interrupt Request
#define nSISR   0x15 // Software Interrupt Summary

#define nRXCS   0x20 // Console Receive Control and Status Register
#define nRXDB   0x21 // Console Receive Data Buffer Register
#define nTXCS   0x22 // Console Transmit Control and Status Register
#define nTXDB   0x23 // Console Transmit Data Buffer Register

#define nMAPEN  0x38 // Map Enable
#define nTBIA   0x39 // TB Clear All Entries
#define nTBIS   0x3A // TB Clear Single Entry
#define nSID    0x3E // System Identification
#define nTBCHK  0x3F // TB Check

#define SCBB_MBZ   0x000001FF // Must be zeros.
#define BR_MASK    0xFFFFFFFC // xxBR Mask
#define LR_MASK    0x003FFFFF // xxLR Mask

// Console Terminal Registers

// RXCS - Console Receive Control and Status Register
#define RXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define RXCS_RDY   0x00000080 // (R)   Ready
#define RXCS_IE    0x00000040 // (R/W) Interrupt Enable

// RXDB - Console Receive Data Buffer Register
#define RXDB_ERROR 0x00008000 // (R)   Error
#define RXDB_ID    0x00000F00 // (R)   Identification
#define RXDB_DATA  0x000000FF // (R)   Data Received

// TXCS - Console Terminal Control and Status Register
#define TXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define TXCS_RDY   0x00000080 // (R)   Ready
#define TXCS_IE    0x00000040 // (R/W) Interrupt Enable

// TXDB - Console Terminal Data Buffer Register
#define TXDB_MBZ   0xFFFFF000 // Must be zeros
#define TXDB_ID    0x00000F00 // (R/W) Identification
#define TXDB_DATA  0x000000FF // (R/W) Data Transmitted

// Read Privileged Register
int32 ka650_ReadRegister(int32 pReg)
{
	int32 data;

	switch (pReg) {
		case nKSP:
			data = (PSL & PSL_IS) ? KSP : SP;
			break;

		case nISP:
			data = (PSL & PSL_IS) ? SP : ISP;
			break;

		case nIPL:
			data = PSB.ipl;
			break;

		case nRXCS:
			data = vax_ReadRXCS(vax->Console);
			break;

		case nRXDB:
			data = vax_ReadRXDB(vax->Console);
			break;

		case nTXCS:
			data = vax_ReadTXCS(vax->Console);
			break;

//		case nTXDB:
//			data = vax_ReadTXDB(vax->Console);
//			break;

		default:
			data = PRN(pReg);
			break;

/*
#ifdef DEBUG
			if (dbg_Check(DBG_TRACE|DBG_DATA)) {
				dbg_Printf("KA650: (R) Unknown privileged register: %08X (%d)\n",
					pReg, pReg);
			}
#endif // DEBUG
			RSVD_OPND_FAULT;
*/
	}

#ifdef DEBUG
//	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
	{
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA650: (R) %s (%02X) => %08X\n", name, pReg, data);
	}
#endif /* DEBUG */

	return data;
}

void ka650_WriteRegister(int32 pReg, int32 data)
{
#ifdef DEBUG
//	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
	{
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA650: (W) %s (%02X) <= %08X\n", name, pReg, data);
	}
#endif /* DEBUG */

	switch (pReg) {
		case nKSP: // Kernel Stack Pointer
			((PSL & PSL_IS) ? KSP : SP) = data;
			break;

		case nESP: // Executive Stack Pointer
		case nSSP: // Supervisor Stack Pointer
		case nUSP: // User Stack Pointer
			PRN(pReg) = data;
			break;

		case nISP: // Interrupt Stack Pointer
			((PSL & PSL_IS) ? SP : ISP) = data;
			break;

		case nP0BR:
		case nP1BR:
		case nSBR:
			PRN(pReg) = data & BR_MASK;
			vax_ClearTBTable(pReg == nSBR);
			break;

		case nP0LR:
		case nP1LR:
		case nSLR:
			PRN(pReg) = data & LR_MASK;
			vax_ClearTBTable(pReg == nSLR);
			break;
		
		case nPCBB:
		case nSCBB:
			PRN(pReg) = data & LALIGN;
			break;

		case nIPL:
			PSB.ipl = data;
//			IPL = data & PSL_M_IPL;
			break;

		case nSIRR:
			if ((data > 0xF) || (data == 0))
				RSVD_OPND_FAULT;
			SISR |= (1 << data);
			break;

		case nSISR:
			SISR = data & SISR_MASK;
			break;

		case nASTLVL:
			if ((uint32)data > AST_MAX)
				RSVD_OPND_FAULT;
			ASTLVL = data;
			break;

		case nRXCS:
			if (data & RXCS_MBZ)
				RSVD_OPND_FAULT;
			vax_WriteRXCS(vax->Console, data);
			break;

//		case nRXDB:
//			vax_WriteRXDB(vax->Console, data);
//			break;

		case nTXCS:
			if (data & TXCS_MBZ)
				RSVD_OPND_FAULT;
			vax_WriteTXCS(vax->Console, data);
			break;

		case nTXDB:
			vax_WriteTXDB(vax->Console, data);
			break;

		case nMAPEN:
			MAPEN = data & 1;
			break;

		case nTBIA:
			vax_ClearTBTable(0);
			break;

		case nTBIS:
			vax_ClearTBEntry(data);
			break;

		case nTBCHK:
			if (vax_CheckTBEntry(data))
				PSL |= CC_V;
			break;

		default:
#ifdef DEBUG
			if (dbg_Check(DBG_TRACE|DBG_DATA)) {
				dbg_Printf("KA650: (W) Unknown privileged register: %08X (%d)\n",
					pReg, pReg);
			}
#endif /* DEBUG */
//			RSVD_OPND_FAULT;
	}
}

uint32 ka650_ReadIO(void *device, int32 pAddr, int32 size)
{
	KA650_DEVICE *ka650 = (KA650_DEVICE *)device;

	if ((pAddr >= 0x20000000) && (pAddr < 0x20020000)) {
		// Nexus Adaptor
	}

	if ((pAddr >= 0x20100000) && (pAddr < 0x20200000)) {
		// Unibus Address
	}

	return 0;
}

void ka650_WriteIO(void *device, int32 pAddr, uint32 data, int32 size)
{
	KA650_DEVICE *ka650 = (KA650_DEVICE *)device;

	if ((pAddr >= 0x20000000) && (pAddr < 0x20020000)) {
		// Nexus Adaptor
	}

	if ((pAddr >= 0x20100000) && (pAddr < 0x20200000)) {
		// Unibus Address
	}
}

// **************************************************************
// ************** CVAX Series System Configuration **************
// **************************************************************

// Create Desired System
void *ka650_Create(void)
{
	KA650_DEVICE *ka650;

	ka650 = (KA650_DEVICE *)calloc(1, sizeof(KA650_DEVICE));

	return (void *)ka650;
}

void ka650_Cleanup(VAX_CPU *vax)
{
}

void ka650_Initialize(VAX_CPU *vax)
{
	TXCS = TXCS_RDY;
	SID  = KA650_SYSID;

	// Set up console TTY device (OPA0:)
	vax->Console = vax_ConsoleInit(vax);

	// Set up functions for KA650 CPU (CVAX Series)
	vax->ReadRegister   = ka650_ReadRegister;
	vax->WriteRegister  = ka650_WriteRegister;
	vax->ReadAligned    = ka650_ReadIO;
	vax->WriteAligned   = ka650_WriteIO;
	vax->ReadUnaligned  = ka650_ReadIO;
	vax->WriteUnaligned = ka650_WriteIO;
}

void ka650_Reset(VAX_CPU *vax)
{
}

// CVAX Series System Definition
DEVINFO vax_System_KA650 =
{
	// Description Information
	KA650_DTNAME,          // Device Type Name
	KA650_NAME,            // Emulator Name
	KA65X_VERSION,         // Emulator Version

	ka650_Devices,         // KA650 Device List
	DT_USE,                // Device Flags
	DT_PROCESSOR,          // Device Type

	NULL, NULL, NULL,      // Commands

	ka650_Create,          // Create Routine
	NULL,                  // Delete Routine
	NULL,                  // Setup Routine
	NULL,                  // Reset Routine
	NULL,                  // Attach Routine
	NULL,                  // Detach Routine
	NULL,                  // Info Routine
	NULL,                  // Boot Routine
	NULL,                  // Execute Routine
	NULL,                  // Debug Routine
};

DEVINFO vax_System_KA655 =
{
	// Description Information
	KA655_DTNAME,          // Device Type Name
	KA655_NAME,            // Emulator Name
	KA65X_VERSION,         // Emulator Version

	ka650_Devices,         // KA655 Device List
	DT_USE,                // Device Flags
	DT_PROCESSOR,          // Device Type

	NULL, NULL, NULL,      // Commands

	ka650_Create,          // Create Routine
	NULL,                  // Delete Routine
	NULL,                  // Setup Routine
	NULL,                  // Reset Routine
	NULL,                  // Attach Routine
	NULL,                  // Detach Routine
	NULL,                  // Info Routine
	NULL,                  // Boot Routine
	NULL,                  // Execute Routine
	NULL,                  // Debug Routine
};
