// ka780.c - VAX-11/780 System Configurations and Routines
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS-10 Emulator.
// See README for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "vax/defs.h"
#include "vax/nexus.h"

#define KA780_KEY     "KA780"
#define KA780_NAME    "VAX-11/780 Emulator"
#define KA780_VERSION "v0.1 (Pre-Alpha)"

// VAX-11/780 Memory Map
//
//            +-----------------------------------------------------+
// 0000 0000: |                  Installed Memory                   |
//     :      |- - - - - - - - - - - - - - - - - - - - - - - - - - -|
// 1FFF FFFF: |                 Up to 512MB Memory                  |
//            +-----------------------------------------------------+   
// 2000 0000: |                  TR0 Adapter Space                  |
//            +-----------------------------------------------------+
// 2000 2000: |                  TR1 Adapter Space                  |
//            |                          :                          |
// 2001 E000: |                  TR15 Adapter Space                 |
//            +-----------------------------------------------------+
// 2002 0000: |                                                     |
//     :      |                      Reserved                       |
// 200F FFFF: |                                                     |
//            +-----------------------------------------------------+
// 2010 0000: |                UNIBUS 0 Address Space               |
//            +-----------------------------------------------------+
// 2014 0000: |                UNIBUS 1 Address Space               |
//            +-----------------------------------------------------+
// 2018 0000: |                UNIBUS 2 Address Space               |
//            +-----------------------------------------------------+
// 201C 0000: |                UNIBUS 3 Address Space               |
//            +-----------------------------------------------------+
// 2020 0000: |                                                     |
//     :      |                      Reserved                       |
// 3FFF FFFF: |                                                     |
//            +-----------------------------------------------------+

typedef struct {
	char *devName; // Device Name (SBIx:)
} SBI780_DEVICE;

typedef struct {
	char *sysName; // System Name
	SBI780_DEVICE Adaptor[NEX780_NSBI];
} KA780_DEVICE;

#ifdef DEBUG
static int  regLength = 0x40;
static char *regNames[] = {
	"KSP",     // (R/W) 00 Kernel Stack Pointer
	"ESP",     // (R/W) 01 Executive Stack Pointer
	"SSP",     // (R/W) 02 Supervisor Stack Pointer
	"USP",     // (R/W) 03 User Stack Pointer
	"ISP",     // (R/W) 04 Interrupt Stack Pointer
	NULL,      //       05 Unknown
	NULL,      //       06 Unknown
	NULL,      //       07 Unknown
	"P0BR",    // (R/W) 08 P0 Base Register
	"P0LR",    // (R/W) 09 P0 Length Register
	"P1BR",    // (R/W) 0A P1 Base Register
	"P1LR",    // (R/W) 0B P1 Length Register
	"SBR",     // (R/W) 0C System Base Register
	"SLR",     // (R/W) 0D System Length Register
	NULL,      //       0E Unknown
	NULL,      //       0F Unknown
	"PCBB",    // (R/W) 10 Process Control Block Base 
	"SCBB",    // (R/W) 11 System Control Block Base
	"IPL",     // (R/W) 12 Interrupt Priority Level
	"ASTLVL",  // (R/W) 13 AST Level
	"SIRR",    // (W)   14 Software Interrupt Request
	"SISR",    // (R/W) 15 Software Interrupt Summary
	NULL,      //       16 Unknown
	NULL,      //       17 Unknown
	"ICCS",    // (R/W) 18 Interval Clock Control
	"NICR",    // (W)   19 Next Interval Count
	"ICR",     // (R)   1A Interval Count
	"TODR",    // (R/W) 1B Time of Year
	NULL,      //       1C Unknown
	NULL,      //       1D Unknown
	NULL,      //       1E Unknown
	NULL,      //       1F Unknown
	"RXCS",    // (R/W) 20 Console Receiver Status
	"RXDB",    // (R)   21 Console Receiver Data Buffer
	"TXCS",    // (R/W) 22 Console Transmit Status
	"TXDB",    // (W)   23 Console Transmit Data Buffer
	NULL,      //       24 Unknown
	NULL,      //       25 Unknown
	NULL,      //       26 Unknown
	NULL,      //       27 Unknown
	"ACCS",    //       28 *Accelerator Control and Status
	"ACCR",    //       29 *Accelerator Maintenance
	NULL,      //       2A Unknown
	NULL,      //       2B Unknown
	"WCSA",    //       2C *Writable-Control-Store Address
	"WCSD",    //       2D *Writable-Control-Store Data
	NULL,      //       2E Unknown
	NULL,      //       2F Unknown
	"SBIFS",   //       30 *SBI Fault Status
	"SBIS",    //       31 *SBI silo
	"SBISC",   //       32 *SBI silo comparator
	"SBIMT",   //       33 *SBI Maintenance
	"SBIER",   //       34 *SBI Error
	"SBITA",   //       35 *SBI Timeout Address
	"SBIQC",   //       36 *SBI Quadword Clear
	NULL,      //       37 Unknown
	"MAPEN",   // (R/W) 38 Memory Management Enable
	"TBIA",    // (W)   39 Translation Buffer Invalidate All
	"TBIS",    // (W)   3A Translation Buffer Invalidate Single
	NULL,      //       3B Unknown
	"MBRK",    //       3C Microprogram Breakpoint
	"PME",     // (R/W) 3D Performance Monitor Enable
	"SID",     // (R)   3E System Identification
	"TBCHK",   // (W)   3F Translation Buffer Check
};
#endif /* DEBUG */

// System ID Longword
//
//  31   24 23 22          15 14   12 11             0
// +-------+--+--------------+-------+----------------+
// |   1   |  |   ECO Level  | Plant |  Serial Number |
// +-------+--+--------------+-------+----------------+
//          |
//          +- 0 = VAX-11/780
//             1 = VAX-11/785

#define KA780_ECO      0
#define KA780_PLANT    0
#define KA780_SN       1

//#define KA780_SYSID    0x0A000000
#define KA780_SYSID    0x01000000
#define KA780_785      0x00800000

#define KA780_P_ECO    15
#define KA780_P_PLANT  12

// Privileged Register for VAX-11/780 system
#define nKSP    0x00 // Kernel Stack Pointer
#define nESP    0x01 // Executive Stack Pointer
#define nSSP    0x02 // Supervisor Stack Pointer
#define nUSP    0x03 // User Stack Pointer
#define nISP    0x04 // Interrupt Stack Pointer

#define nP0BR   0x08 // P0 Base Register
#define nP0LR   0x09 // P0 Length Register
#define nP1BR   0x0A // P1 Base Register
#define nP1LR   0x0B // P1 Length Register
#define nSBR    0x0C // System Base Register
#define nSLR    0x0D // System Length Register

#define nPCBB   0x10 // Process Control Block Base
#define nSCBB   0x11 // System Control Block Base
#define nIPL    0x12 // Interrupt Priority Level
#define nASTLVL 0x13 // AST Level
#define nSIRR   0x14 // Software Interrupt Request
#define nSISR   0x15 // Software Interrupt Summary

#define nRXCS   0x20 // Console Receive Control and Status Register
#define nRXDB   0x21 // Console Receive Data Buffer Register
#define nTXCS   0x22 // Console Transmit Control and Status Register
#define nTXDB   0x23 // Console Transmit Data Buffer Register

#define nMAPEN  0x38 // Map Enable
#define nTBIA   0x39 // TB Clear All Entries
#define nTBIS   0x3A // TB Clear Single Entry
#define nSID    0x3E // System Identification
#define nTBCHK  0x3F // TB Check

#define SCBB_MBZ   0x000001FF // Must be zeros.
#define BR_MASK    0xFFFFFFFC // xxBR Mask
#define LR_MASK    0x003FFFFF // xxLR Mask

// Interval Clock Control Status Registers

#define ICCS_ERR   0x80000000  // Timer Error
#define ICCS_DONE  0x00000080  // Ready/Done (Interrupt)
#define ICCS_IE    0x00000040  // Interrupt Enable
#define ICCS_STEP  0x00000020  // Single Step
#define ICCS_XFER  0x00000010  // Transfer
#define ICCS_RUN   0x00000001  // Run

#define ICCS_CMASK 0x80000001  // Clear Mask
#define ICCS_WMASK 0x00000090  // Write Mask
#define ICCS_MBZ   0x7FFFFF0E  // Must Be Zeros

// Console Terminal Registers

// RXCS - Console Receive Control and Status Register
#define RXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define RXCS_RDY   0x00000080 // (R)   Ready
#define RXCS_IE    0x00000040 // (R/W) Interrupt Enable

// RXDB - Console Receive Data Buffer Register
#define RXDB_ERROR 0x00008000 // (R)   Error
#define RXDB_ID    0x00000F00 // (R)   Identification
#define RXDB_DATA  0x000000FF // (R)   Data Received

// TXCS - Console Terminal Control and Status Register
#define TXCS_MBZ   0xFFFFFF3F // Must be zeros.
#define TXCS_RDY   0x00000080 // (R)   Ready
#define TXCS_IE    0x00000040 // (R/W) Interrupt Enable

// TXDB - Console Terminal Data Buffer Register
#define TXDB_MBZ   0xFFFFF000 // Must be zeros
#define TXDB_ID    0x00000F00 // (R/W) Identification
#define TXDB_DATA  0x000000FF // (R/W) Data Transmitted

// Read Privileged Register
int32 ka780_ReadRegister(int32 pReg)
{
	int32 data;

	switch (pReg) {
		case nKSP:
			data = (PSL & PSL_IS) ? KSP : SP;
			break;

		case nISP:
			data = (PSL & PSL_IS) ? SP : ISP;
			break;

		case nIPL:
			data = PSB.ipl;
			break;

		case nRXCS:
			data = vax_ReadRXCS(vax->Console);
			break;

		case nRXDB:
			data = vax_ReadRXDB(vax->Console);
			break;

		case nTXCS:
			data = vax_ReadTXCS(vax->Console);
			break;

//		case nTXDB:
//			data = vax_ReadTXDB(vax->Console);
//			break;

		default:
			data = PRN(pReg);
			break;

/*
#ifdef DEBUG
			if (dbg_Check(DBG_TRACE|DBG_DATA)) {
				dbg_Printf("KA780: (R) Unknown privileged register: %08X (%d)\n",
					pReg, pReg);
			}
#endif // DEBUG
			RSVD_OPND_FAULT;
*/
	}

#ifdef DEBUG
//	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
	{
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA780: (R) %s (%02X) => %08X\n", name, pReg, data);
	}
#endif /* DEBUG */

	return data;
}

void ka780_WriteRegister(int32 pReg, int32 data)
{
#ifdef DEBUG
//	if (dbg_Check(DBG_TRACE|DBG_DATA)) {
	{
		char *name = "Undefined Register";
		if ((pReg < regLength) && regNames[pReg])
			name = regNames[pReg];
		dbg_Printf("KA780: (W) %s (%02X) <= %08X\n", name, pReg, data);
	}
#endif /* DEBUG */

	switch (pReg) {
		case nKSP: // Kernel Stack Pointer
			((PSL & PSL_IS) ? KSP : SP) = data;
			break;

		case nESP: // Executive Stack Pointer
		case nSSP: // Supervisor Stack Pointer
		case nUSP: // User Stack Pointer
			PRN(pReg) = data;
			break;

		case nISP: // Interrupt Stack Pointer
			((PSL & PSL_IS) ? SP : ISP) = data;
			break;

		case nP0BR:
		case nP1BR:
		case nSBR:
			PRN(pReg) = data & BR_MASK;
			vax_ClearTBTable(pReg == nSBR);
			break;

		case nP0LR:
		case nP1LR:
		case nSLR:
			PRN(pReg) = data & LR_MASK;
			vax_ClearTBTable(pReg == nSLR);
			break;
		
		case nPCBB:
		case nSCBB:
			PRN(pReg) = data & LALIGN;
			break;

		case nIPL:
			PSB.ipl = data;
//			IPL = data & PSL_M_IPL;
			break;

		case nSIRR:
			if ((data > 0xF) || (data == 0))
				RSVD_OPND_FAULT;
			SISR |= (1 << data);
			break;

		case nSISR:
			SISR = data & SISR_MASK;
			break;

		case nASTLVL:
			if ((uint32)data > AST_MAX)
				RSVD_OPND_FAULT;
			ASTLVL = data;
			break;

/*
		case nICCS:
			if (((ICCS & ICCS_RUN) == 0) && (data & ICCS_STEP))
				ICR++
			if (data & ICCS_XFER)
				ICR = NICR;
			ICCS = (ICCS & ~ICCS_WMASK) | data & ICCS_WMASK;
			ICCS &= ~(data & ICCS_CMASK);
			break;

		case nNICR:
		case nICR:
			PRN(pReg) = data;
			break;
*/

		case nRXCS:
			if (data & RXCS_MBZ)
				RSVD_OPND_FAULT;
			vax_WriteRXCS(vax->Console, data);
			break;

//		case nRXDB:
//			vax_WriteRXDB(vax->Console, data);
//			break;

		case nTXCS:
			if (data & TXCS_MBZ)
				RSVD_OPND_FAULT;
			vax_WriteTXCS(vax->Console, data);
			break;

		case nTXDB:
			vax_WriteTXDB(vax->Console, data);
			break;

		case nMAPEN:
			MAPEN = data & 1;
			break;

		case nTBIA:
			vax_ClearTBTable(0);
			break;

		case nTBIS:
			vax_ClearTBEntry(data);
			break;

		case nTBCHK:
			if (vax_CheckTBEntry(data))
				PSL |= CC_V;
			break;

		default:
#ifdef DEBUG
			if (dbg_Check(DBG_TRACE|DBG_DATA)) {
				dbg_Printf("KA780: (W) Unknown privileged register: %08X (%d)\n",
					pReg, pReg);
			}
#endif /* DEBUG */
//			RSVD_OPND_FAULT;
	}
}

uint32 ka780_ReadIO(void *device, int32 pAddr, int32 size)
{
	KA780_DEVICE *ka780 = (KA780_DEVICE *)device;

	if ((pAddr >= 0x20000000) && (pAddr < 0x20020000)) {
		// Nexus Adaptor
	}

	if ((pAddr >= 0x20100000) && (pAddr < 0x20200000)) {
		// Unibus Address
	}

	return 0;
}

void ka780_WriteIO(void *device, int32 pAddr, uint32 data, int32 size)
{
	KA780_DEVICE *ka780 = (KA780_DEVICE *)device;

	if ((pAddr >= 0x20000000) && (pAddr < 0x20020000)) {
		// Nexus Adaptor
	}

	if ((pAddr >= 0x20100000) && (pAddr < 0x20200000)) {
		// Unibus Address
	}
}

// **************************************************************
// ************** VAX-11/780 System Configuration ***************
// **************************************************************

void *ka780_Create(void)
{
	KA780_DEVICE *ka780;

	ka780 = (KA780_DEVICE *)calloc(1, sizeof(KA780_DEVICE));

	return (void *)ka780;
}

void ka780_Cleanup(VAX_CPU *vax)
{
}

void ka780_Initialize(VAX_CPU *vax)
{
	TXCS = TXCS_RDY;
	SID  = KA780_SYSID;

	// Set up console TTY device (OPA0:)
	vax->Console = vax_ConsoleInit(vax);

	// Set up functions for KA780 CPU (VAX-11/780)
	vax->ReadRegister   = ka780_ReadRegister;
	vax->WriteRegister  = ka780_WriteRegister;
	vax->ReadAligned    = ka780_ReadIO;
	vax->WriteAligned   = ka780_WriteIO;
	vax->ReadUnaligned  = ka780_ReadIO;
	vax->WriteUnaligned = ka780_WriteIO;
}

void ka780_Reset(VAX_CPU *vax)
{
}

// VAX-11/780 System Definition
VAX_SYSTEM vax_System_KA780 =
{
	"KA780",
	"VAX-11/780",
	"v0.5 (Alpha)",
	ka780_Create,
	ka780_Cleanup,
	ka780_Initialize,
	ka780_Reset
};
