// memory.c - Memory Configurations
//
// Copyright (c) 2001, Timothy M. Stark
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// TIMOTHY M STARK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Timothy M Stark shall not
// be used in advertising or otherwise to promote the sale, use or other 
// dealings in this Software without prior written authorization from
// Timothy M Stark.

#include "vax/defs.h"

int vax_InitMemory(VAX_CPU *vax, int reqSize)
{
	// Create physical memory
	vax->RAM     = (char *)malloc(reqSize);
	vax->baseRAM = 0L;
	vax->endRAM  = reqSize - 1;
	vax->sizeRAM = reqSize;

	// Clear all physical memory
	memset(vax->RAM, 0, reqSize);

	return VAX_OK;
}

int vax_FreeMemory(VAX_CPU *vax)
{
	// Free physical memory back to host system.
	if (vax->RAM)
		free(vax->RAM);

	// Reset all memory values;
	vax->RAM     = NULL;
	vax->baseRAM = -1L;
	vax->endRAM  = -1L;
	vax->sizeRAM = 0;

	return VAX_OK;
}

// **************************************
// New Memory Routines for more effective
// **************************************

#define IN_RAM(addr) \
	((addr) < vax->sizeRAM)

// Aligned Memory Access
// Note: It only works for little endian machines at this time.
#define BMEM(addr)  ((int8 *)vax->RAM)[addr]
#define WMEM(addr)  ((int16 *)vax->RAM)[addr]
#define LMEM(addr)  ((int32 *)vax->RAM)[addr]

// Unaligned Memory Access
// Note: It only works for little endian machines at this time.
#define BMEMU(addr) *(int8 *)(&vax->RAM[addr])
#define WMEMU(addr) *(int16 *)(&vax->RAM[addr])
#define LMEMU(addr) *(int32 *)(&vax->RAM[addr])

static const align[]  = { 0x00FFFFFF, 0x00FFFFFF, 0x0000FFFF, 0x000000FF };
static const insert[] = { 0x00000000, 0x000000FF, 0x0000FFFF, 0x00FFFFFF };

// Physical Read Aligned
inline uint32 vax_ReadAligned(uint32 pAddr, int32 len)
{
	if (IN_RAM(pAddr)) {
#ifdef TEST_PARITY
		if (vax->TestParity)
			vax->TestParity(vax, pAddr);
#endif /* TEST_PARITY */
		if (len >= LN_LONG)
			return LMEM(pAddr >> 2);
		if (len == LN_WORD)
			return WMEM(pAddr >> 1);
		return BMEM(pAddr);
	} else {
		// I/O Device Access
		return vax->ReadAligned(vax, pAddr, len);
	}
}

// Physical Write Aligned
inline void vax_WriteAligned(uint32 pAddr, uint32 data, int32 len)
{
	if (IN_RAM(pAddr)) {
#ifdef TEST_PARITY
		if (vax->WriteParity)
			vax->WriteParity(vax, pAddr);
#endif /* TEST_PARITY */
		if (len >= LN_LONG)
			LMEM(pAddr >> 2) = data;
		else if (len == LN_WORD)
			WMEM(pAddr >> 1) = data;
		else
			BMEM(pAddr) = data;
	} else {
		// I/O Device Access
		vax->WriteAligned(vax, pAddr, data, len);
	}
}

// Physical Read Unaligned
inline uint32 vax_ReadUnaligned(uint32 pAddr, int32 len)
{
	if (IN_RAM(pAddr)) {
#ifdef TEST_PARITY
		if (vax->TestParity)
			vax->TestParity(vax, pAddr);
#endif /* TEST_PARITY */
		if (len >= LN_LONG)
			return LMEMU(pAddr);
		if (len == LN_WORD)
			return WMEMU(pAddr);
		return BMEMU(pAddr);
	} else {
		// I/O Device Access
		return vax->ReadUnaligned(vax, pAddr, len);
	}
}

// Physical Write Unaligned
inline void vax_WriteUnaligned(uint32 pAddr, uint32 data, int32 len)
{
	if (IN_RAM(pAddr)) {
#ifdef TEST_PARITY
		if (vax->WriteParity)
			vax->WriteParity(vax, pAddr);
#endif /* TEST_PARITY */
		if (len >= LN_LONG)
			LMEMU(pAddr) = data;
		else if (len == LN_WORD)
			WMEMU(pAddr) = data;
		else
			BMEMU(pAddr) = data;
	} else {
		// I/O Device Access
		vax->WriteUnaligned(vax, pAddr, data, len);
	}
}

// Virtual Read
inline uint32 vax_Read(uint32 vAddr, int32 lint, int32 acc)
{
	uint32 vpn, off, tbi;
	TLBENT xpte;
	uint32 pAddr, pAddr1;
	uint32 wl, wh, sc;

	if (MAPEN) {
		// Translate virtual to physcial address
		vpn = VA_GETVPN(vAddr); // Get Virtual Page Number
		off = VA_GETOFF(vAddr); // Get Byte Offset
		tbi = VA_GETTBI(vpn);   // Get TLB Index

		xpte = (vAddr & VA_S0) ? STLB[tbi] : PTLB[tbi];

		if (((xpte.pte & acc) == 0) || (xpte.tag != vpn) ||
		    ((acc & TLB_WACC) && ((xpte.pte & TLB_M) == 0)))
			xpte = vax_Fill(vAddr, lint, acc, NULL);
		pAddr = (xpte.pte & TLB_PFN) | off;
	} else
		pAddr = vAddr & PAMASK;

	// Check address is aligned first.
	if ((pAddr & (lint - 1)) == 0)
		return vax_ReadAligned(pAddr, lint);

	// No, address is unaligned.
	if (MAPEN && ((off + lint) >= VA_PAGESIZE)) {
		// Address is crossing the page.
		vpn = VA_GETVPN(vAddr + lint);
		tbi = VA_GETTBI(vpn);
		
		xpte = (vAddr & VA_S0) ? STLB[tbi] : PTLB[tbi];

		if (((xpte.pte & acc) == 0) || (xpte.tag != vpn) ||
		    ((acc & TLB_WACC) && ((xpte.pte & TLB_M) == 0)))
			xpte = vax_Fill(vAddr + lint, lint, acc, NULL);
		pAddr1 = (xpte.pte & TLB_PFN) | VA_GETOFF(vAddr + lint);
	} else
		// Read unaligned within a page.
		return vax_ReadUnaligned(pAddr, lint);
//		pAddr1 = (pAddr + lint) & PAMASK;

	// Read unaligned for crossing a page.
	wl = vax_ReadAligned(pAddr, lint);
	wh = vax_ReadAligned(pAddr1, lint);
	if (lint == LN_WORD) {
		return SXTW(ZXTB(wl >> 8) | (ZXTB(wh) << 8));
	} else {
		sc = (pAddr & 03) << 3;
		return (wl >> sc) | (wh << (32 - sc));
	}
}

// Virtual Write
inline void vax_Write(uint32 vAddr, uint32 data, int32 lint, int32 acc)
{
	uint32 vpn, off, tbi;
	TLBENT xpte;
	uint32 pAddr, pAddr1;
	uint32 wl, wh, bo, sc;

	if (MAPEN) {
		// Translate virtual to physcial address
		vpn = VA_GETVPN(vAddr); // Get Virtual Page Number
		off = VA_GETOFF(vAddr); // Get Byte Offset
		tbi = VA_GETTBI(vpn);   // Get TLB Index

		xpte = (vAddr & VA_S0) ? STLB[tbi] : PTLB[tbi];

		if (((xpte.pte & acc) == 0) || (xpte.tag != vpn) ||
		    ((xpte.pte & TLB_M) == 0))
			xpte = vax_Fill(vAddr, lint, acc, NULL);
		pAddr = (xpte.pte & TLB_PFN) | off;
	} else
		pAddr = vAddr & PAMASK;


	// Check address is aligned first.
	if ((pAddr & (lint - 1)) == 0)
		return vax_WriteAligned(pAddr, data, lint);

	// No, address is unaligned.
	if (MAPEN && ((off + lint) >= VA_PAGESIZE)) {
		// Address is crossing the page.
		vpn = VA_GETVPN(vAddr + lint);
		tbi = VA_GETTBI(vpn);
		
		xpte = (vAddr & VA_S0) ? STLB[tbi] : PTLB[tbi];

		if (((xpte.pte & acc) == 0) || (xpte.tag != vpn) ||
		    ((xpte.pte & TLB_M) == 0))
			xpte = vax_Fill(vAddr + lint, lint, acc, NULL);
		pAddr1 = (xpte.pte & TLB_PFN) | VA_GETOFF(vAddr + lint);
	} else
		// Write unaligned within a page.
		return vax_WriteUnaligned(pAddr, data, lint);
//		pAddr1 = (pAddr + lint) & PAMASK;

	// Write unaligned that is crossing a page.
	wl = vax_ReadAligned(pAddr, lint);
	wh = vax_ReadAligned(pAddr1, lint);
	if (lint == LN_WORD) {
		wl = (wl & 0x00FF) | (data << 8);
		wh = (wh & 0xFF00) | (data >> 8);
	} else {
		bo = pAddr & 03;
		sc = bo << 3;
		wl = (wl & insert[bo])  | (data << sc);
		wh = (wh & ~insert[bo]) | (data >> (32 - sc));
	}

	vax_WriteAligned(pAddr, wl, lint);
	vax_WriteAligned(pAddr1, wh, lint);
}

// Virtual Access Check
int32 vax_Test(uint32 vAddr, int32 acc, int32 *status)
{
	uint32 vpn, off, tbi;
	TLBENT xpte;

	*status = MM_OK; // Assume Ok.
	if (MAPEN) {
		vpn = VA_GETVPN(vAddr); // Get Virtual Page Number
		off = VA_GETOFF(vAddr); // Get Byte Offset
		tbi = VA_GETTBI(vpn);   // Get TLB Index

		xpte = (vAddr & VA_S0) ? STLB[tbi] : PTLB[tbi];

		if (((xpte.pte & acc) == 0) || (xpte.tag != vpn))
			xpte = vax_Fill(vAddr, LN_BYTE, acc, status);
		return (*status == MM_OK) ? ((xpte.pte & TLB_PFN) | off) : -1;
	} else
		return vAddr & PAMASK;
}

// Prefetch instruction buffer
// Warning: This routine only work for little endian machines

#define BIBUF(idx) *((int8 *)&IBUF[idx])
#define WIBUF(idx) *((int16 *)&IBUF[idx])
#define LIBUF(idx) *((int32 *)&IBUF[idx])

int32 vax_ReadInst(int32 size)
{
	uint32 data;
	int32  t;

	// If buffer is short, fetch data from main memory.
	// To flush instruction buffer, must set IBCNT as 0
	// and IBPPC as -1 to refill instruction buffer 
	// for branch and call instructions.

	if ((IBIDX + size) > IBCNT) {
		// Refill instruction buffer after flush.
		if ((IBPPC < 0) || (VA_GETOFF(IBPPC) == 0))  {
			IBPPC = vax_Test((PC + IBCNT) & ~03, RA, &t);
			if (t != MM_OK)
				vax_Read((PC + IBCNT), LN_LONG, RA); // Force page fault trap
			IBIDX = PC & 03;
		}
		LIBUF(IBCNT) = vax_ReadAligned(IBPPC, LN_LONG);
		IBPPC += LN_LONG;
		IBCNT += LN_LONG;
	}

	// Extract data from the instruction buffer
	if (size == LN_BYTE)
		data = BIBUF(IBIDX);
	else if (size == LN_WORD)
		data = WIBUF(IBIDX);
	else
		data = LIBUF(IBIDX);

	// Increment Program Counter (PC)
	PC    += size;
	IBIDX += size;
	if (IBIDX >= LN_LONG) {
		LIBUF(0) = LIBUF(LN_LONG);
		IBIDX -= LN_LONG;
		IBCNT -= LN_LONG;
	}

#ifdef DEBUG
	if (dbg_Check(DBG_PREFETCH)) {
		dbg_Printf("PRE: Data=%08X PC=%08X PPC=%08X IDX=%d CNT=%d SIZE=%d\n",
			data, PC, IBPPC, IBIDX, IBCNT, size);
		if (IBCNT > LN_LONG)
			dbg_Printf("PRE: Buffer=%08X %08X\n", LIBUF(0), LIBUF(4));
		else
			dbg_Printf("PRE: Buffer=%08X\n", LIBUF(0));
	}
#endif /* DEBUG */

	return data;
}

// ******************* Console Read/Write Access *********************

// Console Read Access
int32 vax_ReadC(register VAX_CPU *vax, uint32 vAddr, uint32 *data, int32 size, uint32 sw)
{
	int32  pAddr, pAddr1; // Physical Address
	uint32 wl, wh, sc;    // Data Low and High
	int32  sts = MM_OK;   // Access Status - Assume Ok.

	if (MAPEN && (sw & SWMASK('v'))) {
		if ((pAddr = vax_Test(vAddr, CRA, &sts)) < 0)
			return sts;

		// Check address is aligned first.
		if ((pAddr & (size - 1)) == 0)
			return vax->ReadCA(vax, pAddr, data, size);

		if ((VA_GETOFF(vAddr) + size) >= VA_PAGESIZE) {
			if ((pAddr1 = vax_Test(vAddr + size, CRA, &sts)) < 0)
				return sts;
		} else
			return vax->ReadCU(vax, pAddr, data, size);

		// Fetch two words for crossing a page.
		if ((sts = vax->ReadCA(vax, pAddr1, &wh, size)) != MM_OK)
			return sts;
		if ((sts = vax->ReadCA(vax, pAddr, &wl, size)) != MM_OK)
			return sts;

		// Extract unaligned data.
		if (size == LN_WORD) {
			*data = (wl >> 8) | (wh << 8);
		} else {
			sc = (pAddr & 03) << 3;
			*data = (wl >> sc) | (wh << (32 - sc));
		}
	} else
		return vax->ReadCU(vax, vAddr, data, size);

	return sts;
}

// Console Write Access
int32 vax_WriteC(register VAX_CPU *vax, uint32 vAddr, uint32 data, int32 size, uint32 sw)
{
	int32  pAddr, pAddr1;  // Physical Address
	uint32 wl, wh, bo, sc; // Unaligned Data
	int32  sts = MM_OK;    // Access Status - Assume Ok.

	if (MAPEN && (sw & SWMASK('v'))) {
		if ((pAddr = vax_Test(vAddr, CRA, &sts)) < 0)
			return sts;

		// Check address is aligned first.
		if ((pAddr & (size - 1)) == 0)
			return vax->WriteCA(vax, pAddr, data, size);

		if ((VA_GETOFF(vAddr) + size) >= VA_PAGESIZE) {
			if ((pAddr1 = vax_Test(vAddr + size, CRA, &sts)) < 0)
				return sts;
		} else
			return vax->WriteCU(vax, pAddr, data, size);

		// Fetch two words for crossing a page.
		if ((sts = vax->ReadCA(vax, pAddr1, &wh, size)) != MM_OK)
			return sts;
		if ((sts = vax->ReadCA(vax, pAddr, &wl, size)) != MM_OK)
			return sts;

		// Insert unaligned data.
		if (size == LN_WORD) {
			wl = (wl & 0x00FF) | (data << 8);
			wh = (wh & 0xFF00) | (data >> 8);
		} else {
			bo = pAddr & 03;
			sc = bo << 3;
			wl = (wl & insert[bo])  | (data << sc);
			wh = (wh & ~insert[bo]) | (data >> (32 - sc));
		}

		// Send them back to memory.
		if ((sts = vax->WriteCA(vax, pAddr1, wh, size)) != MM_OK)
			return sts;
		if ((sts = vax->WriteCA(vax, pAddr, wl, size)) != MM_OK)
			return sts;
	} else
		return vax->WriteCU(vax, vAddr, data, size);

	return sts;
}

// Console Read Access with Instruction Buffer 
// Warning: This routine only work for little endian machines

#define BCIBUF(idx) *((uint8 *)&CIBUF[idx])
#define WCIBUF(idx) *((uint16 *)&CIBUF[idx])
#define LCIBUF(idx) *((uint32 *)&CIBUF[idx])

int32 vax_ReadCI(register VAX_CPU *vax, uint32 vAddr, uint32 *data, int32 size,
int32 sw)
{
	int32 sts; // Memory Access Status

	// If buffer is short, fetch data from main memory.
	// Use different virutal address rather than
	// following address to refill instruction buffer.

	if (vAddr != CVADDR) {
		CIBCNT = 0;
		CIBPPC = -1;
	}

	if ((CIBIDX + size) > CIBCNT) {
		// Refill instruction buffer after flush.
		if ((CIBPPC < 0) || (VA_GETOFF(CIBPPC) == 0))  {
			CIBPPC = vax_Test((vAddr + CIBCNT) & ~03, RA, &sts);
			if (sts != MM_OK)
				return sts;
			CIBIDX = vAddr & 03;
		}
		LCIBUF(CIBCNT) = vax_ReadAligned(CIBPPC, LN_LONG);
		CIBPPC += LN_LONG;
		CIBCNT += LN_LONG;
	}

	// Extract data from the instruction buffer
	if (size == LN_BYTE)
		*data = BCIBUF(CIBIDX);
	else if (size == LN_WORD)
		*data = WCIBUF(CIBIDX);
	else
		*data = LCIBUF(CIBIDX);

	// Increment Program Counter (PC)
	CVADDR += size;
	CIBIDX += size;
	if (CIBIDX >= LN_LONG) {
		LCIBUF(0) = LCIBUF(LN_LONG);
		CIBIDX -= LN_LONG;
		CIBCNT -= LN_LONG;
	}

	return sts;
}
