// system.c - VAX System Configurations
//
// Written by
//  Timothy Stark <sword7@speakeasy.org>
//
// This file is part of the TS-10 Emulator.
// See README for copyright notice.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "vax/defs.h"
#include "emu/socket.h"

#define VAX_KEY     "VAX"
#define VAX_NAME    "VAX-11 Series Emulator"
#define VAX_VERSION "v0.8 (Alpha)"

VAX_CPU *vax = NULL;

// VAX System Table
extern DEVINFO vax_System_KA630;
//extern DEVINFO vax_System_KA650;
//extern DEVINFO vax_System_KA655;
//extern DEVINFO vax_System_KA780;
//extern DEVINFO vax_System_KA790;

DEVINFO *vax_Systems[] =
{
	&vax_System_KA630,  // MicroVAX II Series
//	&vax_System_KA650,  // CVAX Series
//	&vax_System_KA655,  // MicroVAX 3800/3900 Series
//	&vax_System_KA780,  // VAX-11/780 Series
//      &vax_System_KA790,  // VAX 8600 Series
	NULL
};

DEVINFO vax_System; // Look at bottom of this file.

// Load VMB.EXE into VAX memory
int vax_LoadBoot(VAX_SYSTEM *sys, char *fn, uint32 sAddr)
{
	VAX_CPU *vax = sys->Processor;
	int vmbFile;
	int eAddr;
	int idx;
	int st = 1;

	if ((vmbFile = open(fn, O_RDONLY)) < 0)
		return EMU_OPENERR;

	for (idx = sAddr; st > 0; idx += 512) {
		st = read(vmbFile, &vax->RAM[idx], 512);
	}
	eAddr = idx;

	close(vmbFile);

	printf("Bootstrap code %s loaded from %08X to %08X\n",
		fn, sAddr, eAddr);

	R0  = 0;
	R1  = 0;
	R2  = 0;
	R3  = 0;
	R4  = 0;
	R5  = 0;
	R6  = 0;
	R7  = 0;
	R8  = 0;
	R9  = 0;
	R10 = 0;
	R11 = 0;
	AP  = 0;
	FP  = 0;
	SP  = sAddr;
	PC  = sAddr;

	// Initial Settings for VMB.EXE
	IPL    = 31;
	ASTLVL = 4;
	PSL    = 0x04150000;

	return VAX_OK;
}

// Load ROM image into VAX memory
int vax_LoadROM(VAX_SYSTEM *sys, char *fn, uint32 sAddr)
{
	VAX_CPU *vax = sys->Processor;
	struct stat st;
	int romFile;
	int baseAddr, lenAddr;
	int idx;

	if ((romFile = open(fn, O_RDONLY)) < 0)
		return EMU_OPENERR;

	// Set up starting address and ROM image size.
	fstat(romFile, &st);
	baseAddr = sAddr;
	lenAddr  = st.st_size;

	vax->ROM     = (uint8 *)malloc(lenAddr);
	vax->baseROM = baseAddr;
	vax->endROM  = (baseAddr + lenAddr) - 1;
	vax->sizeROM = lenAddr;
	vax->maskROM = lenAddr - 1;

	for (idx = 0; idx < lenAddr; idx += 32768)
		read(romFile, &vax->ROM[idx], 32768);

	close(romFile);

	PC = baseAddr;

	return VAX_OK;
}

// Load NVRAM image into VAX memory
int vax_LoadNVRAM(char *fn)
{
}

// Dump contents into terminal from memory area
int vax_Dump(VAX_CPU *vax, SOCKET *dbg, uint32 *Addr, uint32 eAddr, uint32 sw)
{
	int     idx;
	char    ascBuffer[16];
	char    *pasc;
	uint32  data;

	while (*Addr <= eAddr) {
		SockPrintf(dbg, "%08X: ", *Addr);
		pasc = ascBuffer;
		for (idx = 0; (idx < 16) && (*Addr <= eAddr); idx++) {
			vax_ReadC(vax, (*Addr)++, &data, OP_BYTE, sw);
			SockPrintf(dbg, "%02X%c", data, (idx == 7) ? '-' : ' ');
			*pasc++ = ((data >= 32) && (data < 127)) ? data : '.';
		}
		*pasc = '\0';
		SockPrintf(dbg, " |%-16s|\n", ascBuffer);
	}

	return VAX_OK;
}

void vax_CheckInstructions(void)
{
	char outBuffer[80];
	int  done, total;
	int  idx;

	outBuffer[0] = 0;
	total = 0;
	done  = 0;

	printf("Following instructions that need to be implemented:\n");
	for (idx = 0; vax_Instruction[idx].Name; idx++) {
		total++;
		if (vax_Instruction[idx].Execute == NULL) {
			if (strlen(outBuffer) > 70) {
				outBuffer[strlen(outBuffer)] = '\0';
				printf("%s\n", outBuffer);
				outBuffer[0] = '\0';
			}

			if (outBuffer[0])
				strcat(outBuffer, ", ");
			strcat(outBuffer, vax_Instruction[idx].Name);
		} else
			done++;
	}

	if (outBuffer[0]) {
		outBuffer[strlen(outBuffer)] = '\0';
		printf("%s\n", outBuffer);
	}

	printf("\n%d of %d instructions had been implemented.\n", done, total);
}

// ****************************************************************

// Usage: create <device> <model> ...
void *vax_Create(MAP_DEVICE *nmap, MAP_DEVICE *map, int argc, char **argv)
{
	VAX_SYSTEM  *vax;
	VAX_CPU     *cpu;
	DEVINFO     *dptr, *cptr, *iptr;
#ifdef DEBUG
	DBG_BRKSYS  *newBreaks;
#endif /* DEBUG */
	int         nDevices;
	int         idx;

	if (nmap == NULL) {
		// Set up new VAX system

		if (vax = (VAX_SYSTEM *)calloc(1, sizeof(VAX_SYSTEM))) {

			// Set Up System Identification
			vax->sysName  = VAX_KEY;
			vax->Descrip  = VAX_NAME;
			vax->Version  = VAX_VERSION;

			map->dtName     = vax_System.dtName;
			map->emuName    = vax_System.emuName;
			map->emuVersion = vax_System.emuVersion;
			map->devInfo    = &vax_System;
			map->Device     = vax;
			map->sysDevice  = vax;
			map->sysMap     = map;
		}

		if (vax) {
			printf("Created %s on %s: - %s %s\n",
				map->dtName, map->devName, map->emuName, map->emuVersion);
		} else {
			printf("VAX: *** Can't Intiialize System. Aborting. ***\n");

			if (vax)
				free(vax);

			return NULL;
		}

		return vax;
	} else {
		if (cptr = GetDeviceInfo(map, argc, argv)) {
			vax = (VAX_SYSTEM *)map->Device;
			cpu = (VAX_CPU *)calloc(1, sizeof(VAX_CPU));

			cpu->System    = vax;
			cpu->Processor = cptr->Create(NULL, nmap, argc, argv);
			cpu->Console   = vax_ConsoleInit(cpu);
			vax->Processor = cpu;

			nmap->sysDevice = map->sysDevice;
			nmap->sysMap    = map->sysMap;
#ifdef DEBUG
			nmap->Breaks    = &cpu->Breaks;
#endif /* DEBUG */

			// Initialize Processor
			cptr->Setup(cpu->Processor, cpu);

			vax_InitMemory(cpu, 16384 * 1024);
#ifdef DEBUG
			vax_InitDisasm();
#endif /* DEBUG */

			return cpu->Processor;
		}
		return NULL;
	}
}

DEVINFO vax_System =
{
	VAX_KEY,      // System Device Name
	VAX_NAME,     // Emulator Name
	VAX_VERSION,  // Emulator Version
	vax_Systems,  // Listing of VAX Systems
	DF_SELECT,    // Device Flags
	DT_SYSTEM,    // Device Type

	// Command Table for VAX system
	vax_Commands,     // Commands
	vax_SetCommands,  // Set Commands
	vax_ShowCommands, // Show Commands

	vax_Create,       // Create Routine
	NULL,             // Delete Routine
	NULL,             // Setup Routine
	NULL,             // Reset Routine
	NULL,             // Attach Routine (Not Used)
	NULL,             // Detach Routine (Not Used)
	NULL,             // Info Routine
	NULL,             // Boot Routine
	vax_Execute,      // Execute Routine
};
